/*
	$Id: bookmaker.c,v 1.8 2011/04/18 09:45:59 alex Exp $
	bookmaker.c--- make book
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: bookmaker.c,v 1.8 2011/04/18 09:45:59 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<string.h>
#include	<signal.h>
#include	"edb/edb.h"
#include	"edb/html.h"
#include	"edb/publish.h"

int
edb_make_book_section(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	int	i, t;
	char	str[BUFSIZ];

	if(sec->sec==NULL) return FALSE;
	HTML(sec->ep->doc) if(book->ep != sec->ep) doc_printf(sec->ep->doc, "<h2>%s %s</h2>\n", sec->secno, sec->title);

	for(i=0; sec->sec[i].title; ++i) {
		Section	subsec = &sec->sec[i];
		char	name[BUFSIZ];
		char	name_tmp[BUFSIZ];
		char	secno[BUFSIZ];
		char	title[BUFSIZ];

		if(sec->layer==0) {
			sprintf(subsec->secno, "%s.", edb_book_int2numstr(secno, i+1, sec->number));
		} else {
			sprintf(subsec->secno, "%s%s", sec->secno, edb_book_int2numstr(secno, i+1, sec->number));
		}
		if(subsec->name) {
			HTML(sec->ep->doc) doc_printf(sec->ep->doc, "<li><a href=\"%s.html\">%s</a>\n", subsec->name, subsec->title);
			PLAIN(sec->ep->doc) sprintf(name, "%s.txt", subsec->name);
			HTML(sec->ep->doc) sprintf(name, "%s.html", subsec->name);
			TeX(sec->ep->doc) sprintf(name, "%s.tex", subsec->name);
			subsec->ep = edb_print_dup(sec->ep);
			sprintf(name_tmp, "%s.tmp", subsec->name);
			subsec->ep->doc = doc_open(sec->ep->doc->dml, name_tmp, "w+");
			if(subsec->ep->doc==NULL) {
				fprintf(book->err_fp, "cannot create file.\n");
				return FALSE;
			}
			sprintf(title, "%s %s", subsec->secno, subsec->title);
			HTML(subsec->ep->doc) edb_book_begin_HTML(subsec->ep, book, title);
		} else {
			HTML(sec->ep->doc) doc_printf(sec->ep->doc, "<li>%s\n", subsec->title);
			subsec->ep = sec->ep;
		}
		subsec->layer = sec->layer+1;
		HTML(subsec->ep->doc) {
			if(subsec->func) (*subsec->func)(book, subsec);
		}
		TeX(subsec->ep->doc) {
			doc_printf(subsec->ep->doc, "\\begin{section}{%s}\n", subsec->title);
			if(sec->ep->doc != subsec->ep->doc) doc_printf(sec->ep->doc, "\\input{%s.tex}\n", subsec->name);
			if(subsec->func) (*subsec->func)(book, subsec);
			doc_puts(subsec->ep->doc, "\\end{section}\n");
		}
		if(subsec->name) {
			HTML(subsec->ep->doc) edb_book_end_HTML(subsec->ep, book);
			doc_close(subsec->ep->doc);
			file_update(name_tmp, name);
		}
		TeX(sec->ep->doc) doc_puts(sec->ep->doc, "\\newpage\n");
	}
	return TRUE;
}

int
edb_make_book(EdbContext ec, Book book, int style, int dml, FILE *err_fp)
{
	int	i;
	char	str[BUFSIZ];
	char	name_tmp[BUFSIZ];
	char	name[BUFSIZ];
	Doc	doc, tmpdoc;

	makedir_hierarchy(book->dir);
	if(chdir(book->dir)) {
		perror("edb_make_book");
		return FALSE;
	}
	if(dml==DocML_PLAIN) sprintf(name, "%s.txt", book->name);
	if(dml==DocML_HTML) sprintf(name, "%s.html", book->name);
	if(dml==DocML_TeX) sprintf(name, "%s.tex", book->name);
	sprintf(name_tmp, "%s.tmp", book->name);
	doc = doc_open(dml, name_tmp, "w+");
	if(doc==NULL) {
		fprintf(err_fp, "cannot create file.\n");
		return FALSE;
	}
	book->ep = edb_print_new(ec, doc, style);
	tmpdoc = doc_tmpopen(dml);
	book->tmpep = edb_print_new(ec, tmpdoc, style);
	book->err_fp = err_fp;

	switch(book->ep->doc->dml) {
	case DocML_HTML:
		edb_book_begin_HTML(book->ep, book, book->title);
		doc_printf(book->ep->doc, "<h1>%s</h1>\n", book->title);
		doc_printf(book->ep->doc, "%s\n", book->subtitle);
		doc_puts(book->ep->doc, "<hr />\n");
		doc_puts(book->ep->doc, "<h2>目　次</h2>\n");
		doc_puts(book->ep->doc, "<div class=\"contents\">\n");
		break;
	case DocML_TeX:
		book->ep->ct->nocolor = TRUE;
#if 1
		book->ep->ct->noalert = TRUE;
#endif
		break;
	default:
		break;
	}
	for(i=0; book->part[i].title; ++i) {
		Section	part = &book->part[i];
		char	name_tmp[BUFSIZ];
		char	name[BUFSIZ];
		char	secno[BUFSIZ];
		char	title[BUFSIZ];

		if(book->number == NUM_none) {
			strcpy(part->secno, "");
		} else {
			sprintf(part->secno, "%s.", edb_book_int2numstr(secno, i+1, book->number));
		}
		part->layer = 1;
		if(part->name) {
			HTML(book->ep->doc) doc_printf(book->ep->doc, "<h3><a href=\"%s.html\">%s</a></h3>\n", part->name, part->title);
			PLAIN(book->ep->doc) sprintf(name, "%s.txt", part->name);
			HTML(book->ep->doc) sprintf(name, "%s.html", part->name);
			TeX(book->ep->doc) sprintf(name, "%s.tex", part->name);
			part->ep = edb_print_dup(book->ep);
			sprintf(name_tmp, "%s.tmp", part->name);
			part->ep->doc = doc_open(book->ep->doc->dml, name_tmp, "w+");
			if(part->ep->doc==NULL) {
				fprintf(book->err_fp, "cannot create file.\n");
				edb_print_free(book->ep);
				return FALSE;
			}
			sprintf(title, "%s %s", part->secno, part->title);
			HTML(part->ep->doc) edb_book_begin_HTML(part->ep, book, title);
		} else {
			HTML(book->ep->doc) doc_printf(book->ep->doc, "<h3>%s</h3>\n", part->title);
			part->ep = book->ep;
		}
		PLAIN(part->ep->doc) {
			if(part->func) (*part->func)(book, part);
		}
		HTML(part->ep->doc) {
			if(part->func) (*part->func)(book, part);
		}
		TeX(part->ep->doc) {
			doc_printf(book->ep->doc, "\\begin{章}{%s}\n", part->title);
			if(book->ep->doc != part->ep->doc) doc_printf(book->ep->doc, "\\input{%s.tex}\n", part->name);
			if(part->func) (*part->func)(book, part);
			doc_puts(book->ep->doc, "\\end{章}\n\n");
		}
		if(part->name) {
			HTML(part->ep->doc) edb_book_end_HTML(part->ep, book);
			doc_close(part->ep->doc);
			file_update(name_tmp, name);
		}
	}

	switch(book->ep->doc->dml) {
	case DocML_HTML:
		doc_puts(book->ep->doc, "</div>\n");
		edb_book_end_HTML(book->ep, book);
		break;
	case DocML_TeX:
		break;
	default:
		break;
	}
	doc_close(book->ep->doc);
	file_update(name_tmp, name);
	edb_print_free(book->ep);
	return TRUE;
}


