/*
	$Id: check.c,v 1.33 2011/04/18 09:45:59 alex Exp $
	check.c--- check information
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: check.c,v 1.33 2011/04/18 09:45:59 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<string.h>
#include	<signal.h>
#include	"edb/edb.h"
#include	"edb/html.h"
#include	"edb/publish.h"

#define	STATUS_NORMAL	EP_STATUS_NORMAL
#define	STATUS_WARNING	EP_STATUS_WARNING
#define	STATUS_ALERT	EP_STATUS_ALERT

static	int	check_person __P((Book, Section));
static	int	check_study __P((Book, Section));
static	int	check_academicsocietyactivity __P((Book, Section));
static	int	check_socialactivity __P((Book, Section));
static	int	check_article __P((Book, Section));
static	int	check_patent __P((Book, Section));
static	int	check_researchactivity __P((Book, Section));
static	int	check_conference __P((Book, Section));
static	int	check_prize __P((Book, Section));
static	int	check_collaboration __P((Book, Section));
static	int	check_thesis __P((Book, Section));
static	int	check_lecture __P((Book, Section));
static	int	check_history __P((Book, Section));
static	int	check_posthistory __P((Book, Section));
static	int	check_teaching __P((Book, Section));
static	int	check_picture __P((Book, Section));
static	int	check_default __P((Book, Section));

#if 0
static	eid_t	teacher_title[] = TITLE_TEACHER_LIST ;
#endif

void	edb_book_check_prepare_sections	__P((EdbContext));

int
edb_book_person_is_teacher(EdbContext ec, eid_t eid)
{
	EdbTupleInfo	tuple;
	EdbTupleColumn	tc;
	EdbDatum	datum;
	int	k;

	tuple = edb_tuple_get(ec, eid, 0);
	tc = TupleColumn_person_title(ec, tuple);
	if(tc==NULL) return FALSE;
	datum = TCDatumFirst(tc);
	if(!DatumEIDisValid(datum)) return FALSE;
#if 0
	for(k=0; k<Ns(teacher_title); ++k) {
		if(EIDisVALID(teacher_title[k]) && DatumEID(datum)==teacher_title[k]) return TRUE;
	}
#else
	for(k=0; edb_title_teacher_list[k]; ++k) {
		if(EIDisVALID(edb_title_teacher_list[k]) && DatumEID(datum)==edb_title_teacher_list[k]) return TRUE;
	}
#endif
	return FALSE;
}

struct SECTION	*check_part;

struct SECTION	check_part_special[] = {
{	"person",		"【個人】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_person,		NULL,	0	},
{	"history",		"【履歴】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_history,		NULL,	0	},
{	"teaching",	"【担当授業】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_teaching,	NULL,	0	},
{	"study",		"【研究】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_study,		NULL,	0	},
{	"academicsocietyactivity",		"【学会活動】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_academicsocietyactivity,		NULL,	0	},
{	"socialactivity",			"【社会活動】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_socialactivity,			NULL,	0	},
{	"posthistory",	"【役歴】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_posthistory,	NULL,	0	},
{	"article",		"【著作】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_article,		NULL,	0	},
{	"prize",		"【賞】情報",		ORGAN_UNIV_EID,	NUM_arabic,	check_prize,		NULL,	0	},
{	"picture",		"【画像】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_picture,		NULL,	0	},
{	"patent",		"【特許・実用新案】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_patent,		NULL,	0	},
{	"researchactivity",	"【研究活動】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_researchactivity,	NULL,	0	},
{	"conference",	"【集会】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_conference,	NULL,	0	},
{	"collaboration",	"【共同研究】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_collaboration,	NULL,	0	},
{	"thesis",		"【学位論文】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_thesis,		NULL,	0	},
{	"lecture",	"【講義概要】情報",	ORGAN_UNIV_EID,	NUM_arabic,	check_lecture,	NULL,	0	},
{	NULL,		NULL,			0,		NUM_none,	NULL,			NULL,	0	}
};

struct BOOK	book_check = {
	WEB_DOCUMENT_ROOT "/check",
	"index",
	NAME_OF_DB "登録情報チェック",
	"（登録情報の誤り検出結果です（1日1回更新）．ただし，自動検出が可能なものに限られていますので，完全ではありません）",
	ORGAN_UNIV_EID,
	NUM_none,
	{YMD_DATE(2001, 4, 1), YMD_DATE(2002, 3, 31)},
	{YMD_DATE(2001, 4, 1), YMD_DATE(2002, 3, 31)},
	NULL
};

void
edb_book_check_prepare_sections(EdbContext ec)
{
	Link	l, l_tables;
	int	i;

	l_tables = edb_table_get_tables(ec);
	check_part = (struct SECTION *)malloc(sizeof(struct SECTION)*(link_atoms(l_tables)+1));
	bzero(check_part, sizeof(struct SECTION)*(link_atoms(l_tables)+1));
	i = 0;
	ForLink(l, l_tables) {
		EdbTableInfo ti = l->d;
		char	*nm, *xn;
		int	(*func) __P((Book, struct SECTION *));
		int	k;

		if(ti->auxiliary) continue;
		xn = (char *)TABLE_XN(ti);
		check_part[i].name = xn;
		asprintf(&nm, "【%s】情報", TABLE_Name_ja(ti));
		check_part[i].title = nm;
		check_part[i].organ_eid = ORGAN_UNIV_EID;
		check_part[i].number = NUM_arabic;
		func = check_default;
		for(k=0; k<Ns(check_part_special); ++k) {
			if(!check_part_special[k].name) continue;
			if(!strcmp(xn, check_part_special[k].name)) {
				func = check_part_special[k].func;
				break;
			}
		}
		check_part[i].func = func;
		++i;
	}
	book_check.part = check_part;
	link_free(l_tables);
}

struct CHECK_LIST {
	char	*column;
	char	*description;
	EdbBookPrintFunc func;
};

struct CHECKER {
	EdbTableInfo	ti;
	Link			l_info;
	BiTree		printed;
};

static
void
show_edbclient_button(Doc doc, int eid)
{

#if defined(EDB_PROVIDE_EDBCLIENT)
	doc_printf(doc,
		"<input type=\"button\" value=\"→%s\" title=\"%sを起動します．\" onclick=\"window.open('%s?cap=%d&args=%d','edb_processing','width=384,height=128,resizable=yes,scrollbars=yes');\" style=\"font-size:smaller\"> ",
		EDB_CLIENT_NAME, EDB_CLIENT_NAME,
		EDB_CLIENT_INVOKE_CGI, eid, eid);
#endif	/* EDB_PROVIDE_EDBCLIENT */
}

static
int
check_begin(EdbPrint ep, Section sec, struct CHECKER *chk)
{
	register Doc	doc = ep->doc;
	register EdbContext	ec = ep->ec;
	Link	l;

	doc_printf(doc, "<h2>%s %s</h2>\n", sec->secno, sec->title);
	doc_puts(doc, "<hr />\n");
	chk->ti = edb_table_get_by_name(ec, (utf8_t *)sec->name);
	if(chk->ti==NULL) return FALSE;
	chk->l_info = edb_book_get_tuple_list(ec, chk->ti, NULL);
	doc_printf(doc, "【%s】には現在%d件の情報が登録されています．\n", TABLE_Name_ja(chk->ti), link_atoms(chk->l_info));
	doc_puts(doc, "<hr />\n");
	doc_printf(doc, "(注)下記の検査結果は，学内公開可能な登録内容のみを抽出し，検査した結果です．\n");
	doc_puts(doc, "<hr />\n");
	doc_printf(doc, "<dl>\n");
	chk->printed = NULL;
	doc_printf(doc, "<dt>");
	doc_printf(doc, "<b>項目がひとつも定義されていない情報</b>\n");
	ForLink(l, chk->l_info) {
		EdbTupleInfo	tuple = (EdbTupleInfo)l->d;
		EdbDatum	d;
		char	*img = NULL;

		if(tuple->l_column) continue;

		doc_puts(doc, "<dd style=\"font-size:smaller\">");
		doc_printf(doc, "<a href=\"%s%s?EID=%d\" target=\"_blank\">%s</a>", WEB_SERVER_URL, ec->browser_url, TupleEID(tuple),
			img=html_make_icon_img_tag(ec, EDB_ICON_TUPLE_SMALL, "[%d]", TupleEID(tuple)));
		FREE(img);
		show_edbclient_button(doc, TupleEID(tuple));
		edb_book_print_not_implement(ep, tuple);
		BiTreeAdd(chk->printed, tuple, TupleEID(tuple), NULL, NULL, NULL);
		doc_puts(doc, "\n");
	}
	return TRUE;
}

static
int
check_end(EdbPrint ep, Section sec, struct CHECKER *chk)
{
	register Doc	doc = ep->doc;
	register EdbContext	ec = ep->ec;

	doc_printf(doc, "</dl>\n");
	bitree_free(chk->printed, NULL);
	link_free(chk->l_info);
	return TRUE;
}

static
void
check_column(EdbPrint ep, struct CHECK_LIST *check_list, struct CHECKER *chk)
{
	register Doc	doc = ep->doc;
	register EdbContext	ec = ep->ec;
	Link	l;
	EdbTupleColumn	tc;
	EdbColumnInfo		ci;

	ci = edb_table_seek_column(ec, chk->ti, (utf8_t *)check_list->column);
	if(ci==NULL) {
		edb_trace(ec, "check_column: can't find column (%s).\n", check_list->column);
		return;
	}
	doc_printf(doc, "<dt>");
	doc_puts(doc, "上記以外で，<br />\n");
	doc_printf(doc, "<b>項目[%s]が定義されていない情報</b>\n", ColumnName_ja(ci));
	if(check_list->description) doc_printf(doc, "（%s）\n", check_list->description);

	ForLink(l, chk->l_info) {
		EdbTupleInfo	tuple = (EdbTupleInfo)l->d;
		EdbDatum	d;
		int	print = FALSE;

		if(bitree_search(chk->printed, tuple, TupleEID(tuple), NULL)) continue;
		tc = edb_tuple_seek_column(ec, tuple, (utf8_t *)check_list->column);
		if(tc==NULL || tc->datum==NULL) {
			print = TRUE;
		} else {
			d = tc->datum;
			if(!DatumAvail(d)) print = TRUE;
		}
		if(print) {
			char	*img = NULL;

			doc_puts(doc, "<dd style=\"font-size:90%;\">");
			doc_printf(doc, "<a href=\"%s%s?EID=%d\" target=\"_blank\">%s</a>", WEB_SERVER_URL, ec->browser_url, TupleEID(tuple),
				img=html_make_icon_img_tag(ec, EDB_ICON_TUPLE_SMALL, "[%d]", TupleEID(tuple)));
			FREE(img);
			show_edbclient_button(doc, TupleEID(tuple));
			if(check_list->func) (*check_list->func)(ep, tuple);
			BiTreeAdd(chk->printed, tuple, TupleEID(tuple), NULL, NULL, NULL);
			doc_puts(doc, "</dd>\n");
			doc_puts(doc, "\n");
		}
	}
}

static
void
check_samename(EdbPrint ep, EdbTupleInfo person, Link l_info)
{
	register Doc		doc = ep->doc;
	register EdbContext	ec = ep->ec;
	Link				l;
	EdbTupleColumn	tc;
	EdbDatum			d;
	char	*s = NULL, *g = NULL;

	if(person==NULL || l_info==NULL) return;
	tc = edb_tuple_seek_column(ec, person, (utf8_t *)"@.surname");
	if(tc && (d = tc->datum)) s = (char *)DatumJapanese(d);
	tc = edb_tuple_seek_column(ec, person, (utf8_t *)"@.givenname");
	if(tc && (d = tc->datum)) g = (char *)DatumJapanese(d);
	if(!s && !g) return;

	ForLink(l, l_info) {
		EdbTupleInfo		p = l->d;
		char	*name = NULL;
		char	*img = NULL;

		if(person==p) continue;

		tc = edb_tuple_seek_column(ec, p, (utf8_t *)"@.surname");
		if(tc && (d = tc->datum)) name = (char *)DatumJapanese(d);
		if(!name && s) continue;
		if(name && !s) continue;
		if(name && s && strcmp(name, s)) continue;

		tc = edb_tuple_seek_column(ec, p, (utf8_t *)"@.givenname");
		if(tc && (d = tc->datum)) name = (char *)DatumJapanese(d);
		if(!name && g) continue;
		if(name && !g) continue;
		if(name && g && strcmp(name, g)) continue;

		doc_puts(doc, "<tr style=\"font-size:smaller\"><td>");
		doc_printf(doc, "<a href=\"%s%s?EID=%d\" target=\"_blank\">%s</a>", WEB_SERVER_URL, ec->browser_url, TupleEID(person),
			img=html_make_icon_img_tag(ec, EDB_ICON_TUPLE_SMALL, "[%d]", TupleEID(person)));
		FREE(img);
		show_edbclient_button(doc, TupleEID(person));
		edb_book_print_not_implement(ep, person);
		doc_puts(doc, "</td><td>\n");
		doc_printf(doc, "<a href=\"%s%s?EID=%d\" target=\"_blank\">%s</a>", WEB_SERVER_URL, ec->browser_url, TupleEID(p),
			img=html_make_icon_img_tag(ec, EDB_ICON_TUPLE_SMALL, "[%d]", TupleEID(p)));
		FREE(img);
		show_edbclient_button(doc, TupleEID(p));
		edb_book_print_not_implement(ep, p);
		doc_puts(doc, "</td></tr>\n");
	}
}

static
void
check_print(EdbPrint ep, struct CHECKER *chk, EdbBookPrintFunc func, EdbPrint tmpep)
{
	register EdbContext	ec = ep->ec;
	Link	l;
	EdbTupleInfo	tuple;

	doc_printf(ep->doc, "<dt>");
	doc_puts(ep->doc, "上記以外で，<br />\n");
	doc_printf(ep->doc, "<b>何らかの警告がある情報</b>\n");
	ForLink(l, chk->l_info) {
		int	status;
		char	*img = NULL;

		tuple = (EdbTupleInfo)l->d;
		if(bitree_search(chk->printed, tuple, TupleEID(tuple), NULL)) continue;
		EpStatus(tmpep) = STATUS_NORMAL;
		status = (*func)(tmpep, tuple);
		if(!(EpStatus(tmpep)&STATUS_ALERT)) continue;
		doc_puts(ep->doc, "<dd style=\"font-size:90%;\">");
		doc_printf(ep->doc, "<a href=\"%s%s?EID=%d\" target=\"_blank\">%s</a>", WEB_SERVER_URL, ec->browser_url, TupleEID(tuple),
			img=html_make_icon_img_tag(ec, EDB_ICON_TUPLE_SMALL, "[%d]", TupleEID(tuple)));
		FREE(img);
		show_edbclient_button(ep->doc, TupleEID(tuple));
		(*func)(ep, tuple);
		BiTreeAdd(chk->printed, tuple, TupleEID(tuple), NULL, NULL, NULL);
		doc_puts(ep->doc, "</dd>\n");
	}
}

static
void
check_pronounce(EdbPrint ep, struct CHECKER *chk, char *column, EdbBookPrintFunc func)
{
	register Doc	doc = ep->doc;
	register EdbContext	ec = ep->ec;
	EdbTupleInfo	tuple;
	Link	l;
	EdbTupleColumn	tc;
	int	idx;
	EdbColumnInfo		ci;

	ci = edb_table_seek_column(ec, chk->ti, (utf8_t *)column);
	if(ci==NULL) {
		edb_trace(ec, "check_column: can't find column (%s).\n", column);
		return;
	}

	doc_printf(doc, "<dt>");
	doc_puts(doc, "上記以外で，<br />\n");
	doc_printf(doc, "<b>[%s]の読みが定義されていない情報</b>\n", ColumnName_ja(ci));
	doc_printf(doc, "（%s）\n", "読みが定義されていないと50音順に並べることができません．");
#if 0
	idx = edb_table_seek_column_index(ec, table_ti, (utf8_t *)column);
	for(l=l_info; l; l=l->n) {
		EdbDatum	d;

		tuple = (EdbTupleInfo)l->d;
		if(bitree_search(printed, tuple, TupleEID(tuple), NULL)) continue;
		tc = TupleColumn(tuple, idx);
		if(tc->datum==NULL) continue;
		d = tc->datum;
		if(!DatumJapaneseIsValid(d)) continue;
		if(DatumPronounceIsValid(d)) continue;
		doc_puts(doc, "<dd>");
		if(func) (*func)(ep, tuple);
		BiTreeAdd(printed, tuple, TupleEID(tuple), NULL, NULL, NULL);
		doc_puts(doc, "\n");
	}
#endif
}

static
int
check_person(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	EdbTupleInfo	tuple;
	Link	l;
	int	i, idx;
static struct CHECK_LIST check_list[] = {
{	"@.title",				"[肩書]が定義されていない情報は冊子編簒の際に分類されません",	edb_book_print_not_implement	},
{	"@.degree",			NULL,	edb_print_tuple	},
{	"@.officialnameofdegree",	NULL,	edb_print_tuple	},
{	"@.email",				NULL,	edb_print_tuple	},
{	"@.affiliation",			NULL,	edb_print_tuple	}
};

	check_begin(ep, sec, &chk);
	check_column(ep, &check_list[0], &chk);

	doc_printf(sec->ep->doc, "<dt><b>%s</b>\n", "教員以外を省く．");
	for(l=chk.l_info; l; l=l->n) {
		tuple = (EdbTupleInfo)l->d;
		if(bitree_search(chk.printed, tuple, TupleEID(tuple), NULL)) continue;
		if(edb_book_person_is_teacher(ec, TupleEID(tuple))) continue;
		BiTreeAdd(chk.printed, tuple, TupleEID(tuple), NULL, NULL, NULL);
	}

	for(i=1; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);

	doc_printf(sec->ep->doc, "<dt><b>和名が同姓同名の個人情報</b></dt>\n");
	doc_printf(sec->ep->doc, "<dd style=\"font-size:smaller\">誤った個人情報の登録を発見しやすくするためのもので，必ずしも「同姓同名＝誤り」というのものではありません．</dd>\n");
	doc_printf(sec->ep->doc, "<dd><table border=\"1\">\n");
	ForLink(l, chk.l_info) check_samename(ep, l->d, l->n);
	doc_printf(sec->ep->doc, "</table>\n");

	doc_printf(sec->ep->doc, "</dd>\n");
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_study(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
{	"@.researcher",		"[研究者]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.field",			NULL,	edb_book_print_not_implement	},
{	"@.theme",		NULL,	edb_book_print_not_implement	},
{	"@.keyword",		NULL,	edb_book_print_not_implement	},
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_academicsocietyactivity(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
{	"@.person",			"[個人]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.academicsociety",	NULL,	edb_book_print_not_implement	}
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_socialactivity(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
{	"@.person",	"[個人]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.organization",	NULL,	edb_book_print_not_implement	}
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_article(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
//	{	"@.organization",	"[組織]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.kind",			"[種別]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.author",		NULL,	edb_print_tuple	},
{	"@.title",			NULL,	edb_print_tuple	},
{	"@.date",			"[年月日]が定義されていない情報は冊子に掲載されません",	edb_print_tuple	}
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);

	check_print(ep, &chk, edb_print_tuple, book->tmpep);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_patent(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
//	{	"@.organization",	"[組織]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.kind",	"[種別]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.inventor",	NULL,	edb_print_tuple	},
{	"@.name",	NULL,	edb_print_tuple	}
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_print(ep, &chk, edb_print_tuple, book->tmpep);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_researchactivity(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
//	{	"@.organization",	"[組織]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.researcher",		NULL,											edb_print_tuple	},
{	"@.kind",			NULL,											edb_print_tuple	},
{	"@.sponsor",		NULL,											edb_print_tuple	},
{	"@.budget",		NULL,											edb_print_tuple	},
{	"@.theme",		NULL,											edb_print_tuple	},
{	"@.period",		"[年月日]が定義されていない情報は冊子に掲載されません",	edb_print_tuple	}
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);

	check_print(ep, &chk, edb_print_tuple, book->tmpep);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_conference(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
//	{	"@.organization",	"[組織]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.society",		NULL,											edb_print_tuple	},
{	"@.name",			NULL,											edb_print_tuple	},
{	"@.class1",		NULL,											edb_print_tuple	},
{	"@.class2",		NULL,											edb_print_tuple	},
{	"@.promoter",		NULL,											edb_print_tuple	},
{	"@.promoter.role",	NULL,											edb_print_tuple	},
{	"@.period",		"[年月日]が定義されていない情報は冊子に掲載されません",	edb_print_tuple	},
{	"@.location",		"[開催地]が定義されていない情報は冊子に掲載されません",	edb_print_tuple	}
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_print(ep, &chk, edb_print_tuple, book->tmpep);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_collaboration(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
//	{	"@.organization",			"[組織]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.organization.researcher",	NULL,											edb_print_tuple	},
{	"@.class",					NULL,											edb_print_tuple	},
{	"@.theme",				NULL,											edb_print_tuple	},
{	"@.period",				"[年月日]が定義されていない情報は冊子に掲載されません",	edb_print_tuple	}
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);

	check_print(ep, &chk, edb_print_tuple, book->tmpep);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_prize(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
//	{	"@.organization",	"[組織]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.awarder",		NULL,											edb_print_tuple	},
{	"@.name",			NULL,											edb_print_tuple	},
{	"@.awardee",		NULL,											edb_print_tuple	},
{	"@.date",			"[年月日]が定義されていない情報は冊子に掲載されません",	edb_print_tuple	}
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_print(ep, &chk, edb_print_tuple, book->tmpep);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_thesis(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
{	"@.course",		"[専攻]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.degree",		"[学位]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.name",		NULL,											edb_print_tuple		},
{	"@.title",		NULL,											edb_print_tuple		},
{	"@.date",		"[年月日]が定義されていない情報は冊子に掲載されません",	edb_print_tuple		},
{	"@.teacher",	NULL,											edb_print_tuple		},
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);

	check_pronounce(sec->ep, &chk, "@.name", edb_print_tuple);

	check_print(ep, &chk, edb_print_tuple, book->tmpep);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_lecture(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
{	"@.name",		NULL, 											edb_book_print_not_implement	},
{	"@.course",	"[コース]が定義されていない情報は冊子に掲載されません",	edb_book_print_not_implement	},
{	"@.lecturer",	NULL,											edb_book_print_not_implement	},
{	"@.unit",		NULL,											edb_book_print_not_implement	},
{	"@.target",	NULL,											edb_book_print_not_implement	},
{	"@.outline",	NULL,											edb_book_print_not_implement	},
{	"@.goal",		NULL,											edb_book_print_not_implement	},
{	"@.contact",	NULL,											edb_book_print_not_implement	},
{	"@.student",	NULL,											edb_book_print_not_implement	}
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_pronounce(sec->ep, &chk, "@.name", edb_print_tuple);
	check_print(ep, &chk, edb_print_tuple, book->tmpep);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_history_subject_and_kind(EdbPrint ep, struct CHECKER *chk, int (*func)())
{
	register Doc	doc = ep->doc;
	register EdbContext	ec = ep->ec;
	Link	l;

	ForLink(l, chk->l_info) {
		int	status;
		EdbTupleColumn subject, kind;
		eid_t	subject_eid, kind_eid;
		EdbTupleInfo	tuple, subject_tuple;
		char	*img = NULL;

		tuple = (EdbTupleInfo)l->d;
		if(bitree_search(chk->printed, tuple, TupleEID(tuple), NULL)) continue;

		subject = edb_tuple_seek_column(ec, tuple, (utf8_t *)"主体");
		if(subject && subject->datum) {
			subject_eid = DatumEID(subject->datum);
		} else continue;
		subject_tuple = edb_tuple_get(ec, subject_eid, 0);
		if(subject_tuple==NULL) continue;

		kind = edb_tuple_seek_column(ec, tuple, (utf8_t *)"属性");
		if(kind && kind->datum) {
			kind_eid = DatumEID(kind->datum);
		} else continue;
		if(TupleIsPerson(subject_tuple)) {
			if(kind_eid==10357 || kind_eid==10358 || kind_eid==116050) continue;	/* 学歴, 職歴, 資格 */
		} else if(TupleIsOrganization(subject_tuple)) {
			if(kind_eid==10359) continue;					/* 沿革 */
		} else continue;

		doc_puts(doc, "<dd style=\"font-size:90%;\">");
		doc_printf(doc, "<a href=\"%s%s?EID=%d\" target=\"_blank\">%s</a>", WEB_SERVER_URL, ec->browser_url, TupleEID(tuple),
			img=html_make_icon_img_tag(ec, EDB_ICON_TUPLE_SMALL, "[%d]", TupleEID(tuple)));
		FREE(img);
		show_edbclient_button(doc, TupleEID(tuple));
		(*func)(ep, tuple);
		BiTreeAdd(chk->printed, tuple, TupleEID(tuple), NULL, NULL, NULL);
		doc_puts(doc, "</dd>\n");
	}
	return TRUE;
}

static
int
check_history(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
{	"@.subject",	NULL, 					edb_book_print_not_implement	},
{	"@.date",		NULL,					edb_book_print_not_implement	},
{	"@.attribute",	NULL,					edb_book_print_not_implement	},
{	"@.event",	NULL,					edb_book_print_not_implement	},
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);

	doc_printf(sec->ep->doc, "<dt>上記以外で<br /><b>変な【履歴】情報．</b>\n");
	check_history_subject_and_kind(ep, &chk, edb_book_print_not_implement);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_posthistory(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
{	"@.class",		NULL, 					edb_book_print_not_implement	},
{	"@.post",		NULL,					edb_book_print_not_implement	},
{	"@.person",	NULL,					edb_book_print_not_implement	},
{	"@.period",	NULL,					edb_book_print_not_implement	},
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_teaching(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
{	"@.teacher",	NULL, 					edb_book_print_not_implement	},
{	"@.year",		NULL,					edb_book_print_not_implement	},
{	"@.subject",	NULL,					edb_book_print_not_implement	},
{	"@.course",	NULL,					edb_book_print_not_implement	},
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_picture(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	int	i;
static struct CHECK_LIST check_list[] = {
{	"@.name",		NULL, 					edb_book_print_not_implement	},
{	"@.class",		NULL,					edb_book_print_not_implement	},
{	"@.body",		NULL,					edb_book_print_not_implement	},
};

	check_begin(ep, sec, &chk);
	for(i=0; i<Ns(check_list); ++i) check_column(ep, &check_list[i], &chk);
	check_end(ep, sec, &chk);
	return TRUE;
}

static
int
check_default(Book book, Section sec)
{
	register EdbPrint	ep = sec->ep;
	register EdbContext	ec = ep->ec;
	struct CHECKER chk;
	EdbTableInfo	ti = edb_table_get_by_name(ec, (utf8_t *)sec->name);
	Link	l;

	check_begin(ep, sec, &chk);
	ForLink(l, ti->l_column) {
		EdbColumnInfo	ci = l->d;
		struct CHECK_LIST check_list;
		if(ci->obsoleted) continue;
		if(ci->usage!=USAGE_REQUIRED) continue;

		check_list.column = (char *)ColumnXN(ci);
		check_list.description = NULL;
		check_list.func = edb_book_print_not_implement;
		check_column(ep, &check_list, &chk);
	}

	check_print(ep, &chk, edb_print_tuple, book->tmpep);
	check_end(ep, sec, &chk);
	return TRUE;
}


