/*
	$Id: mod_article.c,v 1.11 2012/04/03 03:51:40 alex Exp alex $
	mod_article.c---
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: mod_article.c,v 1.11 2012/04/03 03:51:40 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<ctype.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/html.h"
#include	"edb/publish.h"

#define	STUDENT_FGC	"#008000"

struct ARTICLE_KIND2MODE {
	eid_t	eid;
#define	ARTICLE_KIND			0x00000001
#define	ARTICLE_AUTHOR		0x00000002
#define	ARTICLE_TITLE			0x00000004
#define	ARTICLE_SUBTITLE		0x00000008
#define	ARTICLE_MAGAZINE		0x00000010
#define	ARTICLE_VOLUME		0x00000020
#define	ARTICLE_NUMBER		0x00000040
#define	ARTICLE_PAGE			0x00000080
#define	ARTICLE_PUBLISHER	0x00000100
#define	ARTICLE_PLACE		0x00000200
#define	ARTICLE_DATE			0x00000400
	int	mode;
	int	ym;
};

static struct ARTICLE_KIND2MODE article_kind2mode[] = {{
		10442,	/* 著書 */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_PUBLISHER | ARTICLE_PLACE,
		TRUE
},{
		10443,	/* 学術論文 */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_MAGAZINE | ARTICLE_VOLUME | ARTICLE_NUMBER | ARTICLE_PAGE,
		FALSE,
},{
		60752,	/* 学術論文 */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_MAGAZINE | ARTICLE_VOLUME | ARTICLE_NUMBER | ARTICLE_PAGE,
		FALSE
},{
		10444,	/* 学術レター */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_MAGAZINE | ARTICLE_VOLUME | ARTICLE_NUMBER | ARTICLE_PAGE,
		FALSE
},{
		10445,	/* 総説・解説 */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_MAGAZINE | ARTICLE_VOLUME | ARTICLE_NUMBER | ARTICLE_PAGE,
		TRUE
},{
		10446,	/* 国際会議 */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_MAGAZINE | ARTICLE_VOLUME | ARTICLE_NUMBER | ARTICLE_PAGE
		| ARTICLE_PLACE,
		TRUE
},{
		20015,	/* 国内講演発表 */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_MAGAZINE | ARTICLE_VOLUME | ARTICLE_NUMBER | ARTICLE_PAGE,
		TRUE
},{
		60753,	/* 報告書 */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_MAGAZINE | ARTICLE_VOLUME | ARTICLE_NUMBER | ARTICLE_PAGE
		| ARTICLE_PLACE,
		TRUE
},{
		60746,	/* 作品等 */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_MAGAZINE | ARTICLE_VOLUME | ARTICLE_NUMBER | ARTICLE_PAGE,
		TRUE
},{
		10447,	/* その他 */
		ARTICLE_AUTHOR | ARTICLE_TITLE | ARTICLE_SUBTITLE | ARTICLE_DATE
		| ARTICLE_MAGAZINE | ARTICLE_VOLUME | ARTICLE_NUMBER | ARTICLE_PAGE,
		TRUE
}};

static	eid_t	language_japanese_list[] = LANGUAGE_JAPANESE_LIST ;

static
int
article_language(EdbPrint ep, EdbTupleInfo tuple)
{
	EdbTupleColumn	tc;
	EdbDatum	d;

	tc = edb_tuple_seek_column(ep->ec, tuple, (utf8_t *)"@.language");
	ForTCDatum(d, tc) {
		int	i;

		if(!DatumEIDisValid(d)) continue;
		for(i=0; i<Ns(language_japanese_list); ++i) {
			if(!EIDisVALID(language_japanese_list[i])) continue;
			if(DatumEID(d)==language_japanese_list[i]) return EP_LANG_Japanese;
		}
		return EP_LANG_English;
	}
	if(!edb_book_item_japanese(ep, tuple, "@.title")) return EP_LANG_English;
	return EP_LANG_Japanese;
}

static
int
author_is_student(EdbContext ec, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	EdbTupleColumn ttc;

	// 学籍番号
	ttc = edb_tuple_seek_TC_by_xmlname(ec, tuple, tc, datum, (utf8_t *)"@.studentcode");
	if(ttc!=NULL && ttc->datum!=NULL && DatumEnglishIsValid(ttc->datum)) return TRUE;

////	// 所属(学生)
////	ttc = edb_tuple_seek_TC_by_xmlname(ec, tuple, tc, datum, (utf8_t *)"@.affiliate");
////	if(ttc==NULL || ttc->datum==NULL || !DatumEIDisValid(ttc->datum)) return FALSE;
////	// 肩書(学生)
////	ttc = edb_tuple_seek_TC_by_xmlname(ec, tuple, tc, datum, (utf8_t *)"@.title");
////	if(ttc==NULL || ttc->datum==NULL || !DatumEIDisValid(ttc->datum)) return FALSE;
////	return TRUE;
	return FALSE;
}

static
int
article_print_author(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	int	st = 0;

	edb_print_push(ep);
	edb_print_tcd_register_func(ep, "article.author", PrintStyle(ep), NULL);
	if(!ep->ct->nocolor && !strcmp((char *)ColumnXN(tc->ci), "article.author") && author_is_student(ep->ec, tuple, tc, datum)) {
		doc_color_begin(ep->doc, STUDENT_FGC);
		st = edb_print_tcd(ep, tuple, tc, datum);
		doc_color_end(ep->doc);
	} else {
		st = edb_print_tcd(ep, tuple, tc, datum);
	}
	edb_print_pop(ep);
	return st;
}

static
int
edb_book_print_article(EdbPrint ep, EdbTupleInfo tuple)
{
	register Doc	doc = ep->doc;
	int	japanese;
	int	kind;
	int	i, mode = ~0, ym = TRUE;
	EdbTupleColumn	tc;
	int	status = EP_STATUS_NORMAL;
	char	*colon, *comma, *last, *delim="";

	edb_print_push(ep);
	edb_print_tuple_prefix(ep, tuple);
	tc = edb_tuple_seek_column(ep->ec, tuple, (utf8_t *)"@.kind");
	if(tc && TCDatumFirst(tc) && DatumEIDisValid(TCDatumFirst(tc))) {
		kind = DatumEID(TCDatumFirst(tc));
		for(i=0; i<Ns(article_kind2mode); ++i) {
			if(article_kind2mode[i].eid == kind) {
				mode = article_kind2mode[i].mode;
				ym = article_kind2mode[i].ym;
				break;
			}
		}
		if(i==Ns(article_kind2mode)) mode = ~0;
	} else {
		mode = ~0;
	}
	edb_print_tcd_register_func(ep, "article.date", PrintStyle(ep), (ym) ? edb_print_type_DATE_YM : edb_print_type_DATE_Y);

	if(LangIsAuto(ep)) EpLang(ep) = article_language(ep, tuple);
	delim = "";
	CSV(ep) {
		mode = ARTICLE_AUTHOR
			| ARTICLE_KIND
			| ARTICLE_TITLE
			| ARTICLE_SUBTITLE
			| ARTICLE_MAGAZINE
			| ARTICLE_VOLUME
			| ARTICLE_NUMBER
			| ARTICLE_PAGE
			| ARTICLE_PUBLISHER
			| ARTICLE_PLACE
			| ARTICLE_DATE ;
	}

	if(LangIsJapanese(ep)) {
		colon = " : ";
		comma = ", ";
		last = ".";
	} else {
		colon = " : ";
		comma = ", ";
		last = ".";
	}
	CSV(ep) {
	} else {
		LaTeX(ep) if(!LangIsJapanese(ep)) doc_puts(doc, "\\英文\n");
	}

	status = FALSE;
	edb_print_push(ep);
	EpLang(ep) = EP_LANG_Japanese;
	if(mode & ARTICLE_KIND) status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.kind", NULL, ", ", "", NULL, EP_ALERT);
	edb_book_print_delimiter(ep, status, colon);
	edb_print_pop(ep);

	if(!ep->ct->noauthor) {
		status = FALSE;
		edb_print_push(ep);
		HTML(ep->doc) edb_print_tcd_register_func(ep, "article.author", PrintStyle(ep), article_print_author);
		ep->ct->column_mode |= EP_COLUMN_MODE_LIST_AND;
		ep->ct->tcd_mode |= EP_TCD_MODE_BOLD;
		if(mode & ARTICLE_AUTHOR) status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.author", NULL, ", ", "", NULL, EP_ALERT);
		edb_print_pop(ep);
		edb_book_print_delimiter(ep, status, colon);
	}

	status = FALSE;
	if(mode & ARTICLE_TITLE) status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.title", NULL, ", ", "", NULL, EP_ALERT);
	edb_book_print_delimiter(ep, status, comma);

	status = FALSE;
	if(mode & ARTICLE_SUBTITLE) {
		tc = edb_tuple_seek_column(ep->ec, tuple, (utf8_t *)"@.subtitle");
		if(tc && TCDatumFirst(tc)
			&& ((!LangIsJapanese(ep) && DatumEnglishIsValid(TCDatumFirst(tc)))
				|| (LangIsJapanese(ep) && DatumJapaneseIsValid(TCDatumFirst(tc))))
			) {
			status = edb_print_tuple_named_column(ep, tuple, NULL, "--- ", " --- ", "@.subtitle", NULL, " ---, ", " ---", "", EP_WARNING);
		}
	}
	edb_book_print_delimiter(ep, status, comma);

	status = FALSE;
	if(mode & ARTICLE_MAGAZINE) {
		edb_print_push(ep);
		ep->ct->column_mode |= EP_COLUMN_MODE_LIST_AND;
		ep->ct->tcd_mode |= EP_TCD_MODE_ITALIC;
		status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.magazine", NULL, ", ", "", NULL, EP_ALERT);
		edb_print_pop(ep);
	}
	edb_book_print_delimiter(ep, status, comma);

	status = FALSE;
	if(mode & ARTICLE_VOLUME) {
		edb_print_push(ep);
		ep->ct->tcd_mode |= EP_TCD_MODE_BOLD;
		status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.volume", NULL, ", ", "", NULL, EP_WARNING);
		edb_print_pop(ep);
	}
	edb_book_print_delimiter(ep, status, comma);

	status = FALSE;
	if(mode & ARTICLE_NUMBER) {
		edb_print_push(ep);
		ep->ct->tcd_mode |= EP_TCD_MODE_ITALIC;
		status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.number", NULL, ", ", "", NULL, EP_WARNING);
		edb_print_pop(ep);
	}
	edb_book_print_delimiter(ep, status, comma);

	status = FALSE;
	if(mode & ARTICLE_PAGE) status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.page", NULL, ", ", "", NULL, EP_WARNING);
	edb_book_print_delimiter(ep, status, comma);

	status = FALSE;
	if(mode & ARTICLE_PUBLISHER) status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.publisher", NULL, ", ", "", NULL, EP_ALERT);
	edb_book_print_delimiter(ep, status, comma);

	status = FALSE;
	if(mode & ARTICLE_PLACE) status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.city", NULL, ", ", "", NULL, EP_ALERT);
	edb_book_print_delimiter(ep, status, comma);

	status = FALSE;
	if(mode & ARTICLE_DATE) status = edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.date", NULL, ", ", "", NULL, EP_ALERT);
	edb_book_print_delimiter(ep, TRUE, last);

	edb_print_tuple_postfix(ep, tuple);
	edb_print_pop(ep);

	return status;
}

static struct EDB_PRINT_MODULE_INFO module_info = {
/* xmlname */	EDB_TABLE_Article_XN,
/* language */	article_language,
/* std */	edb_book_print_article,
/* xml */	NULL,
/* csv */	NULL,
/* item */	NULL,
/* cap */	edb_book_print_article
};
DECLARE_PRINT_MODULE(article, module_info)


