/*
	$Id: mod_lecture.c,v 1.12 2011/04/18 09:45:59 alex Exp $
	mod_lecture.c---
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: mod_lecture.c,v 1.12 2011/04/18 09:45:59 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<ctype.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/html.h"
#include	"edb/publish.h"

static
int
edb_book_TC_has_data(EdbContext ec, EdbTupleColumn tc)
{
	EdbDatum	d;

	if(tc==NULL) return FALSE;
	ForTCDatum(d, tc) {
		if(DatumEIDisValid(d) || DatumEnglishIsUsable(d) || DatumJapaneseIsUsable(d)) {
			return TRUE;
		}
	}
	return FALSE;
}

static
int
edb_book_print_lecture_teachers(EdbPrint ep, EdbTupleInfo tuple)
{
	register Doc	doc = ep->doc;
	EdbTupleColumn	tc;
	int	title_eid;
	EdbDatum	d;

	tc = edb_tuple_seek_column(ep->ec, tuple, (utf8_t *)"@.lecturer");
	title_eid = 0;
	ForTCDatum(d, tc) {
		if(DatumEID(d)) {
			EdbTupleInfo	rtuple, ttuple;
			EdbTupleColumn	ttc;
			EdbDatum	datum;

			rtuple = edb_tuple_get(ep->ec, DatumEID(d), 0);
			if(TupleIsPerson(rtuple)) {
				LaTeX(ep) doc_puts(doc, "\\肩書氏名{");
				ttc = TupleColumn_person_title(ep->ec, rtuple);
				if(d->l_child && d->l_child->d && (datum=TCDatumFirst((EdbTupleColumn)d->l_child->d)) && DatumEIDisValid(datum)) {
					ttc = (EdbTupleColumn)d->l_child->d;
				}
				datum = ttc ? TCDatumFirst(ttc) : NULL;
				if(title_eid==0 || (ttc && (datum=TCDatumFirst(ttc)) && DatumEIDisValid(datum) && title_eid != DatumEID(datum))) {
					if(datum && DatumEIDisValid(datum) && (ttuple = edb_tuple_get(ep->ec, DatumEID(datum), 0))) {
						EdbTupleColumn	tttc;

						tttc = edb_tuple_seek_column(ep->ec, ttuple, (utf8_t *)"名称");
						if(tttc && TCDatumFirst(tttc) && DatumJapaneseIsValid(TCDatumFirst(tttc))) {
							doc_printf(doc, "%s", DatumJapanese(TCDatumFirst(tttc)));
						} else {
							edb_print_alert(ep, EP_ALERT, "名称");
						}
						edb_print_punctuation_puts(ep, "・");
						title_eid = DatumEID(datum);
					} else {
						edb_print_alert(ep, EP_ALERT, "肩書");
						edb_print_punctuation_puts(ep, "・");
					}
				}
				LaTeX(ep) doc_puts(doc, "}{");
				edb_book_print_person_info_name(ep, rtuple, TRUE, TRUE);
				LaTeX(ep) doc_puts(doc, "}");
				LaTeX(ep) title_eid = 0;
			} else {
				LaTeX(ep) doc_puts(doc, "\\肩書氏名{}{");
				edb_book_print_text(ep, rtuple, "名称", 0, 0, EP_ALERT);
				LaTeX(ep) doc_puts(doc, "}");
				title_eid = 0;
			}
		} else {
			edb_print_alert(ep, EP_ALERT, "@.lecturer");
		}
		HTML(doc) doc_printf(doc, "%s", (d->n)? ", " : "");
		pTeX(ep) doc_printf(doc, "%s", (d->n)? ", " : "");
	}
	return TRUE;
}

static
int
edb_book_print_lecture_tcd_contact(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum d)
{
	register Doc	doc = ep->doc;
	Link	l;

	if(DatumEID(d)) {
		EdbTupleInfo rtuple, ttuple;
		EdbTupleColumn ttc;
		EdbDatum	datum;

		rtuple = edb_tuple_get(ep->ec, DatumEID(d), 0);
		if(TupleIsPerson(rtuple)) {
			ttc = TupleColumn_person_surname(ep->ec, rtuple);
			if((datum=TCDatumFirst(ttc))!=NULL) {
				if(DatumJapaneseIsValid(datum)) {
					doc_text_puts(doc, (char *)DatumJapanese(datum));
				} else if(DatumEnglishIsValid(datum)) {
					doc_text_puts(doc, (char *)DatumEnglish(datum));
				} else {
					edb_print_alert(ep, EP_ALERT, "名前");
				}
			} else {
				edb_print_alert(ep, EP_ALERT, "名前");
			}
			doc_puts(doc, "(");
			if(edb_book_print_text(ep, rtuple, "@.room", 0, 0, EP_ALERT)) doc_puts(doc, ", ");
			if(edb_book_print_text(ep, rtuple, "@.phone", 0, 0, EP_ALERT)) doc_puts(doc, ", ");
			edb_book_print_text(ep, rtuple, "@.email", 0, 0, EP_ALERT);
			doc_puts(doc, ")");
		} else {
			edb_book_print_text(ep, rtuple, "名称", 0, 0, EP_ALERT);
		}
	} else {
		if(DatumJapaneseIsValid(d)) {
			doc_text_puts(doc, (char *)DatumJapanese(d));
		} else if(DatumEnglishIsValid(d)) {
			doc_text_puts(doc, (char *)DatumEnglish(d));
		}
	}
	return TRUE;
}

static
int
edb_book_print_lecture(EdbPrint ep, EdbTupleInfo tuple)
{
	register Doc	doc = ep->doc;
	EdbTupleColumn	tc;
	int	i;
	Link	ld;

	edb_print_push(ep);
	if(EpLang(ep)==EP_LANG_Auto) EpLang(ep) = EP_LANG_Japanese;
	edb_print_tuple_prefix(ep, tuple);

	edb_print_tcd_register_func(ep, "lecture.contact", PrintStyle(ep), edb_book_print_lecture_tcd_contact);

	HTML(doc) doc_puts(doc, "<b>");
	pTeX(ep) doc_puts(doc, "{\\bf ");
	LaTeX(ep) doc_puts(doc, "\\名称{");
	edb_book_print_text(ep, tuple, "@.name", 0, 0, EP_ALERT);
	edb_print_punctuation_puts(ep, " / ");
	LaTeX(ep) doc_puts(doc, "}{");
	edb_print_push(ep);
	EpLang(ep) = LangIsJapanese(ep) ? EP_LANG_English : EP_LANG_Japanese;
	edb_book_print_text(ep, tuple, "@.name", 0, 0, EP_ALERT);
	edb_print_pop(ep);
	HTML(doc) doc_puts(doc, "</b>");
	pTeX(ep) doc_puts(doc, "}");
	LaTeX(ep) doc_puts(doc, "}");
	doc_puts(doc, "\n");

	if(edb_tuple_seek_TC_by_xmlname(ep->ec, tuple, NULL, NULL, (utf8_t *)"@.name.topic")) {
		edb_print_punctuation_puts(ep, " --- ");
		LaTeX(ep) doc_puts(doc, "\\題目{");
		edb_book_print_text(ep, tuple, "@.name.topic", 0, 0, 0);
		edb_print_punctuation_puts(ep, " / ");
		LaTeX(ep) doc_puts(doc, "}{");
		edb_print_push(ep);
		EpLang(ep) = LangIsJapanese(ep) ? EP_LANG_English : EP_LANG_Japanese;
		edb_book_print_text(ep, tuple, "@.name.topic", 0, 0, 0);
		edb_print_pop(ep);
		LaTeX(ep) doc_puts(doc, "}");
		doc_puts(doc, "\n");
	}

#if 0

	HTML(doc) doc_puts(doc, "<div align=\"right\">\n");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	LaTeX(ep) doc_puts(doc, "\\教官{");
	edb_book_print_lecture_teachers(ep, tuple);
	edb_print_punctuation_puts(ep, "　");
	LaTeX(ep) doc_puts(doc, "}\n");

	edb_print_tuple_named_column(ep, tuple, NULL, "", "+", "@.unit", NULL, "", "", NULL, EP_ALERT);
	edb_print_punctuation_puts(ep, "単位");
	HTML(doc) doc_puts(doc, "\n</div>");
	edb_book_print_delimiter(ep, TRUE, "\n");

	HTML(doc) doc_puts(doc, "<dl>\n");

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【授業目的】");
	edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.target", NULL, "", "", NULL, EP_ALERT);
	edb_book_print_delimiter(ep, TRUE, "\n");

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【授業概要】");
	edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.outline", NULL, "", "", NULL, EP_ALERT);
	edb_book_print_delimiter(ep, TRUE, "\n");

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_tuple_named_column(ep, tuple, NULL, "【受講要件】", "", "@.requirement", NULL, ", ", "", NULL, EP_ALERT);
	edb_book_print_delimiter(ep, TRUE, "\n");

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【履修上の注意】");
	edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.notice", NULL, "", "", NULL, EP_WARNING);
	edb_print_punctuation_puts(ep, "\n");

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【到達目標】\n");
	HTML(doc) {
		doc_puts(doc, "<dd>\n");
		edb_print_tuple_named_column(ep, tuple, NULL, "<ol>\n<li>", "<li>", "@.goal", NULL, "</li>\n", "</li>\n</ol>\n", NULL, EP_ALERT);
		doc_puts(doc, "</dd>\n");
	} else LaTeX(ep) {
		edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.goal", NULL, "", "", NULL, EP_ALERT);
	} else pTeX(ep) {
		edb_print_tuple_named_column(ep, tuple, NULL, "\\itemitem{・}", "\\itemitem{・}", "@.goal", NULL, "\n", "\n", "\\itemitem{・}", EP_ALERT);
	}

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【授業計画】\n");
	HTML(doc) doc_puts(doc, "<dd>\n");
	HTML(doc) {
		edb_print_tuple_named_column(ep, tuple, NULL, "<ol>\n<li>", "<li>", "@.plan", NULL, "</li>\n", "</li>\n</ol>\n", NULL, EP_ALERT);
	} else LaTeX(ep) {
		edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.plan", NULL, "", "", NULL, EP_ALERT);
	} else pTeX(ep) {
		edb_print_tuple_named_column(ep, tuple, NULL, "\\itemitem{・}", "\\itemitem{・}", "@.plan", NULL, "\n", "\n", "\\itemitem{・}", EP_ALERT);
	}

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【成績評価】");
	edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.evaluation", NULL, "", "", NULL, EP_ALERT);
	edb_book_print_delimiter(ep, TRUE, "\n");

#if 1
	tc = edb_tuple_seek_column(ep->ec, tuple, "@.jabeeevaluation");
	if(tc && edb_book_TC_has_data(ep->ec, tc)) {
		HTML(doc) doc_puts(doc, "<dt>");
		pTeX(ep) doc_puts(doc, "\\item{} ");
		edb_print_punctuation_puts(ep, "【JABEE合格】");
		edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.jabeeevaluation", NULL, "", "", NULL, EP_ALERT);
		edb_book_print_delimiter(ep, TRUE, "\n");
	}

	tc = edb_tuple_seek_column(ep->ec, tuple, "@.jabeerelation");
	if(tc && edb_book_TC_has_data(ep->ec, tc)) {
		HTML(doc) doc_puts(doc, "<dt>");
		pTeX(ep) doc_puts(doc, "\\item{} ");
		edb_print_punctuation_puts(ep, "【学習教育目標との関連】");
		edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.jabeerelation", NULL, "", "", NULL, EP_ALERT);
		edb_book_print_delimiter(ep, TRUE, "\n");
	}
#endif

	tc = edb_tuple_seek_column(ep->ec, tuple, "@.student");
	if(tc && tc->datum) {
		HTML(doc) doc_puts(doc, "<dt>");
		pTeX(ep) doc_puts(doc, "\\item{} ");
		edb_print_punctuation_puts(ep, "【対象学生】");
		edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.student", NULL, "", "", NULL, EP_WARNING);
		edb_print_punctuation_puts(ep, "\n");
	}

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【教科書】");
	edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.textbook", NULL, ", ", "", NULL, EP_ALERT);
	edb_print_punctuation_puts(ep, "\n");

	tc = edb_tuple_seek_column(ep->ec, tuple, "@.reference");
	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【参考資料】");
	edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.reference", NULL, ", ", "", NULL, EP_WARNING);
	edb_print_punctuation_puts(ep, "\n");

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【WEB頁】");
	edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.url", NULL, ", ", "", NULL, EP_WARNING);
	edb_print_punctuation_puts(ep, "\n");

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【連絡先】");
	HTML(doc) {
		edb_print_tuple_named_column(ep, tuple, NULL, "<dd>", "<dd>", "@.contact", NULL, "</dd>", "</dd>", NULL, EP_WARNING);
	} else {
		edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.contact", NULL, ", ", "", NULL, EP_WARNING);
	}
	edb_print_punctuation_puts(ep, "\n");

	HTML(doc) doc_puts(doc, "<dt>");
	pTeX(ep) doc_puts(doc, "\\item{} ");
	edb_print_punctuation_puts(ep, "【備考】");
	edb_print_tuple_named_column(ep, tuple, NULL, "", "　", "@.note", NULL, "<br />", "", NULL, EP_WARNING);

	HTML(doc) doc_puts(doc, "</dl>");
#endif
	edb_print_tuple_postfix(ep, tuple);
	doc_puts(doc, "\n");
	edb_print_pop(ep);

	return TRUE;
}

static struct EDB_PRINT_MODULE_INFO module_info = {
/* xmlname */	EDB_TABLE_Lecture_XN,
/* language */	NULL,
/* std */	edb_book_print_lecture,
/* xml */	NULL,
/* csv */	NULL,
/* item */	NULL,
/* cap */	NULL,
};
#ifdef	THE_UNIVERSITY_OF_TOKUSHIMA
DECLARE_PRINT_MODULE(lecture, module_info)
#endif

