/*
	$Id: mod_person.c,v 1.12 2012/04/03 03:51:40 alex Exp alex $
	mod_person.c---
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: mod_person.c,v 1.12 2012/04/03 03:51:40 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<ctype.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/html.h"
#include	"edb/publish.h"

static
Link
get_eid_list(EdbContext ec, EdbTableInfo ti, char *constraint)
{
	char	*query;
	char	*order;
	int	n;
	Link	l_eid;

	order = (char *)edb_table_make_order(ec, ti, (utf8_t *)"t1");
	asprintf(&query,
		"SELECT %s FROM %s AS t1 WHERE %s %s %s %s %s",
		EDB_BASE_EID,
		ti->sqlname,
		EDB_BASE_AVAILABLE,
		constraint ? "AND" : "",
		constraint ? constraint : "",
		STRisVALID(order) ? "ORDER BY" : "", order);
	FREE(order);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	return l_eid;
}

static
EdbCatalogue
catalogue_get(EdbContext ec, EdbTableInfo table, char *constraint)
{
	register EdbCatalogue	ca;
	register Link	l, l_eid;

	l_eid = get_eid_list(ec, table, constraint);
	ca = edb_catalogue_new();
	ForLink(l, l_eid) {
		register eid_t	eid = ATOI(l->d);
		register EdbTupleInfo	tuple = NULL;

		if(!EIDisVALID(eid)) continue;
		tuple = edb_tuple_get(ec, eid, 0);
		if(tuple) edb_catalogue_append(ca, TupleEID(tuple), tuple);
	}
	link_free(l_eid);
	return ca;
}

static
Link
catalogue_to_eid(EdbCatalogue ca)
{
	register Link	l, l_result = NULL;

	ForCatalogueLink(l, ca) {
		EdbTupleInfo	tuple = CatalogueObject(l);

		LinkAppend(l_result, link_atom_new_eid2s(TupleEID(tuple)));
	}
	return l_result;
}

static
int
ars_person_profile_name(EdbPrint ep, EdbTupleInfo tuple)
{
	register Doc	doc = ep->doc;
	Link	l, l_eid;
	int	n;

	l_eid = edb_eid_is_mapto_from_table(ep->ec, TupleTable(tuple), TupleEID(tuple));

	if(LangIsJapanese(ep)) {
		edb_book_print_text(ep, tuple, "@.surname", 0, 0, EP_ALERT);
		doc_putc(doc, ' ');
		edb_book_print_text(ep, tuple, "@.givenname", 0, 0, EP_ALERT);

		n = 0;
		ForLink(l, l_eid) {
			EdbTupleInfo mtuple = edb_tuple_get(ep->ec, ATOI((char *)l->d), TEMPLATE(TupleTable(tuple)), 0);

			if(mtuple && TupleAvail(mtuple)) {
				doc_puts(doc, (n==0) ? " (" : ", ");
				edb_book_print_text(ep, mtuple, "@.surname", 0, 0, EP_ALERT);
				doc_putc(doc, ' ');
				edb_book_print_text(ep, mtuple, "@.givenname", 0, 0, EP_ALERT);
				++n;
			}
		}
		if(n>0) doc_puts(doc, ")");

		edb_print_push(ep);
		EpLang(ep) = EP_LANG_English;
		doc_puts(doc, ", ");
		edb_book_print_text(ep, tuple, "@.givenname", 0, 0, EP_ALERT);
		doc_putc(doc, ' ');
		edb_book_print_text(ep, tuple, "@.surname", 0, 0, EP_ALERT);

		n = 0;
		ForLink(l, l_eid) {
			EdbTupleInfo mtuple = edb_tuple_get(ep->ec, ATOI((char *)l->d), TEMPLATE(TupleTable(tuple)), 0);

			if(mtuple && TupleAvail(mtuple)) {
				doc_puts(doc, (n==0) ? " (" : ", ");
				edb_book_print_text(ep, mtuple, "@.givenname", 0, 0, EP_ALERT);
				doc_putc(doc, ' ');
				edb_book_print_text(ep, mtuple, "@.surname", 0, 0, EP_ALERT);
				++n;
			}
		}
		if(n>0) doc_puts(doc, ")");
		edb_print_pop(ep);
	} else {
		edb_book_print_text(ep, tuple, "@.givenname", 0, 0, EP_ALERT);
		doc_putc(doc, ' ');
		edb_book_print_text(ep, tuple, "@.surname", 0, 0, EP_ALERT);

		n = 0;
		ForLink(l, l_eid) {
			EdbTupleInfo mtuple = edb_tuple_get(ep->ec, ATOI((char *)l->d), TEMPLATE(TupleTable(tuple)), 0);

			if(mtuple && TupleAvail(mtuple)) {
				doc_puts(doc, (n==0) ? " (" : ", ");
				edb_book_print_text(ep, mtuple, "@.givenname", 0, 0, EP_ALERT);
				doc_putc(doc, ' ');
				edb_book_print_text(ep, mtuple, "@.surname", 0, 0, EP_ALERT);
				++n;
			}
		}
		if(n>0) doc_puts(doc, ")");
	}

	link_free(l_eid);
	return TRUE;
}

static
int
ars_person_profile_email(EdbPrint ep, EdbTupleInfo tuple)
{

	edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.email", NULL, ", ", "", NULL, EP_ALERT);
	return TRUE;
}

static
int
ars_person_profile_post(EdbPrint ep, EdbTupleInfo tuple)
{

	edb_print_tuple_named_column(ep, tuple, NULL, "", "", "@.title", NULL, ", ", "", NULL, EP_ALERT);
	edb_print_tuple_named_column(ep, tuple, NULL, "・", "", "@.officialnameofdegree", NULL, ", ", "", NULL, EP_ALERT);
	return TRUE;
}

static
int
ars_person_profile_asa(EdbPrint ep, EdbTupleInfo tuple)
{
	register Doc	doc = ep->doc;
	Link	l;
	int	num;
	EdbTableInfo	asa_ti;
	char	*condition;
	EdbCatalogue	ca;

	asa_ti = edb_table_get_by_name(ep->ec, (utf8_t *)EDB_TABLE_AcademicSocietyActivity_XN);
	condition = (char *)edb_condition_make(ep->ec, asa_ti, CONDITION_LOGIC_AND,
		ConditionEIDMatch("@.person", TupleEID(tuple)),
		ConditionTerminate);
	ca = catalogue_get(ep->ec, asa_ti, condition);
	FREE(condition);

	num = 0;
	ForCatalogueLink(l, ca) {
		EdbTupleInfo	rti = CatalogueObject(l);
		EdbTupleColumn	tc;

		if(edb_book_print_text(ep, rti, "@.academicsociety", 0, 0, EP_WARNING)) ++num;
		doc_printf(doc, "%s", l->n ? "・" : "");
	}
	edb_catalogue_free(ca);
	return num;
}

static
int
ars_person_profile_sa(EdbPrint ep, EdbTupleInfo tuple)
{
	register Doc	doc = ep->doc;
	Link	l;
	int	num;
	EdbTableInfo	sa_ti;
	char	*condition;
	EdbCatalogue	ca;

	sa_ti = edb_table_get_by_name(ep->ec, (utf8_t *)EDB_TABLE_SocialActivity_XN);
	condition = (char *)edb_condition_make(ep->ec, sa_ti, CONDITION_LOGIC_AND,
		ConditionEIDMatch("@.person", TupleEID(tuple)),
		ConditionTerminate);
	ca = catalogue_get(ep->ec, sa_ti, condition);
	FREE(condition);

	num = 0;
	ForCatalogueLink(l, ca) {
		EdbTupleInfo	rti = CatalogueObject(l);
		EdbTupleColumn	tc;

		if(edb_book_print_text(ep, rti, "@.organization", 0, 0, EP_WARNING)) ++num;
		doc_printf(doc, "%s", l->n ? "・" : "");
	}
	edb_catalogue_free(ca);
	return num;
}

static
int
ars_person_profile_study(EdbPrint ep, EdbTupleInfo tuple, char *column)
{
	register Doc	doc = ep->doc;
	char	*condition;
	int	num = 0;
	EdbTableInfo	study_ti;
	EdbCatalogue	ca;
	Link	l;

	study_ti = edb_table_get_by_name(ep->ec, (utf8_t *)EDB_TABLE_STUDY_XN);
	condition = (char *)edb_condition_make(ep->ec, study_ti, CONDITION_LOGIC_AND,
		ConditionEIDMatch("@.researcher", TupleEID(tuple)),
		ConditionTerminate);
	ca = catalogue_get(ep->ec, study_ti, condition);
	FREE(condition);

	num = 0;
	ForCatalogueLink(l, ca) {
		EdbTupleInfo	study_tuple = CatalogueObject(l);
		EdbTupleColumn	tc;
		EdbDatum d;
		int i;

		tc = edb_tuple_seek_column(ep->ec, study_tuple, (utf8_t *)column);
		ForTCDatum(d, tc) {
			if(num>0) doc_puts(doc, ", ");
			edb_print_tcd(ep, study_tuple, tc, d);
			++num;
		}
	}
	if(num==0) edb_print_alert(ep, EP_ALERT, column);
	edb_catalogue_free(ca);
	return num;
}

static struct PERSON_PROFILE {
	char *name_en;
	char *name_ja;
	int	(*func)();
	char	*arg;
} ars_person_profile_item[] = {
{	"Name",			"氏名",				ars_person_profile_name,	NULL	},
{	"Email",			"Email",				ars_person_profile_email,	NULL	},
{	"Title, Degree",		"職名・学位",			ars_person_profile_post,	NULL	},
{	"Field of Study",	"専門分野",			ars_person_profile_study,	"@.field"	},
{	"Academic Society",	"所属学会",			ars_person_profile_asa,	NULL	},
{	"Social Activity",	"社会活動",			ars_person_profile_sa,	NULL	},
{	"Theme of Study",	"研究テーマ",			ars_person_profile_study,	"@.theme"	},
{	"Keyword",		"キーワード",			ars_person_profile_study,	"@.keyword"},
{	"Collaboratible",		"共同研究可能テーマ",	ars_person_profile_study,	"@.collaborate"}
};


static
int
ars_print_person_profile(EdbPrint ep, EdbTupleInfo tuple)
{
	register EdbContext	ec = ep->ec;
	register Doc	doc = ep->doc;
	int	i, k;
	int	degree, num, post;

	if(!EIDisVALID(TupleEID(tuple))) return FALSE;
	edb_print_push(ep);
#if 0
	EpLang(ep) = EP_LANG_Japanese;
#else
	if(EpLang(ep)==EP_LANG_Auto) EpLang(ep) = EP_LANG_Japanese;
#endif
	ep->ct->column_mode &= ~EP_COLUMN_MODE_AT_ANYTIME;
	edb_print_tuple_prefix(ep, tuple);

	for(i=0; i<Ns(ars_person_profile_item); ++i) {
		doc_puts(doc, "<span style=\"font-weight:bold\">");
		doc_text_printf(doc, "[%s]", LangIsJapanese(ep) ? ars_person_profile_item[i].name_ja : ars_person_profile_item[i].name_en);
		doc_puts(doc, "</span> ");
		(*ars_person_profile_item[i].func)(ep, tuple, ars_person_profile_item[i].arg);
		if(i+1<Ns(ars_person_profile_item)) doc_puts(doc, ", ");
		doc_puts(doc, "\n");
	}

	edb_print_tuple_postfix(ep, tuple);
	doc_puts(ep->doc, "\n");
	edb_print_pop(ep);
	return TRUE;
}

static struct EDB_PRINT_MODULE_INFO module_info = {
/* xmlname */	EDB_TABLE_Person_XN,
/* language */	NULL,
/* std */	ars_print_person_profile,
/* xml */	NULL,
/* csv */	NULL,
/* item */	NULL,
/* cap */	NULL
};
#ifdef	THE_UNIVERSITY_OF_TOKUSHIMA
DECLARE_PRINT_MODULE(person, module_info)
#endif


