/*
	$Id: print.c,v 1.46 2011/04/18 09:45:59 alex Exp $
	print.c--- print
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: print.c,v 1.46 2011/04/18 09:45:59 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/publish.h"

#define	PLAIN(doc)	if(DocIsPLAIN(doc))
#define	HTML(doc)	if(DocIsHTML(doc))
#define	TeX(doc)		if(DocIsTeX(doc))

#define	pTeX(ep)		if(!PrintIsITEMIZE((ep)) && DocIsTeX((ep)->doc))
#define	LaTeX(ep)		if(PrintIsITEMIZE((ep)) && DocIsTeX((ep)->doc))

static
void
print_tcd_info_free(BiTree_obj_t obj)
{
	register struct EDB_PRINT_TCD_INFO	*info = obj;

	FREE(info->xmlname);
	free(info);
}

static
void
print_tuple_info_free(BiTree_obj_t obj)
{
	register struct EDB_PRINT_TUPLE_INFO	*info = obj;

	bitree_free(info->bt_print_tcd_info, print_tcd_info_free);
	free(info);
}

static
struct EDB_PRINT_CONTEXT *
edb_print_context_new()
{
	struct EDB_PRINT_CONTEXT *ct;

	ct = (struct EDB_PRINT_CONTEXT *)malloc(sizeof(struct EDB_PRINT_CONTEXT));
	if(ct==NULL) return NULL;
	bzero(ct, sizeof(struct EDB_PRINT_CONTEXT));
	return ct;
}

static
void
edb_print_context_free(struct EDB_PRINT_CONTEXT *ct)
{

	if(ct->eoi_tobe_free) bitree_free(ct->bt_eoi, free);
	if(ct->print_tuple_tobe_free) bitree_free(ct->bt_print_tuple_info, print_tuple_info_free);
	if(ct->print_type_tobe_free) bitree_free(ct->bt_print_type_info, free);
	free(ct);
}

EdbPrint
edb_print_new(EdbContext ec, Doc doc, int style)
{
	EdbPrint	ep;

	ep = (EdbPrint)malloc(sizeof(struct EDB_PRINT));
	if(ep==NULL) return NULL;
	bzero(ep, sizeof(struct EDB_PRINT));
	ep->ec = ec;
	ep->doc = doc;
	ep->ct = edb_print_context_new();
	if(ep->ct==NULL) {
		free(ep);
		return NULL;
	}
	bzero(ep->ct, sizeof(struct EDB_PRINT_CONTEXT));
	ep->ct->style = style;
	if(doc->codeset==DocCodeset_Latin) {
		ep->ct->lang = EP_LANG_English;
	} else {
		ep->ct->lang = EP_LANG_Auto;
	}
	ep->ct->eoi_tobe_free = TRUE;

#ifdef EDB_READER
#else /* !EDB_READER */
	ep->ct->column_mode |= EP_COLUMN_MODE_AT_ANYTIME;
#endif /* EDB_READER */

	edb_book_print_initialize(ep);
	edb_print_type_initialize(ep);
	return ep;
}

static
int
edb_print_duplicate_eoi(BiTree_obj_t obj, void *data)
{
	register char		*eidstr = obj;
	register EdbPrint	ep = data;

	edb_print_eoi_add(ep, ATOI(eidstr));
	return TRUE;
}

EdbPrint
edb_print_dup(EdbPrint ep)
{
	EdbPrint	new;

	new = edb_print_new(ep->ec, ep->doc, ep->ct->style);
	if(new==NULL) return NULL;
	bcopy(ep->ct, new->ct, sizeof(struct EDB_PRINT_CONTEXT));
	new->ct->style = ep->ct->style;

	new->ct->bt_eoi = NULL;
	bitree_callback(ep->ct->bt_eoi, edb_print_duplicate_eoi, new);
	new->ct->eoi_tobe_free = TRUE;

	new->ct->bt_print_tuple_info = edb_print_tuple_duplicate_info(ep->ct->bt_print_tuple_info);
	new->ct->print_tuple_tobe_free = TRUE;

	new->ct->bt_print_type_info = edb_print_type_duplicate_info(ep->ct->bt_print_type_info);
	new->ct->print_type_tobe_free = TRUE;

	return new;
}

void
edb_print_free(EdbPrint ep)
{

	if(ep->ct->prev) edb_trace(ep->ec, "edb_print_free: context stack is remained.\n");
	edb_print_context_free(ep->ct);
	FREE(ep);
}

void
edb_print_push(EdbPrint ep)
{
	struct EDB_PRINT_CONTEXT *ct;

	ct = (struct EDB_PRINT_CONTEXT *)malloc(sizeof(struct EDB_PRINT_CONTEXT));
	if(ct==NULL) {
		edb_trace(ep->ec, "edb_print_push: can't allocate memory.\n");
	} else {
		bcopy(ep->ct, ct, sizeof(struct EDB_PRINT_CONTEXT));
		ct->eoi_tobe_free = FALSE;
		ct->print_tuple_tobe_free = FALSE;
		ct->print_type_tobe_free = FALSE;
		ct->prev = ep->ct;
		ep->ct = ct;
		EpStatus(ep) = EP_STATUS_NORMAL;
	}
}

void
edb_print_pop(EdbPrint ep)
{

	if(ep->ct->prev==NULL) {
		edb_trace(ep->ec, "edb_print_pop: context stack underflow.\n");
	} else {
		struct EDB_PRINT_CONTEXT *ct = ep->ct->prev;
		int	status;

		status = EpStatus(ep);
		edb_print_context_free(ep->ct);
		ep->ct = ct;
		EpStatus(ep) |= status;
	}
}

int
edb_print_tuple_prefix_register_func(EdbPrint ep, int style, EdbPrintTupleFunc_t print_tuple)
{

	if(style<0 || EP_PrintStyle_Variations<=style) return FALSE;
	ep->ct->print_tuple_prefix[style] = print_tuple;
	return TRUE;
}

int
edb_print_tuple_prefix(EdbPrint ep, EdbTupleInfo tuple)
{
	register Doc	doc = ep->doc;
	register int	style = PrintStyle(ep);
	register EdbPrintTupleFunc_t	print;

	if(style<0 || EP_PrintStyle_Variations<=style) return FALSE;

	if((print = ep->ct->print_tuple_prefix[style])!=NULL) {
		return (*print)(ep, tuple);
	} else if(ep->ct->noeidprefix) {
		return FALSE;
	} else {
		if(PrintStyleIsCSV(ep)) {
			doc_printf(doc, "%d\t%d\t", TupleEID(tuple), TupleEOID(tuple));
		} else LaTeX(ep) {
			doc_printf(doc, "\\begin{%s:%s}%%\n", EP_LaTeX_PREFIX, TupleXN(tuple));
			doc_printf(doc, "\\EdbEID{%d}\\EdbEOID{%d}%%\n", TupleEID(tuple), TupleEOID(tuple));
		} else if(PrintStyleIsITEMIZE(ep)) {
			doc_printf(doc, "EID:\t%d\n", TupleEID(tuple));
			doc_printf(doc, "EOID:\t%d\n", TupleEOID(tuple));
		} else {
			HTML(doc) doc_printf(doc, "<!-- %s=%d %s=%d -->", TupleAvail(tuple) ? "EID" : "eid", TupleEID(tuple), TupleAvail(tuple) ? "EOID" : "eoid", TupleEOID(tuple));
			pTeX(ep) doc_printf(doc, "\\EdbEID{%d}\\EdbEOID{%d}", TupleEID(tuple), TupleEOID(tuple));
		}
	}
	return TRUE;
}

int
edb_print_tuple_postfix_register_func(EdbPrint ep, int style, EdbPrintTupleFunc_t print_tuple)
{

	if(style<0 || EP_PrintStyle_Variations<=style) return FALSE;
	ep->ct->print_tuple_postfix[style] = print_tuple;
	return TRUE;
}

int
edb_print_tuple_postfix(EdbPrint ep, EdbTupleInfo tuple)
{
	register Doc	doc = ep->doc;
	register int	style = PrintStyle(ep);
	register EdbPrintTupleFunc_t	print;

	if(style<0 || EP_PrintStyle_Variations<=style) return FALSE;

	if((print = ep->ct->print_tuple_postfix[style])!=NULL) {
		return (*print)(ep, tuple);
	}
	CSV(ep) return FALSE;
	LaTeX(ep) {
		doc_printf(doc, "\\end{%s:%s}%%\n", EP_LaTeX_PREFIX, TupleXN(tuple));
	}
#ifdef EDB_READER
#else /* EDB_READER */
	if(TupleTable(tuple)->chronological && PrintStyleIsCAPTION(ep) && DocIsHTML(ep->doc)) {
		if(EDB_DATE_MIN < TupleDateFrom(tuple) || TupleDateTo(tuple) < EDB_DATE_MAX) {
			if(!ep->ct->nocolor) doc_puts(doc, "<span style=\"color: green\">");
			doc_puts(doc, "&nbsp;(期間限定: ");
			if(EDB_DATE_MIN < TupleDateFrom(tuple)) {
				doc_printf(doc, "%d年%d月%d日", 
					DATE_Y(TupleDateFrom(tuple)),
					DATE_M(TupleDateFrom(tuple)),
					DATE_D(TupleDateFrom(tuple)));
			}
			doc_puts(doc, "〜");
			if(TupleDateTo(tuple) < EDB_DATE_MAX) {
				doc_printf(doc, "%d年%d月%d日", 
					DATE_Y(TupleDateTo(tuple)),
					DATE_M(TupleDateTo(tuple)),
					DATE_D(TupleDateTo(tuple)));
			}
			doc_puts(doc, ")");
			if(!ep->ct->nocolor) doc_puts(doc, "</span>");
		}
	}
#endif	/* EDB_READER */
	return FALSE;
}

int
edb_print_tuple(EdbPrint ep, EdbTupleInfo tuple)
{
	register EdbPrintTupleInfo	print_info;
	register EdbPrintTupleFunc_t	print = NULL;
	register int	style = PrintStyle(ep);

	if(style<0 || EP_PrintStyle_Variations<=style) return FALSE;
	print_info = bitree_search(ep->ct->bt_print_tuple_info, NULL, TABLE_EID(TupleTable(tuple)), NULL);
	if(print_info) print = print_info->func[style];
	if(print) {
		(*print)(ep, tuple);
		return TRUE;
	} else if(style==EP_PrintStyle_ITEMIZE) {
		edb_book_print_tuple(ep, tuple, NULL);
		return TRUE;
	} else {
		return FALSE;
	}
}

int
edb_print_tuple_language(EdbPrint ep, EdbTupleInfo tuple)
{
	register EdbPrintTupleInfo	print_info;
	register EdbPrintTupleLangFunc_t	lang_func = NULL;

	print_info = bitree_search(ep->ct->bt_print_tuple_info, NULL, TABLE_EID(TupleTable(tuple)), NULL);
	if(print_info) lang_func = print_info->lang;
	if(lang_func) return (*lang_func)(ep, tuple);
	return EP_LANG_Japanese;
}

int
edb_print_tuple_caption(EdbPrint ep, EdbTupleInfo tuple, char *fmt, ...)
{
	va_list	ap;
	register EdbContext	ec = ep->ec;
	register Doc		doc = ep->doc;

	if(tuple==NULL) return FALSE;
	if((ep->ct->mode&EP_TUPLE_OnlyAvail) && !TupleAvail(tuple)) return FALSE;

	edb_print_push(ep);
	PrintStyle(ep) = PrintStyle_CAPTION;

	if(!TupleAvail(tuple)) doc_color_begin(doc, "#808080");

	if(STRisVALID(ep->ct->tuple_tag_prefix)) doc_puts(doc, ep->ct->tuple_tag_prefix);
	if(ep->ct->mode&EP_TUPLE_TableName) doc_printf(doc, "【%s】(", TupleName_ja(tuple));
	ep->ct->nocolor = !TupleAvail(tuple);
	if(edb_print_tuple(ep, tuple)) {
		if(ep->ct->mode&EP_TUPLE_TableName) doc_puts(doc, ") ");
	} else {
		Caption	cap;

		edb_print_tuple_prefix(ep, tuple);
		cap = edb_tuple_make_caption(ec, tuple, CAP_long);
		if(TupleAvail(tuple)) HTML(doc) doc_puts(doc, "<b>");
		doc_text_puts(doc, (char *)cap->c_main);
		if(TupleAvail(tuple)) HTML(doc) doc_puts(doc, "</b>");
		if(ep->ct->mode&EP_TUPLE_TableName) doc_puts(doc, ") ");
		if(ep->ct->mode&EP_TUPLE_LongFormat && STRisVALID(cap->c_sub)) doc_text_puts3(doc, " (", (char *)cap->c_sub, ")");
		edb_caption_free(ec, cap);
	}
	if(fmt) {
		char	*buf;

		va_start(ap, fmt);
		vasprintf(&buf, fmt, ap);
		va_end(ap);
		doc_puts(doc, buf);
		FREE(buf);
	}
#ifdef EDB_READER
#else /* !EDB_READER */
	if(ec->operator!=0 && (ep->ct->mode&EP_TUPLE_PersonInfo) && TupleIsPerson(tuple)) {
		if(edb_auth_registered(ec, TupleEID(tuple))) {
			doc_puts(doc, ".&lt;user");
			if(edb_person_is_personnel(ec, tuple)) doc_puts(doc, "+personnel");
			if(edb_person_is_teacher(ec, tuple)) doc_puts(doc, "+teacher");
			doc_puts(doc, "&gt;");
		}
	}
#endif /* EDB_READER */
	if(ep->ct->mode&EP_TUPLE_RefColumn && EIDisVALID(ep->ct->tuple_ref_eoi)) {
		int	printed = FALSE;
		Link	lc;

		ForLink(lc, tuple->l_column) {
			EdbTupleColumn	tc = (EdbTupleColumn)lc->d;
			EdbDatum d;

			ForTCDatum(d, tc) {
				if(DatumEID(d) == ep->ct->tuple_ref_eoi) {
					if(!printed) {
						doc_text_printf(doc, ".[%s]", ColumnName_ja(tc->ci));
						printed = TRUE;
					} else {
						doc_text_printf(doc, "+[%s]", ColumnName_ja(tc->ci));
					}
				}
			}
		}
	}
	if(ep->ct->mode&(EP_TUPLE_Operator|EP_TUPLE_ModifiedTime)) {
		int	created = !edb_tuple_has_history(ec, tuple);

		HTML(doc) doc_puts(doc, "<span style=\"font-size:80%;\">");
		doc_puts(doc, "---(");
		if(ep->ct->mode&EP_TUPLE_Operator) {
			Caption	cap;

			edb_caption_omit_save(ec);
			edb_caption_omit_append(ec, ORGAN_UNIV_EID);
			cap = edb_make_caption(ec, TupleOperator(tuple), CAP_long);
			edb_caption_omit_restore(ec);
			doc_puts(doc, created ? "Created " : "Modified ");
			doc_puts(doc, "by ");
			doc_text_puts(doc, (char *)cap->c_main);
#if 0
			if(STRisVALID(cap->c_sub)) {
				doc_text_puts(doc, "@");
				doc_text_puts(doc, cap->c_sub);
			}
#endif
			edb_caption_free(ec, cap);
		}
		if((ep->ct->mode&EP_TUPLE_Operator)&& (ep->ct->mode&EP_TUPLE_ModifiedTime)) doc_puts(doc, " ");
		if(ep->ct->mode&EP_TUPLE_ModifiedTime) {
			doc_puts(doc, "at ");
			doc_put_datetime(doc, TupleMtime(tuple));
		}
		doc_puts(doc, ")");
		HTML(doc) doc_puts(doc, "</span>");
	}
	if(!TupleAvail(tuple)) doc_color_end(doc);
	doc_puts(doc, "\n");
	edb_print_pop(ep);
	return TRUE;
}

EdbColumnPrint
edb_column_print_new(utf8_t *xmlname)
{
	EdbColumnPrint	cp;

	cp = (EdbColumnPrint)malloc(sizeof(struct EDB_COLUMN_PRINT));
	if(cp==NULL) return NULL;
	bzero(cp, sizeof(struct EDB_COLUMN_PRINT));
	cp->xmlname = (utf8_t *)strdup((char *)xmlname);
	return cp;
}

void
edb_column_print_free(EdbColumnPrint cp)
{

	FREE(cp->xmlname);
	FREE(cp);
}

/* eid of interest */
void
edb_print_eoi_clear(EdbPrint ep)
{

	if(ep->ct->eoi_tobe_free) bitree_free(ep->ct->bt_eoi, free);
	ep->ct->bt_eoi = NULL;
	ep->ct->eoi_tobe_free = FALSE;
}

int
edb_print_eoi_check(EdbPrint ep, eid_t eid)
{
	char	str[BUFSIZ];

	if(ep->ct->bt_eoi==NULL) return TRUE;
	sprintf(str, "%d", eid);
	return (bitree_search(ep->ct->bt_eoi, str, eid, NULL)!=NULL);
}

eid_t
edb_print_eoi_add(EdbPrint ep, eid_t eid)
{
	char	str[BUFSIZ];

	sprintf(str, "%d", eid);
	if(!bitree_search(ep->ct->bt_eoi, str, eid, NULL)) {
		if(!ep->ct->eoi_tobe_free) {
			BiTree	bt;

			bt = ep->ct->bt_eoi;
			ep->ct->bt_eoi = NULL;
			bitree_callback(bt, edb_print_duplicate_eoi, ep);
		}
		BiTreeAdd(ep->ct->bt_eoi, strdup(str), eid, NULL, NULL, NULL);
		ep->ct->eoi_tobe_free = TRUE;
	}
	return eid;
}

int
edb_print_alert(EdbPrint ep, int alert_level, char *fmt, ...)
{
	va_list	ap;
	register Doc	doc = ep->doc;
	char		*s;
	int		alert;

	if(ep->ct->noalert) return FALSE;
	alert = (alert_level>0) ? alert_level : ep->ct->alert_level;

	va_start(ap, fmt);
	vasprintf(&s, fmt, ap);
	va_end(ap);
	HTML(doc) {
		if(!ep->ct->nocolor) doc_color_begin(ep->doc, (alert==EP_ALERT) ? EP_ALERT_COLOR : EP_WARNING_COLOR);
		doc_text_puts3(doc, "(", s, ")");
		if(!ep->ct->nocolor) doc_color_end(ep->doc);
		EpStatus(ep) |= (alert==EP_ALERT) ? EP_STATUS_ALERT : EP_STATUS_WARNING;
		return TRUE;
	}
	TeX(doc) {
		if(alert==EP_ALERT) doc_text_puts3(doc, "\\EdbAlert{", s, "}");
		EpStatus(ep) |= (alert==EP_ALERT) ? EP_STATUS_ALERT : EP_STATUS_WARNING;
		return TRUE;
	}
	FREE(s);
	return FALSE;
}

int
edb_print_alert_column(EdbPrint ep, int alert_level, EdbColumnInfo ci)
{

	return edb_print_alert(ep, alert_level, "%s", ColumnName_ja(ci));
}

int
edb_print_punctuation_puts(EdbPrint ep, char *s)
{

	LaTeX(ep) {
	} else {
		doc_puts(ep->doc, s);
	}
	return 0;
}


