/*
	$Id: print_tuple.c,v 1.44 2012/04/03 03:51:40 alex Exp alex $
	print-tuple.c--- print tuple information
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: print_tuple.c,v 1.44 2012/04/03 03:51:40 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	<sys/param.h>
#if defined(__DARWIN__)
#else
#include	<fetch.h>
#endif
#include	<ctype.h>
#include	"edb/edb.h"
#include	"edb/tex.h"
#include	"edb/publish.h"
#include	"edb/pki.h"

#define	TUIR_COOPERATION
#if defined(TUIR_COOPERATION)
#include	<sys/types.h>
#include	<sha512.h>
#endif	/* TUIR_COOPERATION */

#if defined(__DARWIN__)

int	edb_print_issn_information(EdbContext ec, Doc doc, char *issn) { return FALSE; }
int	edb_print_issn_information_body(EdbContext ec, Doc doc, char *issn) { return FALSE; }
int	edb_print_magazine_information(EdbContext ec, Doc doc, eid_t eid, char *omitIssn, int horizontal) { return FALSE; }

#else

static
struct REPGET {
	char	*url;
	struct timeval	acctime;
} repgets[] = {{
	"http://utlf1.db.tokushima-u.ac.jp/cgi-bin/rep-get?", { 0L, 0L }
},{
	"http://utlf2.db.tokushima-u.ac.jp/cgi-bin/rep-get?", { 0L, 0L }
}};
typedef struct	REPGET	*RepGet;

struct REPGET_CTXT {
	char	*url;
	int	status;
	int	cerr;		// communication error
	FILE	*fp;
};

static
void	*
RepGetRetriever(void *arg)
{
	struct REPGET_CTXT *ctxt = arg;

	ctxt->fp = fetchGetURL(ctxt->url, "");
	if(ctxt->fp==NULL) {
		switch(fetchLastErrCode) {
		case FETCH_OK:
		case FETCH_UNAVAIL:
			break;
		default:
			ctxt->cerr = fetchLastErrCode;
			break;
		}
	}
	ctxt->status = EDB_PROCESS_TERMINATED;
	return ctxt;
}

static
char *
cmd_REPGET(EdbContext ec, char *idarg)
{
	int			status = FALSE;
	int			n, idx, num;
	RepGet		repget;
	FILE			*ffp = NULL;

	if(Ns(repgets)==0) return NULL;
	if(!STRisValid(idarg)) return NULL;

	num = Ns(repgets);
	repget = &repgets[idx=(random()%num)];
	for(n=0; n<num; ++n) {
		RepGet		rg = &repgets[n];
		if(timercmp(&repget->acctime, &rg->acctime, >)) {
			repget = rg;
			idx = n;
		}
	}

	for(n=0; n<num; ++n) {
		struct REPGET_CTXT	*ctxt = malloc(sizeof(struct REPGET_CTXT));
		bzero(ctxt, sizeof(struct REPGET_CTXT));
		pthread_t	thread;
		struct timeval	tv_start, tv_present, tv, tv_limit;

		repget = &repgets[(idx+n)%num];
		asprintf(&ctxt->url, "%s%s", repget->url, idarg);

		gettimeofday(&tv_start, NULL);
		bzero(&tv_present, sizeof(struct timeval));
		tv_present.tv_sec = 5;	// 5sec
		timeradd(&tv_start, &tv_present, &tv_limit);

		ctxt->status = EDB_PROCESS_EXECUTING;
		if(pthread_create(&thread, &ec->process_attr, RepGetRetriever, ctxt)!=0) return NULL;
		pthread_detach(thread);
		while(ctxt->status!=EDB_PROCESS_TERMINATED) {
			gettimeofday(&tv_present, NULL);
			if(timercmp(&tv_limit, &tv_present, <)) break;
			pthread_yield();
		}

		if(ctxt->status!=EDB_PROCESS_TERMINATED || ctxt->cerr) {
			// Not connected.
			if(ctxt->status==EDB_PROCESS_TERMINATED) {
				fprintf(stderr, "FetchErrCode: %d", ctxt->cerr);
			}
			bzero(&tv, sizeof(struct timeval));
			tv.tv_sec = 60*60; // 1 hour (penalty)
			timeradd(&tv, &repget->acctime, &repget->acctime);
		} else {
			gettimeofday(&tv_present, NULL);
			timersub(&tv_present, &tv_start, &tv);
			timeradd(&tv, &repget->acctime, &repget->acctime);
			ffp = ctxt->fp;
			FREE(ctxt->url);
			FREE(ctxt);
			break;
		}
	}
	if(ffp) {
		Link	l_str = NULL;
		while(!feof(ffp)) {
			char	*s;

			if((s=file_getline(ffp))==NULL) break;
			LinkAppend(l_str, link_atom_new(s, free));
		}
		fclose(ffp);

		char	*s = str_concatenate(l_str, '\n');
		EdbLinkFree(l_str);
		return s;
	}
	return NULL;
}

#define	NodeNsURI(n)	(((n)&&((n)->ns)&&((n)->ns->href)) ? (n)->ns->href : (xmlChar *)"")
#define	NodeName(n)	(((n)&&((n)->name)) ? (n)->name : (xmlChar *)"")
#define	UTLF_XMLNS_URI	"http://utlf.db.tokushima-u.ac.jp/UTLF/utlf#"

static
xmlNodePtr
utlf_get_node1(xmlDocPtr xdoc, xmlNodePtr cur, xmlChar *uri, xmlChar *en)
{

	for( ; cur!=NULL; cur=cur->next) {
		if(!xmlStrcmp(NodeNsURI(cur), uri) && !xmlStrcmp(cur->name, en)) return cur;
	}
	return NULL;
}

static
xmlChar *
utlf_get_dict_text(xmlDocPtr xdoc, xmlChar *dv, xmlNodePtr dict, ...)
{
	va_list	ap;
	xmlChar	*key;
	xmlNodePtr	cur;

	va_start(ap, dict);
	key = va_arg(ap, xmlChar *);

	cur = dict->xmlChildrenNode;
	while(key && cur) {
		if(strcmp((char *)NodeNsURI(cur), UTLF_XMLNS_URI) || strcmp((char *)NodeName(cur), "key")) {
			cur = cur->next;
			continue;
		}
		xmlChar *keyname = xmlNodeListGetString(xdoc, cur->xmlChildrenNode, 1);
		if(!keyname) keyname = (xmlChar *)"";
		if(!keyname || strcmp((char *)key, (char *)keyname)) {
			cur = cur->next;
			continue;
		}

		key = va_arg(ap, xmlChar *);
		if(key==NULL) break;
		if(cur->next==NULL) break;
		cur = cur->next->xmlChildrenNode;
	}
	va_end(ap);

	if(key==NULL && cur && cur->next) {
		return  xmlNodeListGetString(xdoc, cur->next->xmlChildrenNode, 1);
	}
	return dv;
}

static
Link
utlf_get_dict_text_list(xmlDocPtr xdoc, xmlNodePtr dict, ...)
{
	va_list	ap;
	xmlChar	*key;
	xmlNodePtr	cur;

	va_start(ap, dict);
	key = va_arg(ap, xmlChar *);

	cur = dict->xmlChildrenNode;
	while(key && cur) {
		if(strcmp((char *)NodeNsURI(cur), UTLF_XMLNS_URI) || strcmp((char *)NodeName(cur), "key")) {
			cur = cur->next;
			continue;
		}
		xmlChar *keyname = xmlNodeListGetString(xdoc, cur->xmlChildrenNode, 1);
		if(!keyname) keyname = (xmlChar *)"";
		if(!keyname || strcmp((char *)key, (char *)keyname)) {
			cur = cur->next;
			continue;
		}

		key = va_arg(ap, xmlChar *);
		if(key==NULL) break;
		if(cur->next==NULL) break;
		cur = cur->next->xmlChildrenNode;
	}
	va_end(ap);

	Link	list = NULL;
	if(key==NULL && cur && cur->next) {
		cur = cur->next;
		if(!strcmp((char *)NodeNsURI(cur), UTLF_XMLNS_URI) && !strcmp((char *)NodeName(cur), "array")) {
			for(cur = cur->xmlChildrenNode;  cur ; cur = cur->next) {
				xmlChar *s = xmlNodeListGetString(xdoc, cur->xmlChildrenNode, 1);
				if(s) LinkAppend(list, link_atom_new(strdup((char *)s), free));
			}
		} else {
			xmlChar *s = xmlNodeListGetString(xdoc, cur->xmlChildrenNode, 1);
			if(s) LinkAppend(list, link_atom_new(strdup((char *)s), free));
		}
	}
	return list;
}

static
Link
utlf_get_dict_list(xmlDocPtr xdoc, xmlNodePtr dict, ...)
{
	va_list	ap;
	xmlChar	*key;
	xmlNodePtr	cur;

	va_start(ap, dict);
	key = va_arg(ap, xmlChar *);

	cur = dict->xmlChildrenNode;
	while(key && cur) {
		if(strcmp((char *)NodeNsURI(cur), UTLF_XMLNS_URI) || strcmp((char *)NodeName(cur), "key")) {
			cur = cur->next;
			continue;
		}
		xmlChar *keyname = xmlNodeListGetString(xdoc, cur->xmlChildrenNode, 1);
		if(!keyname) keyname = (xmlChar *)"";
		if(!keyname || strcmp((char *)key, (char *)keyname)) {
			cur = cur->next;
			continue;
		}

		key = va_arg(ap, xmlChar *);
		if(key==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(key==NULL && cur && cur->next) {
		cur = cur->next;
		Link	list = NULL;
		if(!strcmp((char *)NodeNsURI(cur), UTLF_XMLNS_URI) && !strcmp((char *)NodeName(cur), "array")) {
			for(cur=cur->xmlChildrenNode; cur; cur=cur->next) {
				LinkAppend(list, link_atom_new(cur, NULL));
			}
		} else {
			LinkAppend(list, link_atom_new(cur, NULL));
		}
		return list;
	}
	return NULL;
}

static
xmlNodePtr
utlf_get_content_dict(xmlDocPtr xdoc)
{

	xmlNodePtr	root = xmlDocGetRootElement(xdoc);
	if(root==NULL) return NULL;
	xmlNodePtr	node = utlf_get_node1(xdoc, root, (xmlChar *)"http://www.w3.org/1999/02/22-rdf-syntax-ns#", (xmlChar *)"RDF");
	if(node==NULL) return NULL;
	xmlNodePtr	desc = utlf_get_node1(xdoc, node->xmlChildrenNode, (xmlChar *)"http://www.w3.org/1999/02/22-rdf-syntax-ns#", (xmlChar *)"Description");
	if(desc==NULL) return NULL;
	xmlNodePtr	contentDict = utlf_get_node1(xdoc, desc->xmlChildrenNode, (xmlChar *)UTLF_XMLNS_URI, (xmlChar *)"content");
	if(desc==NULL) return NULL;
	return contentDict;
}

static
int
checkISSN(char *issn)
{

	if(!STRisValid(issn)) return FALSE;
	if(strlen(issn)!=9) return FALSE;
	if(!isdigit(issn[0])) return FALSE;
	if(!isdigit(issn[1])) return FALSE;
	if(!isdigit(issn[2])) return FALSE;
	if(!isdigit(issn[3])) return FALSE;
	if(issn[4]!='-') return FALSE;
	if(!isdigit(issn[5])) return FALSE;
	if(!isdigit(issn[6])) return FALSE;
	if(!isdigit(issn[7])) return FALSE;
	if(!(isdigit(issn[8]) || issn[8]=='X')) return FALSE;

	// check check-digit
	int sum = 0;
	sum += (int)(issn[0]-'0')*8;
	sum += (int)(issn[1]-'0')*7;
	sum += (int)(issn[2]-'0')*6;
	sum += (int)(issn[3]-'0')*5;
	sum += (int)(issn[5]-'0')*4;
	sum += (int)(issn[6]-'0')*3;
	sum += (int)(issn[7]-'0')*2;

	int	check1 = (issn[8]=='X') ? 10 : (issn[8]-'0');
	int	check2 = 11-(sum%11);
	if(check2==11) check2 = 0;
	return (check1==check2);
}

static
int
checkORCID(char *orcid)
{

	if(!STRisValid(orcid)) return FALSE;
	if(strlen(orcid)!=19) return FALSE;
	for(int i=0; i<18; ++i) {
		if(i==4 || i==9 || i==14) {
			if(orcid[i]!='-') return FALSE;
		} else {
			if(!isdigit(orcid[i])) return FALSE;
		}
	}
	if(!(isdigit(orcid[18]) || orcid[18]=='X')) return FALSE;

	// check check-digit
	int sum = 0;
	for(int i=0; i<18; ++i) {
		if(i==4 || i==9 || i==14) continue;
		sum = (int)(sum+(orcid[i]-'0'))*2;
	}
	int	check1 = (orcid[18]=='X') ? 10 : (orcid[18]-'0');
	int	check2 = (12-(sum%11))%11;
	return (check1==check2);
}

static struct ExtraDBViewer {
	char	*title;
	char *sysID;
	char	*prefix;
	char	*url;
} extraDBViewerTemplates[] = {
{
	"NLM Catalog",
		"NLM.nih.gov",
		"NlmId=",
//		"http://ncbi.nlm.nih.gov/sites/entrez?db=nlmcatalog&term=%%22%s%%22%%5BNLM%%20Unique%%20ID%%5D",
		"https://www.ncbi.nlm.nih.gov/nlmcatalog/%s",
},{
	"Webcat Plus",
		"CiNii.nii.ac.jp",
		"NCID=",
		"http://webcatplus.nii.ac.jp/webcatplus/details/book/ncid/NCID:%s.html",
},{
	"医中誌Web",
		"JAMAS.or.jp",
		"収載誌コード=",
		"http://www.jamas.or.jp/user/database/Search/detail/scode/%s",
},{
	"J-STAGE",
		"JSTAGE.jst.go.jp",
		"資料コード=",
		"https://www.jstage.jst.go.jp/browse/%s",
},{
	"IEEE",
		"IEEE.org",
		"punumber=",
		"http://ieeexplore.ieee.org/xpl/RecentIssue.jsp?punumber=%s",
},{
	"CrossRef",
		"CrossRef.org",
		"ISSN=",
		"http://www.crossref.org/titleList/?datatype=all&search=issn&journal=%s",
},{
	"Wiley",
		"Wiley.com",
		"DOI=",
		"http://onlinelibrary.wiley.com/journal/%s",
},{
	"Scopus",
		"Scopus.com",
		"Source=",
		"http://www.scopus.com/source/sourceInfo.url?sourceId=%s&origin=inward",
}};

int
edb_print_issn_information(EdbContext ec, Doc doc, char *issn)
{

#if 0
	doc_printf(doc, "<iframe src=\"/cgi-bin/edb_browse?ACT=ISSN&D_ARG1=%s\" width=\"640\" height=\"256\" frameborder=\"0\" marginwidth=\"0\" marginheight=\"0\" style=\"background-color:transparent;\">\n", issn);
	doc_printf(doc, "</iframe>\n");
	return TRUE;
#else
	return edb_print_issn_information_body(ec, doc, issn);
#endif
}

int
edb_print_issn_information_body(EdbContext ec, Doc doc, char *issn)
{
	char	*id, *content;

	if(!STRisValid(issn)) {
		return FALSE;
	} else if(!checkISSN(issn)) {
		doc_text_puts3(doc, "<span style=\"color:#808080;\">", "(not ISSN)", "</span><br />\n");
		return FALSE;
	}
	if(EIDisValid(ec->operator)) {
#if defined(WOS_ENGAGED)
		asprintf(&id, "ID=utlf:ISSN=%s/ISSN.org&DEPTH=1", issn);
#else
		asprintf(&id, "ID=utlf:ISSN=%s/ISSN.org&DEPTH=1", issn);
#endif
	} else {
		asprintf(&id, "ID=utlf:ISSN=%s/ISSN.org", issn);
	}
	content = cmd_REPGET(ec, id);
	FREE(id);
	if(content==NULL) return FALSE;

	xmlParserCtxtPtr ctxt= xmlNewParserCtxt();
	xmlInitParserCtxt(ctxt);
	xmlCtxtUseOptions(ctxt, XML_PARSE_NOBLANKS);

	xmlDocPtr xdoc = xmlCtxtReadMemory(ctxt, (char *)content, xmlStrlen((xmlChar *)content), NULL, NULL, 0);
	if(xdoc==NULL) {
		fprintf(stderr, "edb_test: unable to parse.\n");
		if(ctxt) xmlFreeParserCtxt(ctxt);
		FREE(content);
		return FALSE;
	}

	xmlNodePtr contentDict = utlf_get_content_dict(xdoc);
	if(contentDict==NULL) {
		if(xdoc) xmlFreeDoc(xdoc);
		if(ctxt) xmlFreeParserCtxt(ctxt);
		FREE(content);
		return FALSE;
	}

//	doc_text_puts3(doc, "<span style=\"color:red;\">", "(Now, under construction)", "</span><br />\n");
	doc_text_puts3(doc, "ISSN: <b>", (char *)utlf_get_dict_text(xdoc, (utf8_t *)"", contentDict, "ISSN", NULL), "</b>\n");

	utf8_t *pissn = utlf_get_dict_text(xdoc, (utf8_t *)"", contentDict, "pISSN", NULL);
	utf8_t *eissn = utlf_get_dict_text(xdoc, (utf8_t *)"", contentDict, "eISSN", NULL);
	if(STRisValid(pissn) || STRisValid(eissn)) {
		doc_puts(doc, "(");
		if(STRisValid(pissn)) doc_text_puts3(doc, "pISSN: <b>", (char *)pissn, "</b>");
		if(STRisValid(pissn) && STRisValid(eissn)) doc_puts(doc, ", ");
		if(STRisValid(eissn)) doc_text_puts3(doc, "eISSN: <b>", (char *)eissn, "</b>");
		doc_puts(doc, ")<br />");
	}
	utf8_t *t = utlf_get_dict_text(xdoc, (utf8_t *)"", contentDict, "Title", NULL);
	if(STRisValid(t)) doc_text_puts3(doc, "Title: <b>", (char *)t, "</b><br />");
	t = utlf_get_dict_text(xdoc, (utf8_t *)"", contentDict, "IsoAbbr", NULL);
	if(STRisValid(t)) doc_text_puts3(doc, "Title(ISO): <b>", (char *)t, "</b><br />");
	t = utlf_get_dict_text(xdoc, (utf8_t *)"", contentDict, "Supplier", NULL);
	if(STRisValid(t)) doc_text_puts3(doc, "Supplier: <b>", (char *)t, "</b><br />");
	t = utlf_get_dict_text(xdoc, (utf8_t *)"", contentDict, "Publisher", NULL);
	if(STRisValid(t)) doc_text_puts3(doc, "Publisher: <b>", (char *)t, "</b><br />");

	{
		Link	le, l_extraId = utlf_get_dict_text_list(xdoc, contentDict, "ExtraID", NULL);
		ForLink(le, l_extraId) {
			xmlChar	*id = le->d;
			if(!STRisValid(id)) continue;
			int	i;
			for(i=0; i<Ns(extraDBViewerTemplates); ++i) {
				struct ExtraDBViewer	*ev = &extraDBViewerTemplates[i];
				char *tmp = STRDUP((char *)id);
				if(strncmp(tmp, "utlf:", 5)) continue;
				tmp = &tmp[5];
				if(strncmp(tmp, ev->prefix, strlen(ev->prefix))) continue;
				tmp = &tmp[strlen(ev->prefix)];
				char	*s = rindex(tmp, '/');
				if(!s) continue;
				if(strcmp(&s[1], ev->sysID)) continue;
				*s = '\0';
				doc_puts(doc, "<span style=\"font-size:80%;\">&nbsp;(<a href=\"");
				doc_printf(doc, ev->url, tmp);
				doc_puts(doc, "\" target=\"_blank\">");
				doc_text_puts(doc, ev->title);
				doc_puts(doc, "</a>)&nbsp;</span>\n");
			}
		}
		link_free(l_extraId);
	}

#if defined(WOS_ENGAGED)
#define	JCR_YEAR_begin	2004
#define	JCR_YEAR_end		2014

	Link	lj, l_jcr = utlf_get_dict_list(xdoc, contentDict, "JCR", NULL);
	if(l_jcr==NULL || link_atoms(l_jcr)==0) {
		doc_text_puts3(doc, "", "(No JCR information.)", "<br />");
	} else if(!EIDisValid(ec->operator)) {	// EDB にログインしないユーザにはJCR情報は見せない．
		doc_text_puts3(doc, "", "(JCR information is found. [need login])", "<br />");
	} else {
		ForLink(lj, l_jcr) {
			int	y;
			xmlNodePtr	jcrDict = lj->d;
			if(jcrDict==NULL) continue;

			char *cellStyle = " style=\"border-width:1px; border-style:solid; border-color:#808080; background-color:white;\"";
			char *naStyle = " style=\"border-width:1px; border-style:solid; border-color:#808080;\"";
			doc_puts(doc, "<table style=\"border-width:1px; border-color:black;border-style:solid;\">\n");

			doc_printf(doc, "<tr><td colspan=\"%d\" style=\"background-color:" EXTDB_WOS_BGC ";\">", JCR_YEAR_end-JCR_YEAR_begin+2);
			doc_text_puts3(doc, "Title (JCR): <b>", (char *)utlf_get_dict_text(xdoc, (utf8_t *)"(not defined)", jcrDict, "Title", NULL), "</b>\n");
			doc_puts(doc, "</td></tr>\n");

			doc_puts(doc, "<tr>\n");
			doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\"></th>");
			for(y=JCR_YEAR_begin; y<=JCR_YEAR_end; ++y) doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\">%d</th>", y);
			doc_puts(doc, "</tr>\n");

			doc_puts(doc, "<tr>\n");
			doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\">IF</th>");
			for(y=JCR_YEAR_begin; y<=JCR_YEAR_end; ++y) {
				char	*ystr;
				asprintf(&ystr, "%d", y);
				char *jif = (char *)utlf_get_dict_text(xdoc, NULL, jcrDict, ystr, "ImpactFactor", NULL);
				if(jif!=NULL) {
					doc_printf(doc, "<td%s align=\"right\">", cellStyle);
					doc_text_puts3(doc, "", jif, "</td>");
				} else {
					doc_printf(doc, "<td%s align=\"center\">n/a</td>", naStyle);
				}
				FREE(ystr);
			}
			doc_puts(doc, "</tr>\n");

			doc_puts(doc, "<tr>\n");
			doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\">5Y-IF</th>");
			for(y=JCR_YEAR_begin; y<=JCR_YEAR_end; ++y) {
				char	*ystr;
				asprintf(&ystr, "%d", y);
				char *jif = (char *)utlf_get_dict_text(xdoc, NULL, jcrDict, ystr, "5YearImpactFactor", NULL);
				if(jif!=NULL) {
					doc_printf(doc, "<td%s align=\"right\">", cellStyle);
					doc_text_puts3(doc, "", jif, "</td>");
				} else {
					doc_printf(doc, "<td%s align=\"center\">n/a</td>", naStyle);
				}
				FREE(ystr);
			}
			doc_puts(doc, "</tr>\n");


			Link	lc, l_categories = utlf_get_dict_list(xdoc, jcrDict, "Category", NULL);
			ForLink(lc, l_categories) {
				xmlNodePtr catDict = lc->d;

				doc_printf(doc, "<tr><td colspan=\"%d\" style=\"background-color:" EXTDB_WOS_BGC ";\">", JCR_YEAR_end-JCR_YEAR_begin+2);
				doc_text_puts3(doc, "SubjectCategory (JCR): <b>", (char *)utlf_get_dict_text(xdoc, (utf8_t *)"n/a", catDict, "Category", NULL), "</b><br />\n");
				doc_puts(doc, "</td></tr>\n");

				cellStyle = " style=\"border-width:1px; border-style:solid; border-color:#808080;background-color:#f0f0ff\"";

				doc_puts(doc, "<tr>\n");
				doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\">A-IF</th>");
				for(y=JCR_YEAR_begin; y<=JCR_YEAR_end; ++y) {
					char	*ystr;
					asprintf(&ystr, "%d", y);
					char *aif = (char *)utlf_get_dict_text(xdoc, NULL, catDict, ystr, "AggregateImpactFactor", NULL);
					if(aif!=NULL) {
						doc_printf(doc, "<td%s align=\"right\">", cellStyle);
						doc_text_puts3(doc, "", aif, "</td>");
					} else {
						doc_printf(doc, "<td%s align=\"center\">n/a</td>", naStyle);
					}
					FREE(ystr);
				}
				doc_puts(doc, "</tr>\n");
			}
			doc_puts(doc, "</table>\n");
		}
		doc_puts(doc, "<div style=\"line-height:1;\">");
		doc_printf(doc, "<a href=\"%s\" target=\"_blank\" style=\"border-bottom-style:none; color:#606060;\">%s</a>", EXTDB_WOS_JCR_explanation_link, EXTDB_WOS_JCR_explanation);
		doc_puts(doc, "</div>");
	}
#endif

#define	Scopus_YEAR_begin	2011
#define	Scopus_YEAR_end	2020

#if defined(Scopus_ENGAGED)
	Link	ls, l_scopus = utlf_get_dict_list(xdoc, contentDict, "Scopus", NULL);
	if(l_scopus==NULL || link_atoms(l_scopus)==0) {
		doc_text_puts3(doc, "", "(No Scopus information.)", "<br />");
	} else if(!EIDisValid(ec->operator)) {	// EDB にログインしないユーザにはScopus情報は見せない．
		doc_text_puts3(doc, "", "(Scopus information is found. [need login])", "<br />");
	} else {
		ForLink(ls, l_scopus) {
			int	y;
			xmlNodePtr	scopusDict = ls->d;
			if(scopusDict==NULL) continue;

			char *cellStyle = " style=\"border-width:1px; border-style:solid; border-color:#808080; background-color:white;\"";
			char *naStyle = " style=\"border-width:1px; border-style:solid; border-color:#808080;\"";
			doc_puts(doc, "<table style=\"border-width:1px; border-color:black;border-style:solid;\">\n");

			doc_printf(doc, "<tr><td colspan=\"%d\" style=\"background-color:" EXTDB_Scopus_BGC ";\">", Scopus_YEAR_end-Scopus_YEAR_begin+2);
			doc_text_puts3(doc, "Title (Scopus): <b>", (char *)utlf_get_dict_text(xdoc, (utf8_t *)"(not defined)", scopusDict, "Title", NULL), "</b>\n");
			doc_puts(doc, "</td></tr>\n");

			doc_puts(doc, "<tr>\n");
			doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\"></th>");
			for(y=Scopus_YEAR_begin; y<=Scopus_YEAR_end; ++y) doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\"  align=\"center\">%d</th>", y);
			doc_puts(doc, "</tr>\n");

			doc_puts(doc, "<tr>\n");
			doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\"><a href=\"%s\" target=\"_blank\">SJR</a></th>", EXTDB_Scopus_SJR_explanation_link);
			for(y=Scopus_YEAR_begin; y<=Scopus_YEAR_end; ++y) {
				char	*ystr;
				asprintf(&ystr, "%d", y);
				char *sjr = (char *)utlf_get_dict_text(xdoc, NULL, scopusDict, ystr, "SJR", NULL);
				if(sjr!=NULL) {
					doc_printf(doc, "<td%s align=\"right\">", cellStyle);
					doc_text_puts3(doc, "", sjr, "</td>");
				} else {
					doc_printf(doc, "<td%s align=\"center\">n/a</td>", naStyle);
				}
				FREE(ystr);
			}
			doc_puts(doc, "</tr>\n");

#if 0		// IPP
			doc_puts(doc, "<tr>\n");
			doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\"><a href=\"%s\" target=\"_blank\">IPP</a></th>", EXTDB_Scopus_IPP_explanation_link);
			for(y=Scopus_YEAR_begin; y<=Scopus_YEAR_end; ++y) {
				char	*ystr;
				asprintf(&ystr, "%d", y);
				char *ipp = (char *)utlf_get_dict_text(xdoc, NULL, scopusDict, ystr, "IPP", NULL);
				if(ipp!=NULL) {
					doc_printf(doc, "<td%s align=\"right\">", cellStyle);
					doc_text_puts3(doc, "", ipp, "</td>");
				} else {
					doc_printf(doc, "<td%s align=\"center\">n/a</td>", naStyle);
				}
				FREE(ystr);
			}
			doc_puts(doc, "</tr>\n");
#else	// CiteScore
			doc_puts(doc, "<tr>\n");
			doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\"><a href=\"%s\" target=\"_blank\">CiteScore</a></th>", EXTDB_Scopus_CiteScore_explanation_link);
			for(y=Scopus_YEAR_begin; y<=Scopus_YEAR_end; ++y) {
				char	*ystr;
				asprintf(&ystr, "%d", y);
				char *citescore = (char *)utlf_get_dict_text(xdoc, NULL, scopusDict, ystr, "CiteScore", NULL);
				if(citescore==NULL) citescore = (char *)utlf_get_dict_text(xdoc, NULL, scopusDict, ystr, "CiteScore", "", NULL);
				if(citescore!=NULL) {
					doc_printf(doc, "<td%s align=\"right\">", cellStyle);
					doc_text_puts3(doc, "", citescore, "</td>");
				} else {
					doc_printf(doc, "<td%s align=\"center\">n/a</td>", naStyle);
				}
				FREE(ystr);
			}
			doc_puts(doc, "</tr>\n");
#endif

			doc_puts(doc, "<tr>\n");
			doc_printf(doc, "<th style=\"border-width:1px; border-style:solid; border-color:#808080;\" align=\"center\"><a href=\"%s\" target=\"_blank\">SNIP</a></th>", EXTDB_Scopus_SNIP_explanation_link);
			for(y=Scopus_YEAR_begin; y<=Scopus_YEAR_end; ++y) {
				char	*ystr;
				asprintf(&ystr, "%d", y);
				char *snip = (char *)utlf_get_dict_text(xdoc, NULL, scopusDict, ystr, "SNIP", NULL);
				if(snip!=NULL) {
					doc_printf(doc, "<td%s align=\"right\">", cellStyle);
					doc_text_puts3(doc, "", snip, "</td>");
				} else {
					doc_printf(doc, "<td%s align=\"center\">n/a</td>", naStyle);
				}
				FREE(ystr);
			}
			doc_puts(doc, "</tr>\n");

			doc_puts(doc, "</table>\n");
		}
		doc_puts(doc, "<div style=\"line-height:1; font-size:90%;\">");
		doc_printf(doc, "<div style=\"float:right;\">%s</div>", EXTDB_Scopus_poweredByScopus);
		doc_printf(doc, "<a href=\"%s\" target=\"_blank\" style=\"border-bottom-style:none; color:#606060;\">%s</a>", EXTDB_Scopus_SJR_explanation_link, EXTDB_Scopus_SJR_explanation);
		doc_puts(doc, "</div>");
#if 0		// IPP
		doc_puts(doc, "<div style=\"line-height:1; font-size:90%;\">");
		doc_printf(doc, "<div style=\"float:right;\">%s</div>", EXTDB_Scopus_poweredByScopus);
		doc_printf(doc, "<a href=\"%s\" target=\"_blank\" style=\"border-bottom-style:none; color:#606060;\">%s</a>", EXTDB_Scopus_IPP_explanation_link, EXTDB_Scopus_IPP_explanation);
		doc_puts(doc, "</div>");
#else	// CiteScore
		doc_puts(doc, "<div style=\"line-height:1; font-size:90%;\">");
		doc_printf(doc, "<div style=\"float:right;\">%s</div>", EXTDB_Scopus_poweredByScopus);
		doc_printf(doc, "<a href=\"%s\" target=\"_blank\" style=\"border-bottom-style:none; color:#606060;\">%s</a>", EXTDB_Scopus_CiteScore_explanation_link, EXTDB_Scopus_CiteScore_explanation);
		doc_puts(doc, "</div>");
#endif
		doc_puts(doc, "<div style=\"line-height:1; font-size:90%;\">");
		doc_printf(doc, "<div style=\"float:right;\">%s</div>", EXTDB_Scopus_poweredByScopus);
		doc_printf(doc, "<a href=\"%s\" target=\"_blank\" style=\"border-bottom-style:none; color:#606060;\">%s</a>", EXTDB_Scopus_SNIP_explanation_link, EXTDB_Scopus_SNIP_explanation);
		doc_puts(doc, "</div>");
	}
#endif

	if(xdoc) xmlFreeDoc(xdoc);
	if(ctxt) xmlFreeParserCtxt(ctxt);

	FREE(content);
	return TRUE;
}

int
edb_print_magazine_information(EdbContext ec, Doc doc, eid_t eid, char *omitIssn, int horizontal)
{
	EdbTupleColumn mtc;

	EdbTupleInfo	magazine = edb_tuple_get(ec, eid, 0);
	if(magazine==NULL) return FALSE;

	char *pissn=NULL, *eissn=NULL;

	mtc = edb_tuple_seek_column_by_xmlname(ec, magazine, (utf8_t *)"@.issn");
	if(mtc && TCDatumFirst(mtc)) pissn = (char *)DatumEnglish(TCDatumFirst(mtc));

	mtc = edb_tuple_seek_column_by_xmlname(ec, magazine, (utf8_t *)"@.eissn");
	if(mtc && TCDatumFirst(mtc)) eissn = (char *)DatumEnglish(TCDatumFirst(mtc));

	if(!(STRisValid(pissn) || STRisValid(eissn))) return FALSE;
	if(STRisValid(omitIssn)) {
		if((STRisValid(pissn) && !strcmp(pissn, omitIssn))
			|| (STRisValid(eissn) && !strcmp(eissn, omitIssn))) {
			doc_text_puts(doc, "(");
			if(STRisValid(pissn)) {
				doc_printf(doc, "pISSN: <span style=\"color:%s;\">%s</span>",
					!strcmp(pissn, omitIssn) ? "blue":"black", pissn);
			}
			if(STRisValid(pissn) && STRisValid(eissn)) doc_text_puts(doc, ", ");
			if(STRisValid(eissn)) {
				doc_printf(doc, "eISSN: <span style=\"color:%s;\">%s</span>",
					!strcmp(eissn, omitIssn) ? "blue":"black", eissn);
			}
			doc_text_puts(doc, ")");
			return TRUE;
		}
	}

	doc_puts(doc, "<table><tr><td valign=\"top\">");
	if(STRisValid(pissn)) {
		doc_text_puts3(doc, "(resolved by ", pissn, ")<br />");
		edb_print_issn_information(ec, doc, pissn);
	}
	doc_puts(doc, "</td>");
	if(!horizontal) doc_puts(doc, "</tr><tr>");
	doc_puts(doc, "<td valign=\"top\">");
	if(STRisValid(eissn)) {
		doc_text_puts3(doc, "(resolved by ", eissn, ")<br />");
		edb_print_issn_information(ec, doc, eissn);
	}
	doc_puts(doc, "</td></tr></table>");
	return TRUE;
}
#endif

#if defined(TUIR_COOPERATION)
static	char *	tuir_name;
static	char *	tuir_home;
static	char *	tuir_uploader;
static	Link		tuir_l_kinds = NULL;
static	int		tuir_date_from;
static	char *	tuir_secret;

static
int
tuir_load_config()
{
	if(STRisValid(tuir_secret)) return TRUE;
	Link l, l_lines = file_load_as_linked_str(EDBDIR "/etc/tuir.conf", stderr);
	if(l_lines==NULL) return FALSE;

	ForLink(l, l_lines) {
		char *s = l->d;
		if(s==NULL) continue;
		char *ss = index(s, '#');
		if(ss!=NULL) *ss = '\0';

		Link l_str = str_split(s, '\t');
		if(link_atoms(l_str)>=2) {
			char *tag = l_str->d;
			char *value = l_str->n->d;
			str_oneline(tag);
			str_oneline(value);
			if(!strcmp("name", tag)) {
				tuir_name = strdup(value);
			} else if(!strcmp("home", tag)) {
				tuir_home = strdup(value);
			} else if(!strcmp("uploader", tag)) {
				tuir_uploader = strdup(value);
			} else if(!strcmp("kind", tag)) {
				eid_t	kind = atoi(value);
				if(EIDisValid(kind)) {
					int *p_eid = (eid_t *)malloc(sizeof(eid_t));
					*p_eid = kind;
					LinkAppend(tuir_l_kinds, link_atom_new(p_eid, free));
				}
			} else if(!strcmp("date-from", tag)) {
				tuir_date_from = atoi(value);
			} else if(!strcmp("secret", tag)) {
				tuir_secret = strdup(value);
			}
		}
		EdbLinkFree(l_str);
	}
	EdbLinkFree(l_lines);
	return TRUE;
}

static
char *
tuir_make_token(char *eid, char *time, char *uid, char *remote, char *secret)
{
	char *data;

	SHA512_CTX context;
	SHA512_Init(&context);
	asprintf(&data, "/eid=%s/time=%s/uid=%s/remote=%s/secret=%s/", eid, time, uid, remote, secret);
	SHA512_Update(&context, (unsigned char *)data, strlen(data));
	free(data);
	return SHA512_End(&context, NULL);
}

static
char *
tuir_env_remote[] = {
	"HTTP_X_FORWARDED_FOR",
	"HTTP_CLIENT_IP",
	"REMOTE_ADDR",
};

char *
tuir_create_url(EdbContext ec, EdbTupleInfo tuple) {
	char *url;
	char *str_eid = NULL;
	char *str_time = NULL;
	char *str_uid = NULL;
	char *str_remote = NULL;
	char *str_token = NULL;
	if(!tuir_load_config()) return NULL;		// 設定ファイル読み込み失敗
	if(!EIDisValid(ec->operator)) return NULL;	// 無効なオペレータ
	if(!edb_auth_writable(ec, TupleEID(tuple))) return NULL;	// 書き込みできないオペレータ
	if(tuple==NULL || !EIDisValid(TupleEID(tuple)) || !TupleAvail(tuple)) return NULL;	// 無効な情報
	if(!TupleIsArticle(tuple)) return NULL;	// 著作のみ
	// 著作種別
	{
		int	matched = FALSE;
		eid_t kind = 0;
		Link	l;
		EdbTupleColumn tc = edb_tuple_seek_column_by_xmlname(ec, tuple, (utf8_t *)"@.kind");
		if(tc==NULL) return NULL;
		EdbDatum datum = TCDatumFirst(tc);
		if(datum==NULL) return NULL;
		kind = DatumEID(datum);
		if(!EIDisValid(kind)) return NULL;
		ForLink(l, tuir_l_kinds) {
			if((*((eid_t *)l->d))==kind) {
				matched = TRUE;
				break;
			}
		}
		if(!matched) return NULL;
	}
	// 発表年月日
	{
		EdbTupleColumn tc = edb_tuple_seek_column_by_xmlname(ec, tuple, (utf8_t *)"@.date");
		if(tc==NULL) return NULL;
		EdbDatum datum = TCDatumFirst(tc);
		if(datum==NULL) return NULL;
		if(edb_datum_get_DATE(ec, datum)<tuir_date_from) return NULL;
	}
	// eid
	asprintf(&str_eid, "%d", TupleEID(tuple));
	// time
	{
		time_t t;
		struct tm tm;
		time(&t);
		bcopy(gmtime(&t), &tm, sizeof(struct tm));
		asprintf(&str_time, "%04d-%02d-%02dT%02d:%02d:%02dZ",
			tm.tm_year+1900,
			tm.tm_mon + 1,
			tm.tm_mday,
			tm.tm_hour,
			tm.tm_min,
			tm.tm_sec);
	}
	// uid
	asprintf(&str_uid, "S%d", ec->operator);
	// remote
	for(int i=0; i<Ns(tuir_env_remote); ++i) {
		char *s = getenv(tuir_env_remote[i]);
		if(s==NULL) continue;
		str_remote = strdup(s);
		break;
	}
	// token
	str_token = tuir_make_token(str_eid, str_time, str_uid, str_remote, tuir_secret);

	asprintf(&url, "%s?eid=%s&time=%s&uid=%s&token=%s", tuir_uploader, str_eid, str_time, str_uid, str_token);
	FREE(str_eid);
	FREE(str_uid);
	FREE(str_time);
	FREE(str_remote);
	FREE(str_token);
	return url;
}

#define TUIR_FRAME_HEIGHT_MIN	80
#define TUIR_FRAME_HEIGHT_MAX	384
void
tuir_show_iframe(EdbContext ec, char *tag, EdbTupleInfo tuple, char *url, Doc doc)
{
	doc_printf(doc, "<%s>", tag);
	doc_printf(doc, "<a href=\"%s\" target=\"_blank\">", tuir_home);
	doc_text_puts(doc, tuir_name);
	doc_puts(doc, "</a>");
	doc_text_printf(doc, " 連携");

////	doc_puts(doc, "<span style=\"color:red;\">");
////	doc_text_printf(doc, " (試行中)");
////	doc_puts(doc, "</span>");

	doc_printf(doc, " <input type=\"button\" value=\"\u21f3\" title=\"広げる\" onclick=\"document.getElementById('tuir_frame').height=(document.getElementById('tuir_frame').height<%d)?%d:%d;\" />\n",
		TUIR_FRAME_HEIGHT_MAX, TUIR_FRAME_HEIGHT_MAX, TUIR_FRAME_HEIGHT_MIN);
	doc_printf(doc, " <input type=\"button\" value=\"別窓で開く\" title=\"フレームの内容を別ウィンドウで開く\" onclick=\"window.open('%s', '_blank')\" />\n", url);
	doc_printf(doc, "</%s>\n", tag);
	doc_puts(doc, "<div class=\"contents\">\n");
	doc_puts(doc, " <span style=\"font-size:90%;\">");
	doc_text_printf(doc, "（次のフレームは本著作に関して%sに問い合わせた結果を表示しています．カーソルを乗せるとフレームが縦に広がります．）", tuir_name);
	doc_puts(doc, "</span>");
	doc_printf(doc, "<iframe id=\"tuir_frame\" name=\"tuir_frame\" src=\"%s\" width=\"100%%\" height=\"%d\" frameborder=\"1\" scrolling=\"auto\""
		" onmouseover=\"if(this.height<%d-1) this.height=%d-1;\" onmouseout=\"if(this.height<%d) this.height=%d;\">'\n",
		url,
		TUIR_FRAME_HEIGHT_MIN,
		TUIR_FRAME_HEIGHT_MAX,
		TUIR_FRAME_HEIGHT_MAX,
		TUIR_FRAME_HEIGHT_MAX,
		TUIR_FRAME_HEIGHT_MIN);
	doc_puts(doc, "'<span style=\"color:red;\">インラインフレーム(&lt;iframe&gt;)に対応したブラウザをお使いください．</span>'\n");
	doc_puts(doc, "</iframe>\n");
	doc_puts(doc, "</div>\n");
}

#endif	/* TUIR_COOPERATION */


static
int
edb_print_tuple_check_matched(EdbContext ec, EdbCatalogue ca, EdbTableInfo ti, EdbColumnInfo ci, EdbMapLookup map, char *str, int lang)
{
	Link	l, l_tuple;
	char	*s, *target = NULL;
	ucs4_t	*u, *ustr;
	int	matched = FALSE;

	if(!map) return FALSE;
	if(str==NULL || strlen(str)==0) return FALSE;

	ustr = UTF8toUCS4((utf8_t *)str);
	for(u=ustr; *u; ++u) {
		if(ucs4_isascii(*u) && !ucs4_isalnum(*u)) *u = ' ';
	}
	ucs4_str_oneline(ustr);
	if(!UCS4_STRisVALID(ustr)) return FALSE;
	for(u=ustr; *u; ++u) {
		if(*u==' ') *u = '%';
	}
	str = (char *)UCS4toUTF8(ustr);

	asprintf(&target, "%s", str);

	l_tuple = edb_table_search_tuples_byText(ec, ti, (utf8_t *)target, lang);
	FREE(target);
	if(l_tuple) matched = TRUE;
	ForLink(l, l_tuple) {
		EdbTupleInfo	tuple = (EdbTupleInfo)l->d;
		if(tuple) edb_catalogue_append(ca, TupleEID(tuple), tuple);
	}
	link_free(l_tuple);
	return matched;
}

static
void
edb_print_tuple_column_YEAR(EdbPrint ep, int y)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	int	ymd = y*10000+401;

	HTML(doc) {
		static struct {
			char	*name;
			int	d;
		} era_ymd[] = {
			{"明治", 18680908},
			{"大正", 19120730},
			{"昭和", 19261225},
			{"平成", 19890108},
			{"令和", 20190501}
		};
		int	i, year;

		year = DATE_Y(ymd);
		doc_printf(doc, "西暦 %d年", year);
		for(i=Ns(era_ymd)-1; i>=0; --i) {
			if(ymd>=era_ymd[i].d) {
				doc_puts(doc, "<span style=\"font-size:90%;\">");
				doc_printf(doc, " (%s ", era_ymd[i].name);
				year -= (era_ymd[i].d/10000 - 1);
				if(year==1) {
					doc_printf(doc, "元年");
				} else {
					doc_printf(doc, "%d年", year);
				}
				doc_printf(doc, ") ");
				doc_puts(doc, "</span>");
				break;
			}
		}
	}
	PLAIN(doc) doc_printf(doc, "西暦 %d年", DATE_Y(ymd));
	TeX(doc) doc_printf(doc, "%d", DATE_Y(y));

}

static
void
edb_print_tuple_column_MINUTE(EdbPrint ep, int m)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;

	HTML(doc) {
		doc_printf(doc, "%d分", m);
		if(m>60) doc_printf(doc, " (%d時間%02d分)", m/60, m%60);
	}
	PLAIN(doc) doc_printf(doc, "%d分", m);
	TeX(doc) doc_printf(doc, "%d", m);


}

static
void
edb_print_tuple_column_DATE(EdbPrint ep, int ymd)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;

	HTML(doc) {
		static struct {
			char	*name;
			int	d;
		} era_ymd[] = {
			{"明治", 18680908},
			{"大正", 19120730},
			{"昭和", 19261225},
			{"平成", 19890108},
			{"令和", 20190501}
		};
		int	i, year, month, day;

		if(ymd<=EDB_DATE_MIN) return;
		if(ymd>=EDB_DATE_MAX) return;
		year = DATE_Y(ymd);
		month = DATE_M(ymd);
		day = DATE_D(ymd);
		doc_printf(doc, "西暦 %d年 %d月 ", year, month);
		if(day==0) {
			doc_printf(doc, " 初日");
		} else if(day>=32) {
			doc_printf(doc, " 末日");
		} else {
			doc_printf(doc, " %d日", day);
		}
		for(i=Ns(era_ymd)-1; i>=0; --i) {
			if(ymd>=era_ymd[i].d) {
				doc_puts(doc, "<span style=\"font-size:90%;\">");
				doc_printf(doc, " (%s ", era_ymd[i].name);
				year -= (era_ymd[i].d/10000 - 1);
				if(year==1) {
					doc_printf(doc, "元年");
				} else {
					doc_printf(doc, "%d年", year);
				}
				doc_printf(doc, " %d月", month);
				if(day==0) {
					doc_printf(doc, " 初日");
				} else if(day>=32) {
					doc_printf(doc, " 末日");
				} else {
					doc_printf(doc, " %d日", day);
				}
				doc_printf(doc, ") ");
				doc_puts(doc, "</span>");
				break;
			}
		}
	}
	PLAIN(doc) {
		if(ymd<=EDB_DATE_MIN) return;
		if(ymd>=EDB_DATE_MAX) return;
		doc_printf(doc, "西暦 %d年 %d月 %d日", DATE_Y(ymd), DATE_M(ymd), DATE_D(ymd));
	}
	TeX(doc) doc_printf(doc, "\\DATE{%d}{%d}{%d}", DATE_Y(ymd), DATE_M(ymd), DATE_D(ymd));

}

static
void
edb_print_tuple_datum_period(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;

	if(!tc->ci->chronological) return;
	if(EDB_DATE_MIN < DatumDateFrom(datum) || DatumDateTo(datum) < EDB_DATE_MAX) {
		doc_puts(doc, "<span style=\"font-size:90%;\">");
		doc_puts(doc, "(期間: ");
		if(EDB_DATE_MIN < DatumDateFrom(datum)) {
			int day;
			doc_printf(doc, "%d年%d月", 
				DATE_Y(DatumDateFrom(datum)),
				DATE_M(DatumDateFrom(datum)));
			day = DATE_D(DatumDateFrom(datum));
			if(day==0) {
				doc_printf(doc, " 初日");
			} else if(day>=32) {
				doc_printf(doc, " 末日");
			} else {
				doc_printf(doc, " %d日", day);
			}
		}
		doc_puts(doc, "〜");
		if(DatumDateTo(datum) < EDB_DATE_MAX) {
			int day;
			doc_printf(doc, "%d年%d月", 
				DATE_Y(DatumDateTo(datum)),
				DATE_M(DatumDateTo(datum)));
			day = DATE_D(DatumDateTo(datum));
			if(day==0) {
				doc_printf(doc, " 初日");
			} else if(day>=32) {
				doc_printf(doc, " 末日");
			} else {
				doc_printf(doc, " %d日", day);
			}
		}
		doc_puts(doc, ")&nbsp;");
		doc_puts(doc, "</span>\n");
	}
}

void
edb_print_tuple_datum_permission(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	EdbColumnInfo	ci = tc->ci;
	EdbTableInfo	ti = TupleTable(tuple);
	char	*s, *fgc = "#404040";
	int	l;

	if(datum && DatumRead(datum)!=EDB_AUTH_INHERIT) {
		s = (char *)edb_auth_string(ec, l=DatumRead(datum), FALSE);
		fgc = "#ff4040";
	} else if(ColumnTupleRead(ci)!=EDB_AUTH_INHERIT) {
		s = (char *)edb_auth_string(ec, l=ColumnTupleRead(ci), FALSE);
	} else {
		s = (char *)edb_auth_string(ec, l=TABLE_TupleRead(ti), TRUE);
	}
	if(STRisVALID(s)) {
		if(STRisVALID(fgc)) doc_printf(doc, "<span style=\"color:%s;\">", fgc);
		doc_puts(doc, "&nbsp;&nbsp;[公開範囲: ");
		doc_puts(doc, s);
		doc_puts(doc, "] ");
		if(TupleRead(tuple)>l && TupleRead(tuple)<=EDB_AUTH_USER) {
			doc_printf(doc, "<span style=\"color:%s\">", "#ff4040");
			doc_text_puts3(doc, "(", (char *)edb_auth_string(ec, TupleRead(tuple), TRUE), ")");
			doc_puts(doc, "</span>");
		}
		if(STRisVALID(fgc)) doc_puts(doc, "</span>");
	}
}

static
void
edb_print_tuple_datum_permission0(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	EdbColumnInfo	ci = tc->ci;
	EdbTableInfo	ti = TupleTable(tuple);
	char	*s, *fgc = "#404040";
	int	l;

#if 0
	if(datum && DatumRead(datum)!=EDB_AUTH_INHERIT) {
		s = (char *)edb_auth_string(ec, l=DatumRead(datum), FALSE);
		fgc = "#ff4040";
	} else if(ColumnTupleRead(ci)!=EDB_AUTH_INHERIT) {
		s = (char *)edb_auth_string(ec, l=ColumnTupleRead(ci), FALSE);
	} else {
		s = (char *)edb_auth_string(ec, l=TABLE_TupleRead(ti), TRUE);
	}
#else
	if(datum && DatumRead(datum)!=EDB_AUTH_INHERIT) {
		s = (char *)edb_auth_string(ec, l=DatumRead(datum), FALSE);
		fgc = "#ff4040";
	} else {
		s = (char *)edb_auth_string(ec, l=ColumnTupleRead(ci), FALSE);
	}
#endif
	if(STRisVALID(s)) {
		if(STRisVALID(fgc)) doc_printf(doc, "<span style=\"color:%s\">", fgc);
		doc_puts(doc, "<span style=\"font-size:90%;\">");
		doc_text_puts3(doc, "[", s, "]");
		if(TupleRead(tuple)>l && TupleRead(tuple)<=EDB_AUTH_USER) {
			doc_printf(doc, "<span style=\"color:%s\">", "#ff4040");
			doc_text_puts3(doc, "(", (char *)edb_auth_string(ec, TupleRead(tuple), TRUE), ")");
			doc_puts(doc, "</span>");
		}
		doc_puts(doc, "</span>");
		if(STRisVALID(fgc)) doc_puts(doc, "</span>");
	}
}

void
edb_print_tuple_column_datum(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum d, int mode)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	Caption	cap;
	register eid_t	eid;
	int32_t	date1, date2;

	eid = DatumEID(d);
	if(EIDisValid(eid)) {
		edb_print_tuple_datum_period(ep, tuple, tc, d);
		if((cap = edb_make_mapped_caption(ec, tc->ci, eid, CAP_long|((mode&EDB_PRINT_MODE_CAPNOMAPTO)?CAP_nomapto:0)))!=NULL) {
			if(mode&(EDB_PRINT_MODE_LINK|EDB_PRINT_MODE_WWWLINK)) {
				doc_printf(doc, "<a href=\"%s?EID=%d\"", ec->browser_url, eid);
				if(mode&EDB_PRINT_MODE_BLANKTARGET) doc_puts(doc, " target=\"_blank\"");
				doc_puts(doc, ">");
				doc_text_puts(doc, (char *)cap->c_main);
				doc_puts(doc, "</a>");
			} else {
				doc_text_puts(doc, (char *)cap->c_main);
			}
			if(STRisVALID(cap->c_sub)) {
				HTML(doc) {
					doc_text_puts3(doc, " <span style=\"font-size:90%;\">(", (char *)cap->c_sub, ")</span>");
				}
			}
			edb_caption_free(ec, cap);
		} else if(DatumEnglishIsValid(d)) {
			doc_text_printf(doc, "%s", DatumEnglish(d));
		} else if(DatumJapaneseIsValid(d)) {
			doc_text_printf(doc, "%s", DatumJapanese(d));
		} else if(DatumPronounceIsValid(d)) {
			doc_text_printf(doc, "%s", DatumPronounce(d));
		}
#if 0
		edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
	} else {
		switch(tc->ci->typeInfo->basicTypeID) {
		case EDB_TYPE_NAME:
			HTML(doc) {
				doc_text_puts3(doc, " <span style=\"font-size:80%;\">(英)</span> ", (char *)DatumEnglish(d), NULL);
				doc_text_puts3(doc, " <span style=\"font-size:80%;\">(日)</span> ", (char *)DatumJapanese(d), NULL);
				doc_text_puts3(doc, " <span style=\"font-size:80%;\">(読)</span> ", (char *)DatumPronounce(d), NULL);
			}
			PLAIN(doc) {
				doc_text_puts3(doc, " (英) ", (char *)DatumEnglish(d), NULL);
				doc_text_puts3(doc, " (日) ", (char *)DatumJapanese(d), NULL);
				doc_text_puts3(doc, " (読) ", (char *)DatumPronounce(d), NULL);
			}
			TeX(doc) {
				doc_text_puts3(doc, "\\NAME{", (char *)DatumEnglish(d), "}");
				doc_text_puts3(doc, "{", (char *)DatumJapanese(d), "}");
				doc_text_puts3(doc, "{", (char *)DatumPronounce(d), "}");
			}
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		case EDB_TYPE_SENTENCE:
			HTML(doc) {
				doc_text_puts3(doc, "<span style=\"font-size:80%;\">(英)</span> ", (char *)DatumEnglish(d), "&nbsp;&nbsp;");
				doc_text_puts3(doc, "<span style=\"font-size:80%;\">(日)</span> ", (char *)DatumJapanese(d), "&nbsp;&nbsp;");
				edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
				edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			}
			PLAIN(doc) {
				doc_text_puts3(doc, " (英) ", (char *)DatumEnglish(d), NULL);
				doc_text_puts3(doc, " (日) ", (char *)DatumJapanese(d), NULL);
			}
			TeX(doc) {
				doc_text_puts3(doc, "\\SENTENCE{", (char *)DatumEnglish(d), "}");
				doc_text_puts3(doc, "{", (char *)DatumJapanese(d), "}");
			}
			break;
		case EDB_TYPE_EMAIL:
			if((mode&EDB_PRINT_MODE_LINK) || (mode&EDB_PRINT_MODE_WWWLINK)) {
#if 0
				doc_printf(doc, "<a href=\"mailto:%s\"", DatumEnglish(d));
				if(mode&EDB_PRINT_MODE_BLANKTARGET) doc_puts(doc, " target=\"_blank\"");
				doc_printf(doc, ">%s</a>", DatumEnglish(d));
#else
				edb_print_type_EMAIL_in_HTML_avoid_SPAM(ep, tuple, tc, d);
#endif
			} else {
				HTML(doc) doc_printf(doc, "%s", DatumEnglish(d));
				PLAIN(doc) doc_printf(doc, "%s", DatumEnglish(d));
				TeX(doc) doc_printf(doc, "\\EMAIL{%s}", DatumEnglish(d));
			}
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		case EDB_TYPE_DATE:
			edb_print_tuple_column_DATE(ep, ATOI((char *)DatumEnglish(d)));
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		case EDB_TYPE_DATE2:
			edb_datum_get_DATE2(ec, d, &date1, &date2);
			edb_print_tuple_column_DATE(ep, date1);
			doc_puts(doc, "〜");
			edb_print_tuple_column_DATE(ep, date2);
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		case EDB_TYPE_REAL:
			if(tc->ci->typeInfo->typeID==EDB_TYPE_MONETARY) {
				doc_printf(doc, "%.1f千円", ATOF((char *)DatumEnglish(d)));
			} else {
				doc_printf(doc, "%f", ATOF((char *)DatumEnglish(d)));
			}
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		case EDB_TYPE_URL:
			HTML(doc) {
				if((mode&EDB_PRINT_MODE_LINK) || (mode&EDB_PRINT_MODE_WWWLINK)) {
					doc_printf(doc, "<a href=\"%s\" target=\"_blank\">", DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");
				} else {
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "");
				}
			}
			PLAIN(doc) doc_text_puts3(doc, NULL, (char *)DatumEnglish(d), NULL);
			TeX(doc) 	doc_text_puts3(doc, "\\URL{", (char *)DatumEnglish(d), "}");
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		case EDB_TYPE_PHONE:
			HTML(doc) doc_text_puts(doc, (char *)DatumEnglish(d));
			PLAIN(doc) doc_text_puts(doc, (char *)DatumEnglish(d));
			TeX(doc) doc_text_puts3(doc, "\\PHONE{", (char *)DatumEnglish(d), "}");
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		case EDB_TYPE_PICTURE:
			doc_puts(doc, "<table border=\"0\"><tr><td>");
			doc_text_printf(doc, "\\P{%d; height=8}", TupleEID(tuple));
			doc_puts(doc, "</td><td valign=\"top\" style=\"font-size:80%;\">");
			{
				EdbPicture	picture;

				picture = edb_picture_new(ec, tuple);
				if(picture) {
					doc_text_printf(doc, " mimetype = \"%s\";\n", picture->mimetype);
					doc_text_printf(doc, " xsize = %d [pixel];\n", picture->xsize);
					doc_text_printf(doc, " ysize = %d [pixel];\n", picture->ysize);
					doc_text_printf(doc, " size = %d [byte];\n", picture->bytes);
					if(picture->dpi != 0) {
						doc_text_printf(doc, " --- \n");
						doc_text_printf(doc, " actual xsize = %.1f [mm] (= %.2f [inch]);\n", (double)picture->xsize*MM_PER_INCH/picture->dpi, (double)picture->xsize/picture->dpi);
						doc_text_printf(doc, " actual ysize = %.1f [mm] (= %.2f [inch]);\n", (double)picture->ysize*MM_PER_INCH/picture->dpi, (double)picture->ysize/picture->dpi);
					}
					if(picture->xsize>0) doc_text_printf(doc, " aspect ratio (ysize/xsize) = %.2f ;\n", (double)picture->ysize/picture->xsize);
					edb_picture_free(ec, picture);
				}
			}
			doc_puts(doc, "</td></tr></table>");
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		case EDB_TYPE_PASSPHRASE:
			if(STRisValid(ColumnTypeCryptography(tc->ci))) {
				if(DatumEnglishIsValid(d)) {
					doc_text_puts(doc, "(暗号化パスフレーズが設定されています)");
				}
			} else {
				doc_text_puts(doc, (char *)DatumEnglish(d));
			}
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		case EDB_TYPE_CERTIFICATE:
			{
				EdbPkiCertificateInfo	info = edb_pki_certificate_info(ec, DatumEnglish(d));
				int	valid = edb_pki_verify_certificate(ec, DatumEnglish(d));

				if(info && STRisValid(info->S.DN.CN) && mode&(EDB_PRINT_MODE_LINK|EDB_PRINT_MODE_WWWLINK)) {
					if(valid) {
						doc_printf(doc, "<a href=\"/PKI/CN/%s.crt\" target=\"_blank\">", info->S.DN.CN);
						doc_puts(doc, "<img align=\"left\" valign=\"top\" alt=\"Certificate\" border=\"0\" src=\"" EDB_PKI_ICON_CERTIFICATE "\" />");
						doc_puts(doc, "</a>");
					} else {
						doc_puts(doc, "<img align=\"left\" valign=\"top\" alt=\"Revoked\" border=\"0\" src=\"" EDB_PKI_ICON_CERTIFICATE_REVOKED "\" />");
					}
				} else {
					if(valid) {
						doc_puts(doc, "<img align=\"left\" valign=\"top\" alt=\"Certificate\" border=\"0\" src=\"" EDB_PKI_ICON_CERTIFICATE "\" />\n");
					} else {
						doc_puts(doc, "<img align=\"left\" valign=\"top\" alt=\"Revoked\" border=\"0\" src=\"" EDB_PKI_ICON_CERTIFICATE_REVOKED "\" />\n");
					}
				}
				if(info) {
					doc_puts(doc, "<span style=\"font-size:80%;\">");
					if(!valid) doc_color_begin(doc, EP_WARNING_COLOR);

					if(!strcmp((char *)info->I.DN.CN, EDB_PKI_ROOT_CN)) {
						doc_text_puts3(doc, "<span style=\"font-size:120%; font-weight:bold\">", EDB_PKI_CERT_STRING, "</span><br />\n");
					} else if(!strcmp((char *)info->I.DN.CN, EDB_PKI2_ROOT_CN)) {
						doc_text_puts3(doc, "<span style=\"font-size:120%; font-weight:bold\">", EDB_PKI2_CERT_STRING, "</span><br />\n");
					}

					doc_text_puts3(doc, "countryName (C) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.C, "</span>\"<br />\n");
					doc_text_puts3(doc, "stateOrProvinceName (ST) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.ST, "</span>\"<br />\n");
					doc_text_puts3(doc, "localityName (L) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.L, "</span>\"<br />\n");
					doc_text_puts3(doc, "organizationName (O) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.O, "</span>\"<br />\n");
					doc_text_puts3(doc, "organizationalUnitName (OU) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.OU, "</span>\"<br />\n");
					doc_text_puts3(doc, "commonName (CN) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.CN, "</span>\"<br />\n");
#if 0
					doc_text_puts3(doc, "emailAddress = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.Email, "</span>\"<br />\n");
#else
					doc_puts(doc, "emailAddress = \"<span style=\"font-weight:bold\">");
					if(STRisValid(info->S.DN.Email)) edb_print_emailaddress_in_HTML_avoid_SPAM(doc, (char *)info->S.DN.Email);
					doc_puts(doc, "</span>\"<br />\n");
#endif
					if(info->l_subjectAltNames!=NULL) {
						Link	l;
						ForLink(l, info->l_subjectAltNames) {
							doc_text_puts3(doc, "subjectAltName = \"<span style=\"font-weight:bold\">", (char *)l->d, "</span>\"<br />\n");
						}
					}
					if(!valid) doc_color_end(doc);
					doc_puts(doc, "</span>\n");
				}
				edb_pki_certificate_info_free(ec, info);
			}
			edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
			edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
			break;
		default:
			switch(tc->ci->typeInfo->typeID) {
			case EDB_TYPE_EXTID:
				if(!strcmp((char *)ColumnXN(tc->ci), "article.doi")) {
					doc_printf(doc, "<a href=\"%s%s\" target=\"_blank\">", EXTDB_DOI_Article_BASE_URL, DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");

#if defined(WOS_ENGAGED)
					doc_puts(doc, "<span style=\"font-size:90%;\">&nbsp;&nbsp;&nbsp;&nbsp;(→<a href=\"");
					doc_puts(doc, EXTDB_WOSbyDOI_Article_BASE_URL);
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "\" target=\"_blank\">");
					doc_puts(doc, "Web of Science<sup>TM</sup>で検索");
					doc_puts(doc, "</a>)</span>");
#endif

#if defined(Scopus_ENGAGED)
					doc_puts(doc, "<span style=\"font-size:90%;\">&nbsp;&nbsp;&nbsp;&nbsp;(→<a href=\"");
					doc_puts(doc, EXTDB_ScopusByDOI_Article_BASE_URL);
					char *s, *doi = (char *)DatumEnglish(d);
					int len = 3+1+strlen(doi)+1;
					doc_printf(doc, "&sl=%d", len);
					doc_text_puts(doc, "&s=DOI%28");
					for(s=doi; *s; ++s) {
						if(isalnum(*s) || *s=='.') {
							doc_putc(doc, *s);
						} else {
							doc_printf(doc, "%%%02x", (*s)&0xff);
						}
					}
					doc_text_puts(doc, "%29");
					doc_puts(doc, "\" target=\"_blank\">");
					doc_text_puts(doc, "Scopusで検索");
					doc_puts(doc, "</a>)</span>");
#endif
				} else if(!strcmp((char *)ColumnXN(tc->ci), "article.pmid")) {
					doc_printf(doc, "<a href=\"%s%s\" target=\"_blank\">", EXTDB_PubMed_Article_BASE_URL, DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");

#if defined(WOS_ENGAGED)
					doc_puts(doc, "<span style=\"font-size:90%;\">&nbsp;&nbsp;&nbsp;&nbsp;(→<a href=\"");
					doc_puts(doc, EXTDB_WOSbyPMID_Article_BASE_URL);
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "\" target=\"_blank\">");
					doc_puts(doc, "Web of Science<sup>TM</sup>で検索");
					doc_puts(doc, "</a>)</span>");
#endif

#if defined(Scopus_ENGAGED)
					doc_puts(doc, "<span style=\"font-size:90%;\">&nbsp;&nbsp;&nbsp;&nbsp;(→<a href=\"");
					doc_puts(doc, EXTDB_ScopusByDOI_Article_BASE_URL);
					char *s, *pmid = (char *)DatumEnglish(d);
					int len = 4+1+strlen(pmid)+1;
					doc_printf(doc, "&sl=%d", len);
					doc_text_puts(doc, "&s=PMID%28");
					for(s=pmid; *s; ++s) {
						if(isalnum(*s) || *s=='.') {
							doc_putc(doc, *s);
						} else {
							doc_printf(doc, "%%%02x", (*s)&0xff);
						}
					}
					doc_text_puts(doc, "%29");
					doc_puts(doc, "\" target=\"_blank\">");
					doc_text_puts(doc, "Scopusで検索");
					doc_puts(doc, "</a>)</span>");
#endif
				} else if(!strcmp((char *)ColumnXN(tc->ci), "article.naid")) {
					doc_printf(doc, "<a href=\"%s%s\" target=\"_blank\">", EXTDB_CiNii_Article_BASE_URL, DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");
				} else if(!strcmp((char *)ColumnXN(tc->ci), "article.wos")) {
					doc_printf(doc, "<a href=\"" EXTDB_WOS_Article_BASE_URL "\" target=\"_blank\">", DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");
				} else if(!strcmp((char *)ColumnXN(tc->ci), "article.scopus")) {
					doc_printf(doc, "<a href=\"" EXTDB_Scopus_Article_BASE_URL "\" target=\"_blank\">", DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");
				} else if(!strcmp((char *)ColumnXN(tc->ci), "article.tuir")) {
					doc_printf(doc, "<a href=\"" EXTDB_TUIR_Article_BASE_URL "\" target=\"_blank\">", DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");
				} else if(!strcmp((char *)ColumnXN(tc->ci), "person.wos")) {
					doc_printf(doc, "<a href=\"" EXTDB_WOS_Researcher_BASE_URL "\" target=\"_blank\">", DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");
				} else if(!strcmp((char *)ColumnXN(tc->ci), "person.scopus")) {
					doc_printf(doc, "<a href=\"" EXTDB_Scopus_Author_BASE_URL "\" target=\"_blank\">", DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");
				} else if(!strcmp((char *)ColumnXN(tc->ci), "person.orcid")) {
					doc_printf(doc, "<a href=\"" EXTDB_ORCID_BASE_URL "\" target=\"_blank\">", DatumEnglish(d));
					doc_text_puts(doc, (char *)DatumEnglish(d));
					doc_puts(doc, "</a>");
				} else {
					doc_text_puts(doc, (char *)DatumEnglish(d));
				}
				edb_print_tuple_datum_period(ep, tuple, tc, d);
				break;
			case EDB_TYPE_YEAR:
				edb_print_tuple_column_YEAR(ep, ATOI((char *)DatumEnglish(d)));
				edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
				edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
				break;
			case EDB_TYPE_MINUTE:
				edb_print_tuple_column_MINUTE(ep, ATOI((char *)DatumEnglish(d)));
				edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
				edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
				break;
			default:
				doc_text_puts(doc, (char *)DatumEnglish(d));
				edb_print_tuple_datum_period(ep, tuple, tc, d);
#if 0
				edb_print_tuple_datum_permission(ep, tuple, tc, d);
#endif
				break;
			}
		}
	}
}

static
int
check_studentcode(utf8_t *code)
{
	if(!STRisValid(code)) return FALSE;
	if(strlen((char *)code)!=10) return FALSE;
	utf8_t *c;
	for(c=code; *c; ++c) {
		if(!isdigit(*c)) return FALSE;
	}
	int i, s=0, sum = 0;
	for(i=0; i<9; ++i) {
		if(!isdigit(code[i])) return FALSE;
		s += code[i]-'0';
		sum += s;
	}
	sum += code[9]-'0';
	sum %= 11;
	return (sum==0 || sum==1);
}

void
edb_print_student_info(EdbContext ec, StudentInfo info, int full, Doc doc)
{

	doc_puts(doc, " 《");
	if(full) doc_printf(doc, "%s: ", info->v_code);
	doc_printf(doc, "%s (%s", info->v_name_ja, info->v_name_en);
	if(full && STRisValid(info->v_name_pr)) {
		if(STRisValid(info->v_name_en)) doc_puts(doc, ", ");
		doc_printf(doc, "%s", info->v_name_pr);
	}
	doc_puts(doc, ") @ ");
	utf8_t *org = NULL;
	if(!STRisValid(org) && full) org = info->v_division;
	if(!STRisValid(org) && full) org = info->v_department;
	if(!STRisValid(org)) org = info->v_faculty;
	if(STRisValid(org)) doc_printf(doc, "%s", org);
	doc_printf(doc, " [%s]", info->v_title);
	if(full || TRUE) {
		doc_puts(doc, " ");
		if(info->v_enter>0) doc_printf(doc, "%d年%d月", DATE_Y(info->v_enter), DATE_M(info->v_enter));
		doc_puts(doc, "〜");
		if(info->v_leave>0) doc_printf(doc, "%d年%d月", DATE_Y(info->v_leave), DATE_M(info->v_leave));
	}
	doc_puts(doc, "》");
}

void
edb_print_student_list(EdbContext ec, utf8_t *code, int full, char *class, char *style, int limits, Doc doc)
{

	Link l_info = edb_table_get_student_info(ec, code);
	HTML(doc) {
		doc_puts(doc, "<span");
		if(STRisValid(class)) doc_printf(doc, " class=\"%s\"", class);
		doc_puts(doc, " style=\"");
		if(STRisValid(style)) doc_printf(doc, "%s; ", style);
	}
	if(l_info) {
		HTML(doc) doc_puts(doc, "color:#008000;\">");
		Link	l;
		ForLink(l, l_info) {
			if(limits-->0) {
				edb_print_student_info(ec, (StudentInfo)l->d, full, doc);
			} else {
				doc_puts(doc, "...");
				break;
			}
		}
		link_free(l_info);
	} else {
		HTML(doc) doc_puts(doc, "color:#c00000;\">");
		if(!check_studentcode(code)) {
			doc_printf(doc, " 《学籍番号ではありません》");
		} else {
			doc_printf(doc, " 《学籍番号に対応する情報が見つかりません》");
		}
	}
	HTML(doc) doc_puts(doc, "</span>");
}

static
void
edb_print_tuple_column_index(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, int idx, char *tag, int mode)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	int	i;
	Caption	cap;
	int	e = idx;
	int	auth;
	EdbDatum	d;

	if(ec==NULL || tuple == NULL || tc == NULL || doc==NULL) return;
	d = edb_tcdatum_get_datum(ec, tc, e);
#ifdef EDB_READER
#else /* !EDB_READER */
	if(d && DatumAvail(d)) {
		if(!edb_auth_readable_TCD(ec, tuple, tc, d)) {
			if(mode&EDB_PRINT_MODE_ANCHOR) doc_printf(doc, "<a name=\"%s-%d\">", ColumnSN(tc->ci), idx+1);
			doc_printf(doc, "%s****", tag);
			if(mode&EDB_PRINT_MODE_ANCHOR) doc_puts(doc, "</a>");
			return;
		}
	} else {
		if(!edb_auth_readable_tuple_column(ec, tuple, tc->ci)) {
			if(mode&EDB_PRINT_MODE_ANCHOR) doc_printf(doc, "<a name=\"%s-%d\">", ColumnSN(tc->ci), idx+1);
			doc_printf(doc, "%s****", tag);
			if(mode&EDB_PRINT_MODE_ANCHOR) doc_puts(doc, "</a>");
			return;
		}
	}
#endif /* EDB_READER */

	if(d && DatumAvail(d)) {
		register eid_t	eid;

		if(mode&EDB_PRINT_MODE_ANCHOR) doc_printf(doc, "<a name=\"%s-%d\">", ColumnSN(tc->ci), idx+1);
		doc_printf(doc, "%s", tag);
		edb_print_tuple_column_datum(ep, tuple, tc, d, mode);
		if((ec->operator!=0 || ec->university) && DatumEnglishIsUsable(d) && edb_table_column_is_studentcode(ec, tc->ci)) {
			HTML(doc) edb_print_student_list(ec, DatumEnglish(d), FALSE, "", "font-size:95%", 16, doc);
		}
#if 1
		if(mode&EDB_PRINT_MODE_CHECKMATCHED) {
			// マッチするEIDのチェック．
			if(tc->ci->l_maplookup && !DatumEIDisValid(d)) {
				EdbCatalogue	ca = edb_catalogue_new();
				Link	lm;

				ForLink(lm, tc->ci->l_maplookup) {
					EdbTableInfo	ti;
					eid_t	teid;

					teid = edb_maplookup_table_eid(ec, (EdbMapLookup)lm->d);
					ti = edb_table_get(ec, teid);
					if(DatumEnglishIsValid(d)) edb_print_tuple_check_matched(ec, ca, ti, tc->ci, (EdbMapLookup)lm->d, (char *)DatumEnglish(d), 0);
					if(DatumJapaneseIsValid(d)) edb_print_tuple_check_matched(ec, ca, ti, tc->ci, (EdbMapLookup)lm->d, (char *)DatumJapanese(d), 1);
					if(DatumPronounceIsValid(d)) edb_print_tuple_check_matched(ec, ca, ti, tc->ci, (EdbMapLookup)lm->d, (char *)DatumPronounce(d), 2);
				}
				if(edb_catalogue_size(ca)>0) {
					Link	l;
					int	person_is_exist = FALSE;

					doc_puts(doc, "<div class=\"cautionBox\">\n");
					doc_puts(doc, "<dl><dt>↑(データベース中に適合する可能性のある以下の情報を発見!!)</dt>");
					ForCatalogueLink(l, ca) {
						EdbTupleInfo	mtuple = CatalogueObject(l);
						Caption	cap = edb_make_caption(ec, TupleEID(mtuple), CAP_long);

						if(TupleIsPerson(mtuple)) person_is_exist = TRUE;
						doc_puts(doc, "<dd>");
						doc_text_puts3(doc, "【", (char *)TupleName_ja(mtuple), "】(");
						doc_printf(doc, "<a href=\"%s?EID=%d\"", ec->browser_url, TupleEID(mtuple));
						if(mode&EDB_PRINT_MODE_BLANKTARGET) doc_puts(doc, " target=\"_blank\"");
						doc_text_puts3(doc, ">", (char *)cap->c_main, "</a>");
						if(STRisVALID(cap->c_sub)) doc_text_puts3(doc, " <span style=\"font-size:90%;\">(", (char *)cap->c_sub, ")</span>");
						doc_puts(doc, ")</dd>\n");
						edb_caption_free(ec, cap);
					}
					doc_puts(doc, "</dl>\n");
					if(person_is_exist) {
						doc_color_begin(doc, EP_WARNING_COLOR);
						doc_text_puts(doc, "この情報が個人の業績等に分類されるためには参照登録が必要です．テキスト記載の個人が上記に掲げた個人本人ならば，参照登録を行ってください．");
						doc_color_end(doc);
					}
					doc_puts(doc, "</div>\n");
				}
				edb_catalogue_free(ca);
			}
			if(tc->ci->caption == EDB_CAPTION_SHORT && DatumRead(d)!=EDB_AUTH_INHERIT) {
				doc_puts(doc, "<div class=\"cautionBox\">\n");
//				doc_color_begin(doc, EP_WARNING_COLOR);
				doc_puts(doc, "↑(見出しに利用される項目が公開制限されています．この情報を参照する他の情報を公開したときに，正しく表示されない可能性があります．)");
//				doc_color_end(doc);
				doc_puts(doc, "</div>\n");
			}
		}
#if 1		// add by Alex @ Aug. 23, 2011
		if(!strcmp((char *)ColumnXN(tc->ci), "article.magazine.issn")
			|| !strcmp((char *)ColumnXN(tc->ci), "magazine.issn")
			|| !strcmp((char *)ColumnXN(tc->ci), "magazine.eissn")) {
			doc_prefix_set(doc, "<div class=\"infoBox\" style=\"font-size:80%;\">\n");
			if(!edb_print_issn_information(ep->ec, ep->doc, (char *)DatumEnglish(d))) {
				doc_prefix_reset(doc);
			} else {
				doc_puts(doc, "</div>\n");
			}
		} else if((!strcmp((char *)ColumnXN(tc->ci), "article.magazine") || !strcmp((char *)ColumnXN(tc->ci), "magazine.successor")) && DatumEIDisValid(d)) {
			char	*issn = NULL;
			if(!strcmp((char *)ColumnXN(tc->ci), "article.magazine")) {
				EdbTupleColumn issnTc = edb_tuple_seek_TC_by_xmlname(ec, tuple, tc, d, (utf8_t *)"@.issn");
				if(issnTc) {
					EdbDatum	issnDatum;
					ForTCDatum(issnDatum, issnTc) {
						if(DatumEnglishIsUsable(issnDatum)) {
							issn = (char *)DatumEnglish(issnDatum);
							break;
						}
					}
				}
			}
			doc_prefix_set(doc, "<div class=\"infoBox\" style=\"font-size:70%;\">\n");
			if(!edb_print_magazine_information(ep->ec, ep->doc, DatumEID(d), issn, FALSE)) {
				doc_prefix_reset(doc);
			} else {
				doc_puts(doc, "</div>\n");
			}
		}
#endif
#endif
		if(mode&EDB_PRINT_MODE_ANCHOR) doc_puts(doc, "</a>");
	} else {
		if(mode&EDB_PRINT_MODE_ANCHOR) doc_printf(doc, "<a name=\"%s-%d\">", ColumnSN(tc->ci), idx+1);
		doc_printf(doc, "%s", tag);
		if(mode&EDB_PRINT_MODE_ANCHOR) doc_puts(doc, "</a>");
	}
}

#define	MAPTO_COLOR	"teal"
#define	DIFF_COLOR		"#ff0000"

void
edb_print_tuple_column_head(EdbPrint ep)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;

	HTML(doc) doc_puts(doc, "<table border=\"0\" style=\"font-size:100%\" align=\"left\" style=\"text-align:left; display:inline\">\n");
}

void
edb_print_tuple_column_foot(EdbPrint ep)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;

	HTML(doc) doc_puts(doc, "</table>\n");
}

void
edb_print_tuple_column_information(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn super_tc, EdbDatum super_datum, Link l_column, int mode)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	EdbTupleColumn	tc;
	Link	l;
	int	exclusive = FALSE;
	char	*fgc = "black";

	if(ec==NULL || tuple==NULL || l_column==NULL || doc==NULL) return;

	ForLink(l, l_column) {
		EdbColumnInfo		ci = l->d;
		EdbTupleColumn	tc = edb_tuple_seek_TC_by_xmlname(ec, tuple, super_tc, super_datum, ColumnXN(ci));
		int	elements;
		int	e, num = 0;
		char	*cmodestr = "";
		int	usage = 0;

		if(exclusive && ci->exclusive) continue;
		if(ci->obsoleted && (tc==NULL || tc->datum==NULL)) continue;
		usage = edb_tuple_column_usage(ec, tuple, ci);
		if(usage==USAGE_UNUSED && (tc==NULL || tc->datum==NULL)) continue;

		fgc = (!TupleAvail(tuple)) ? "#808080" : "black";
		if(TupleMapto(tuple) && tc && tc->mapped) {
			fgc = MAPTO_COLOR;
		}
		HTML(doc) doc_printf(doc, "<tr valign=\"top\" style=\"text-align:left; color:%s\"><td><nobr>", fgc);
		switch(usage) {
		case USAGE_REQUIRED:		cmodestr = "(必須)";	break;
		case USAGE_PREFERRED:	cmodestr = "(推奨)";	break;
		case USAGE_OPTIONAL:		cmodestr = "(任意)";	break;
		case USAGE_UNUSED:		cmodestr = "(不用)";	break;
		}
		if(mode&EDB_PRINT_MODE_ANCHOR) {
			doc_printf(doc, "<a name=\"%s\">○<b>%s</b></a> <span style=\"font-size:90%%;\">%s</span>: ", ColumnXN(ci), ColumnName_ja(ci), cmodestr);
		} else {
			HTML(doc) doc_printf(doc, "○<b>%s</b> <span style=\"font-size:90%%;\">%s</span>: ", ColumnName_ja(ci), cmodestr);
			PLAIN(doc) doc_printf(doc, "%s: ", ColumnName_ja(ci));
		}
		HTML(doc) doc_puts(doc, "</nobr></td>");
		elements = edb_table_column_xml_elements(ec, ColumnTable(ci), ColumnSuper(ci), ci);
		if(tc) {
			num = elements>0 ? elements : edb_tcdatum_num(ec, tc);
			for(e=0; e<num; ++e) {
				EdbDatum	d = edb_tcdatum_get_datum(ec, tc, e);
				int	c;

				if(e>0) HTML(doc) {
					doc_puts(doc, "</td></tr>\n");
					doc_printf(doc, "<tr valign=\"top\" style=\"text-align:left; color:%s\"><td></td>", fgc);
				}
				if(elements!=1) {
					HTML(doc) doc_printf(doc, "<td align=\"right\">%d.</td>", e+1);
				} else {
					HTML(doc) doc_printf(doc, "<td align=\"center\">□</td>");
				}
				HTML(doc) doc_puts(doc, "<td style=\"text-align:left\">");
				TeX(doc) {
					char	*texcmd;

					texcmd = (char *)tex_column_to_sqlname(ec, ci);
					doc_printf(doc, "\\%s{", texcmd);
					FREE(texcmd);
				}
				edb_print_tuple_column_index(ep, tuple, tc, e, "", mode);
				HTML(doc) doc_puts(doc, "\n");
				PLAIN(doc) doc_puts(doc, "\n");
				TeX(doc) doc_puts(doc, "}\n");
				if(d && ci->l_child) {
					HTML(doc) doc_puts(doc, "<br />\n");
					edb_print_tuple_column_head(ep);
					edb_print_tuple_column_information(ep, tuple, tc, d, ci->l_child, mode);
					edb_print_tuple_column_foot(ep);
				}
				HTML(doc) {
					doc_printf(doc, "</td><td nowrap=\"nowrap\">");
					edb_print_tuple_datum_permission0(ep, tuple, tc, d);
				}
			}
			HTML(doc) doc_puts(doc, "</td>\n");
		} else if(mode&EDB_PRINT_MODE_CHECKMATCHED) {
			HTML(doc) {
				// 学籍番号のチェック．
				if(edb_table_column_is_studentcode(ec, ci) && super_datum) {
					EdbDatum sdatum = super_datum;
					Link	l_found = edb_table_search_students_byDatum(ec, sdatum, 0);
					if(l_found!=NULL) {
						Link	l;
						doc_printf(doc, "<td align=\"right\"></td></tr><tr>");
						doc_puts(doc, "<td colspan=\"4\" style=\"text-align:left\">");
						doc_puts(doc, "<div class=\"cautionBox\">\n");
						doc_puts(doc, "<dl><dt>↑(適合する可能性のある以下の学生情報を発見!!)</dt>");
#define	LISTING_STUDENT_LIMITs	16
						int	num = LISTING_STUDENT_LIMITs;
						ForLink(l, l_found) {
							if(num-->0) {
								doc_puts(doc, "<dd style=\"color:#008000;\">");
								edb_print_student_info(ec, (StudentInfo)l->d, FALSE, doc);
								doc_puts(doc, "</dd>\n");
							} else {
								doc_text_puts3(doc, "<dd style=\"color:#008000;\">", "---(他)---", "</dd>\n");
								break;
							}
						}
						doc_puts(doc, "</dl>\n");
						doc_color_begin(doc, EP_WARNING_COLOR);
						doc_text_puts(doc, "この情報が学生所属組織の業績等に分類されるためには学籍番号の登録が必要です．テキスト記載の個人が上記に掲げた個人本人ならば，学籍番号の登録を行ってください．（編集画面では学籍番号が表示されます）");
						doc_color_end(doc);
						doc_puts(doc, "</div>\n");
						link_free(l_found);
						doc_printf(doc, "</td>\n");
					}
				}
			}
		}
		if(!exclusive && ci->exclusive && num>0) exclusive = TRUE;
		if(num==0) doc_puts(doc, "\n");
		HTML(doc) doc_puts(doc, "</tr>\n");
	}
}

static
char *
edb_print_tuple_tc_make_xml_for_diff(EdbContext ec, EdbTupleInfo tuple, EdbTupleColumn tc, int completion)
{
	EdbTableInfo	ti = TupleTable(tuple);
	Link	l, l_xml;
	char	*xml = NULL;
	EdbColumnInfo		ci;
	register EdbDatum	datum;

	l_xml = NULL;
	if(tc==NULL) return NULL;
	ci = tc->ci;
	if(!completion && tc->mapped) return NULL;
#if 1
	if(TCDatumFirst(tc)==NULL && TCDatumNothing(tc)) {
		asprintf((char **)&xml, "<%s:%s %s/>", EDB_XMLNS, ColumnXN(ci), XML_AN_ASSERT "=\"" XML_AV_ASSERT_NOTHING "\"");
		return xml;
	}
#endif
	ForTCDatum(datum, tc) {
		char	*datum_xml=NULL, *child_xml=NULL, *m_attr = NULL, *p_attr = NULL;
		Caption	cap;

		xml = NULL;
		if(datum->l_child) child_xml = (char *)edb_tuple_datum_make_xml(ec, tuple, datum->l_child, completion);
		if(completion && DatumEIDisValid(datum) && ci->typeInfo->multilingual && (cap = edb_make_caption(ec, DatumEID(datum), CAP_norecurse))) {
			EdbDatum	d = edb_datum_dup(ec, NULL, datum);

#if 0
			FREE(d->d_english);
			FREE(d->d_japanese);
			FREE(d->d_pronounce);
			d->d_english = strdup_null(cap->c_main_english);
			d->d_japanese = strdup_null(cap->c_main_japanese);
			d->d_pronounce = strdup_null(cap->c_main_pronounce);
			if(ci->typeInfo) datum_xml = (char *)edb_datum_make_xml(ec, ColumnCType(ci), d, &m_attr, &p_attr, completion);
			edb_datum_free(ec, d);
#else
			edb_datum_set(ec, d, DatumEID(d), cap->c_main_english, cap->c_main_japanese, cap->c_main_pronounce);
			if(ci->typeInfo) datum_xml = (char *)edb_datum_make_xml(ec, ColumnCType(ci), d, (utf8_t **)&m_attr, (utf8_t **)&p_attr, completion);
			edb_datum_free(ec, d);
#endif
			edb_caption_free(ec, cap);
		} else {
			if(ci->typeInfo) datum_xml = (char *)edb_datum_make_xml(ec, ColumnCType(ci), datum, (utf8_t **)&m_attr, (utf8_t **)&p_attr, completion);
		}

		if(child_xml==NULL && datum_xml==NULL) {
			if(m_attr) asprintf(&xml, "<%s:%s %s%s%s/>", EDB_XMLNS, ColumnXN(ci), m_attr, p_attr?" ":"", p_attr ? p_attr : "");
		} else {
			asprintf(&xml, "<%s:%s%s%s%s%s>%s%s</%s:%s>", EDB_XMLNS, ColumnXN(ci),
				m_attr?" ":"",
				m_attr ? m_attr : "",
				p_attr?" ":"",
				p_attr ? p_attr : "",
				datum_xml ? datum_xml : "",
				child_xml ? child_xml : "",
				EDB_XMLNS, ColumnXN(ci));
		}
		FREE(child_xml);
		FREE(datum_xml);
		FREE(m_attr);
		FREE(p_attr);
		if(xml) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	xml = str_concatenate0(l_xml);
	link_free(l_xml);
	if(STRisVALID(xml)) return xml;
	FREE(xml);
	return NULL;
}

static
int
edb_print_tuple_column_diff_compare(EdbContext ec, EdbTupleInfo new, EdbTupleColumn new_tc, EdbTupleInfo old, EdbTupleColumn old_tc)
{
	char	*new_xml, *old_xml;
	int	st;

	new_xml = edb_print_tuple_tc_make_xml_for_diff(ec, new, new_tc, FALSE);
	old_xml  = edb_print_tuple_tc_make_xml_for_diff(ec, old, old_tc, FALSE);

	if(new_xml==NULL && old_xml==NULL) {
		st = 0;
	} else if(new_xml==NULL) {
		st = -1;
	} else if(old_xml==NULL) {
		st = 1;
	} else {
		st = strcmp(new_xml, old_xml);
	}
	FREE(new_xml);
	FREE(old_xml);
	return st;
}

static
void
edb_print_tuple_column_diff_information(EdbPrint ep, EdbTupleInfo new, EdbTupleInfo old, Link l_column, int mode)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	EdbTupleColumn	tc;
	Link	l;
	int	exclusive = FALSE;
	int	differ;
	char	*fgc = "black";

	if(ec==NULL || new==NULL || l_column==NULL || doc==NULL) return;

	ForLink(l, l_column) {
		EdbColumnInfo		ci = l->d;
		EdbTupleColumn	new_tc = edb_tuple_seek_TC_by_xmlname(ec, new, NULL, NULL, ColumnXN(ci));
		EdbTupleColumn	old_tc = edb_tuple_seek_TC_by_xmlname(ec, old, NULL, NULL, ColumnXN(ci));
		int	elements;
		int	e, num = 0;

		if(exclusive && ci->exclusive) continue;
		if(ci->obsoleted && new_tc && new_tc->datum==NULL) continue;
		differ = edb_print_tuple_column_diff_compare(ec, new, new_tc, old, old_tc);
		fgc = (!TupleAvail(new)) ? "#808080" : "black";
		if(differ) {
			fgc = DIFF_COLOR;
		} else if(TupleMapto(new) && new_tc && new_tc->mapped) {
			fgc = MAPTO_COLOR;
		}
		HTML(doc) doc_printf(doc, "<tr valign=\"top\" style=\"text-align:left;color:%s\"><td><nobr>", fgc);
		if(mode&EDB_PRINT_MODE_ANCHOR) {
			doc_printf(doc, "<a name=\"%s\">○<b>%s</b></a> : \n", ColumnXN(ci), ColumnName_ja(ci));
		} else {
			HTML(doc) doc_printf(doc, "○<b>%s</b> : \n", ColumnName_ja(ci));
			PLAIN(doc) doc_printf(doc, "%s: ", ColumnName_ja(ci));
		}
		HTML(doc) doc_puts(doc, "</nobr></td>");
		if(new_tc) {
			if(TCDatumFirst(new_tc)==NULL && TCDatumNothing(new_tc)) {
				HTML(doc) doc_printf(doc, "<td align=\"center\">□</td><td style=\"font-size:90%;\">(該当無し)</td><td nowrap=\"nowrap\">");
			} else {
				elements = edb_table_column_xml_elements(ec, ColumnTable(ci), ColumnSuper(ci), ci);
				num = elements>0 ? elements : edb_tcdatum_num(ec, new_tc);
				for(e=0; e<num; ++e) {
					EdbDatum	d = edb_tcdatum_get_datum(ec, new_tc, e);
					int	c;

					if(e>0) HTML(doc) doc_printf(doc, "</td></tr>\n<tr valign=\"top\" style=\"text-align:left;color:%s\"><td></td>", fgc);
					if(elements!=1) {
						HTML(doc) doc_printf(doc, "<td align=\"right\">%d.</td><td>", e+1);
					} else {
						HTML(doc) doc_printf(doc, "<td align=\"center\">□</td><td>");
					}
					TeX(doc) {
						char	*texcmd;

						texcmd = (char *)tex_column_to_sqlname(ec, ci);
						doc_printf(doc, "\\%s{", texcmd);
						FREE(texcmd);
					}
					edb_print_tuple_column_index(ep, new, new_tc, e, "", mode);
					HTML(doc) doc_puts(doc, "\n");
					PLAIN(doc) doc_puts(doc, "\n");
					TeX(doc) doc_puts(doc, "}\n");
					if(d && d->l_child) {
						HTML(doc) doc_puts(doc, "<br />\n");
						edb_print_tuple_column_head(ep);
						edb_print_tuple_column_information(ep, new, new_tc, d, ci->l_child, mode);
						edb_print_tuple_column_foot(ep);
					}
					HTML(doc) {
						doc_printf(doc, "</td><td nowrap=\"nowrap\">");
						edb_print_tuple_datum_permission0(ep, new, new_tc, d);
					}
				}
			}
		}
		if(!exclusive && ci->exclusive && num>0) exclusive = TRUE;
		if(num==0) doc_puts(doc, "\n");
		HTML(doc) doc_puts(doc, "</td></tr>\n");
	}
}

static
void
edb_print_tuple_column_diff_column(EdbPrint ep, EdbTupleInfo new, EdbTupleInfo old, Link l_column, int mode)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	EdbTupleColumn	tc;
	Link	l;
	int	exclusive = FALSE;
	int	differ;
	char	*fgc = "black";

	if(ec==NULL || new==NULL || l_column==NULL || doc==NULL) return;

	ForLink(l, l_column) {
		EdbColumnInfo		ci = l->d;
		EdbTupleColumn	new_tc = edb_tuple_seek_TC_by_xmlname(ec, new, NULL, NULL, ColumnXN(ci));
		EdbTupleColumn	old_tc = edb_tuple_seek_TC_by_xmlname(ec, old, NULL, NULL, ColumnXN(ci));
		int	elements;
		int	e, num = 0;

		if(exclusive && ci->exclusive) continue;
		if(ci->obsoleted && new_tc && new_tc->datum==NULL) continue;
		differ = edb_print_tuple_column_diff_compare(ec, new, new_tc, old, old_tc);
		fgc = (!TupleAvail(new)) ? "#808080" : "black";
		if(differ) {
			fgc = DIFF_COLOR;
		} else if(TupleMapto(new) && new_tc && new_tc->mapped) {
			fgc = MAPTO_COLOR;
		}
		if(differ) {
			doc_printf(doc, "<b style=\"color:%s;\">[%s]</b> \n", fgc, ColumnName_ja(ci));
		} else {
			doc_printf(doc, "<span style=\"color:%s;\">[%s]</span> \n", fgc, ColumnName_ja(ci));
		}
		if(!exclusive && ci->exclusive && num>0) exclusive = TRUE;
	}
}

void
edb_print_tuple_information(EdbPrint ep, EdbTupleInfo tuple, int mode)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;

	if(tuple == NULL || doc==NULL) return;

	HTML(doc) {
		doc_puts(doc, "<table border=\"5\" width=\"100%\" style=\"");
		if(!TupleAvail(tuple)) doc_puts(doc, "color:#808080;");
		doc_puts(doc, "\">\n");
	} else {
		if(!TupleAvail(tuple)) doc_color_begin(doc, "#808080");
	}
	TeX(doc) doc_printf(doc, "\\begin{%s}\n", TupleName_en(tuple));
	HTML(doc) {
		doc_puts(doc, "<tr><td>");
		doc_printf(doc, "EID=<span title=\"" NAME_OF_DB "におけるこの情報の識別子\" style=\"font-weight:bold;font-size:x-large\">%d</span>", TupleEID(tuple));
		doc_puts(doc, "</td><td style=\"font-size:90%;\">");
		edb_base_print_information(ec, TupleBase(tuple), mode, doc);
		doc_puts(doc, "</td></tr><tr><td colspan=\"2\" style=\"text-align:left\">\n");
	} else {
		edb_base_print_information(ec, TupleBase(tuple), mode, doc);
	}
	edb_print_tuple_column_head(ep);
	edb_print_tuple_column_information(ep, tuple, NULL, NULL, TupleTable(tuple)->l_column, mode);
	edb_print_tuple_column_foot(ep);
	TeX(doc) doc_printf(doc, "\\end{%s}\n", TupleName_en(tuple));
	if(TupleMapto(tuple)) {
		HTML(doc) doc_puts(doc, "</td></tr><tr><td colspan=\"2\">\n");
		doc_color_begin(doc, MAPTO_COLOR);
		doc_puts(doc, "<p>●この色で表示されている項目はマップによって参照された内容です．</p>\n");
		doc_color_end(doc);
	}
	HTML(doc) {
		doc_puts(doc, "\n");
		doc_puts(doc, "</td></tr></table>\n");
	}
}

#if 0
static
void
edb_tuple_xml_indent(EdbContext ec, int indent, Doc doc)
{
	int	i;

	for(i=0; i<indent; ++i) doc_puts(doc, "&nbsp;&nbsp;&nbsp;&nbsp;");
}

static
void
edb_tuple_xml_begin(EdbContext ec, EdbTupleInfo tuple, int indent, Doc doc)
{

	edb_tuple_xml_indent(ec, indent, doc);
	doc_text_printf(doc, "<?xml version=\"%s\" encoding=\"utf-8\"?>\n", XML_DEFAULT_VERSION);
	doc_text_printf(doc, "<!DOCTYPE edb:%s SYSTEM '%s%s.dtd'>\n", TupleXN(tuple), EDB_DTD_URI, TupleXN(tuple));
	doc_text_printf(doc, "\n");
	doc_text_printf(doc, "<%s:%s xmlns:%s=\"%s\">\n", EDB_XMLNS, TupleXN(tuple), EDB_XMLNS, EDB_XMLNS_URI);
}

static
void
edb_tuple_xml_end(EdbContext ec, EdbTupleInfo tuple, int indent, Doc doc)
{

	edb_tuple_xml_indent(ec, indent, doc);
	doc_text_printf(doc, "</%s:%s>\n", EDB_XMLNS, TupleXN(tuple));
}

static
void
edb_tuple_column_xml_begin(EdbContext ec, EdbTupleColumn tc, int indent, Doc doc)
{

	edb_tuple_xml_indent(ec, indent, doc);
	doc_text_printf(doc, "<%s:%s>\n", EDB_XMLNS, ColumnXN(tc->ci));
}

static
void
edb_tuple_column_xml_end(EdbContext ec, EdbTupleColumn tc, int indent, Doc doc)
{

	edb_tuple_xml_indent(ec, indent, doc);
	doc_text_printf(doc, "</%s:%s>\n", EDB_XMLNS, ColumnXN(tc->ci));
}

static
void
edb_tuple_xml_print_one_element(EdbContext ec, char *xmlname, char *value, int indent, Doc doc)
{

	edb_tuple_xml_indent(ec, indent, doc);
	doc_text_printf(doc, "<%s:%s>%s</%s:%s>\n", EDB_XMLNS, xmlname, value, EDB_XMLNS, xmlname);
}

static	void	edb_print_tuple_column_in_xml		__P((EdbContext, EdbTupleInfo, Link, int, Doc));
static
void
edb_print_tuple_column_datum_in_xml(EdbContext ec, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum d, int indent, Doc doc)
{
	int	i;
	char	*s1, *s2;

	if(ec==NULL || tuple == NULL || tc == NULL || doc==NULL || d==NULL) return;
	if(!edb_auth_readable_TCD(ec, tuple, tc, d)) {
		edb_tuple_xml_print_one_element(ec, ColumnXN(tc->ci), "****", indent, doc);
		return;
	}
	if(DatumEIDisValid(d)) {
		edb_tuple_xml_indent(ec, indent, doc);
		if(d->l_child) {
			doc_text_printf(doc, "<%s:%s %s=\"%d\">\n", EDB_XMLNS, ColumnXN(tc->ci), XML_AN_MAPTO, DatumEID(d));
			edb_print_tuple_column_in_xml(ec, tuple, d->l_child, indent+1, doc);
			edb_tuple_column_xml_end(ec, tc, indent, doc);
		} else {
			doc_text_printf(doc, "<%s:%s %s=\"%d\"/>\n", EDB_XMLNS, ColumnXN(tc->ci), XML_AN_MAPTO, DatumEID(d));
		}
	} else if(DatumEnglishIsValid(d) || DatumJapaneseIsValid(d) || DatumPronounceIsValid(d)) {
		edb_tuple_column_xml_begin(ec, tc, indent, doc);
		switch(tc->ci->typeInfo->basicTypeID) {
		case EDB_TYPE_NAME:
		case EDB_TYPE_SENTENCE:
		case EDB_TYPE_PARAGRAPH:
			if(DatumEnglishIsValid(d)) edb_tuple_xml_print_one_element(ec, XML_EN_ENGLISH, DatumEnglish(d), indent+1, doc);
			if(DatumJapaneseIsValid(d)) edb_tuple_xml_print_one_element(ec, XML_EN_JAPANESE, DatumJapanese(d), indent+1, doc);
			if(DatumPronounceIsValid(d)) edb_tuple_xml_print_one_element(ec, XML_EN_PRONOUNCE, DatumPronounce(d), indent+1, doc);
			break;
		case EDB_TYPE_DATE2:
		case EDB_TYPE_EMAIL:
		case EDB_TYPE_DATE:
		case EDB_TYPE_REAL:
		case EDB_TYPE_URL:
		case EDB_TYPE_PHONE:
		default:
			edb_tuple_xml_print_one_element(ec, XML_EN_ENGLISH, DatumEnglish(d), indent+1, doc);
			break;
		}
		if(d->l_child) edb_print_tuple_column_in_xml(ec, tuple, d->l_child, indent+1, doc);
		edb_tuple_column_xml_end(ec, tc, indent, doc);
	}
}

static
void
edb_print_tuple_column_in_xml(EdbContext ec, EdbTupleInfo tuple, Link l_column, int indent, Doc doc)
{
	EdbTupleColumn	tc;
	Link	l;
	int	exclusive = FALSE;

	if(ec==NULL || tuple==NULL || l_column==NULL || doc==NULL) return;

	ForLink(l, l_column) {
		EdbTupleColumn	tc = (EdbTupleColumn)l->d;
		EdbColumnInfo		ci = tc->ci;
		int				elements;
		int	e, num = 0;

		if(exclusive && ci->exclusive) continue;
		num = edb_tcdatum_num(ec, tc);
		elements = edb_table_column_xml_elements(ec, ColumnTable(ci), ColumnSuper(ci), ci);
		if(elements>0 && num>elements) num = elements;
		for(e=0; e<num; ++e) {
			EdbDatum	d = edb_tcdatum_get_datum(ec, tc, e);

			if(d==NULL) break;
			edb_print_tuple_column_datum_in_xml(ec, tuple, tc, d, indent, doc);
			if(!exclusive && ci->exclusive) exclusive = TRUE;
		}
	}
}
#endif

void
edb_print_tuple_in_xml(EdbPrint ep, EdbTupleInfo tuple, int mode)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
#if 1
	char *xml;
	xmlChar	*xmlout;
	xmlDocPtr	xmldoc;
	int	len;
	Link	l, lstr;

	if(tuple == NULL || doc==NULL) return;
	xml = (char *)edb_tuple_make_xml(ec, tuple, (mode&EDB_PRINT_MODE_COMPLETION)!=0);

	xmldoc = xmlParseMemory(xml, xmlStrlen((xmlChar *)xml));
	if(xmldoc==NULL) return;
#if 1
	xmlDocDumpFormatMemoryEnc(xmldoc, &xmlout, &len, "utf-8", 1);
#else
	xmlDocDumpMemoryEnc(xmldoc, &xmlout, &len, "utf-8");
#endif
	lstr = str_split((char *)xmlout, '\n');
	ForLink(l, lstr) {
		char *s = (char *)l->d;

		while(*s==' ') {
			HTML(doc) {
				doc_puts(doc, "&nbsp;&nbsp;&nbsp;&nbsp;");
			} else {
				doc_puts(doc, "    ");
			}
			++s;
		}
#if 0
		doc_text_puts(doc, s);
#else
		HTML(doc) {
			s = (char *)xml_text_encode(ec, (xmlChar *)s);
			doc_puts(doc, s);
			FREE(s);
			doc_puts(doc, "<br />\n");
		} else {
			doc_puts(doc, s);
			doc_puts(doc, "\n");
		}
#endif
	}
	link_free(lstr);

	xmlFreeDoc(xmldoc);
	FREE(xml);
	FREE(xmlout);
#else
	Link	l;

	if(tuple == NULL || doc==NULL) return;
	if(!TupleAvail(tuple)) doc_color_begin(doc, "#808080");
	edb_tuple_xml_begin(ec, tuple, 0, doc);
	edb_tuple_xml_indent(ec, 1, doc);
	edb_base_print_in_xml(ec, TupleBase(tuple), FALSE, doc);
	doc_text_puts(doc, "\n");
	edb_print_tuple_column_in_xml(ec, tuple, tuple->l_column, 1, doc);
	edb_tuple_xml_end(ec, tuple, 0, doc);
	if(!TupleAvail(tuple)) doc_color_end(doc);
#endif
}

void
edb_print_tuple_diff_information(EdbPrint ep, EdbTupleInfo tuple, EdbTupleInfo old, int mode)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	int	c, c1, diff;

	if(tuple == NULL || old == NULL || doc==NULL) return;
	HTML(doc) doc_puts(doc, "<table border=\"3\" width=\"100%\"><tr><td>\n");
	if(!TupleAvail(tuple)) doc_color_begin(doc, "#808080");
	TeX(doc) doc_printf(doc, "\\begin{%s}\n", TupleName_en(tuple));

	diff = edb_base_compare(ec, TupleBase(tuple), TupleBase(old));
	if(diff) doc_color_begin(doc, DIFF_COLOR);
	edb_base_print_information(ec, TupleBase(tuple), mode, doc);
	if(diff) doc_color_end(doc);

	HTML(doc) doc_puts(doc, "</td></tr><tr><td>\n");
	edb_print_tuple_column_head(ep);
	edb_print_tuple_column_diff_information(ep, tuple, old, TupleTable(tuple)->l_column, mode);
	edb_print_tuple_column_foot(ep);
	TeX(doc) doc_printf(doc, "\\end{%s}\n", TupleName_en(tuple));
	if(TupleMapto(tuple)) {
		HTML(doc) doc_puts(doc, "</td></tr><tr><td>\n");
		doc_color_begin(doc, MAPTO_COLOR);
		doc_puts(doc, "<p>●この色で表示されている項目はマップによって参照された内容です．</p>\n");
		doc_color_end(doc);
	}
	if(!TupleAvail(tuple)) doc_color_end(doc);
	HTML(doc) doc_puts(doc, "</td></tr></table>\n");
	doc_puts(doc, "<p>\n");
	doc_color_begin(doc, DIFF_COLOR);
	doc_puts(doc, "●この色で表示されている項目が変更されています．\n");
	doc_color_end(doc);
	doc_puts(doc, "</p>\n");
}

void
edb_print_tuple_diff_column(EdbPrint ep, EdbTupleInfo tuple, EdbTupleInfo old, int mode)
{
	EdbContext	ec = ep->ec;
	Doc			doc = ep->doc;
	int	c, c1, diff;

	if(tuple == NULL || doc==NULL) return;
	HTML(doc) doc_puts(doc, "<table border=\"3\" width=\"100%\"><tr><td colspan=\"2\">\n");
	if(!TupleAvail(tuple)) doc_color_begin(doc, "#808080");
	TeX(doc) doc_printf(doc, "\\begin{%s}\n", TupleName_en(tuple));

	diff = FALSE;
	if(old) diff = edb_base_compare(ec, TupleBase(tuple), TupleBase(old));

	if(diff) doc_color_begin(doc, DIFF_COLOR);
	edb_base_print_information(ec, TupleBase(tuple), mode, doc);
	if(diff) doc_color_end(doc);

	if(old) {
		HTML(doc) {
			doc_puts(doc, "</td></tr><tr><td width=\"2%\" nowrap=\"nowrap\" style=\"vertical-align:top;\">\n");
			doc_text_puts(doc, "変更項目");
			doc_puts(doc, "</td><td>\n");
		}
		edb_print_tuple_column_diff_column(ep, tuple, old, TupleTable(tuple)->l_column, mode);
	}
	if(!TupleAvail(tuple)) doc_color_end(doc);
	HTML(doc) doc_puts(doc, "</td></tr></table>\n");
}

static
int
edb_print_duplicate_print_tuple_info_func(BiTree_obj_t obj, void *data)
{
	register struct EDB_PRINT_TUPLE_INFO	*src = obj;
	register BiTree						*new_bt = data;
	register struct EDB_PRINT_TUPLE_INFO	*dst;

	dst = malloc(sizeof(struct EDB_PRINT_TUPLE_INFO));
	bcopy(src, dst, sizeof(struct EDB_PRINT_TUPLE_INFO));
	dst->bt_print_tcd_info = edb_print_tcd_duplicate_info(src->bt_print_tcd_info);
#if 0
	BiTreeAdd(*new_bt, dst, TABLE_EID(dst->ti), NULL, NULL, NULL);
#else
	BiTreeAdd(*new_bt, dst, dst->teid, NULL, NULL, NULL);
#endif
	return TRUE;
}

BiTree
edb_print_tuple_duplicate_info(BiTree bt)
{
	BiTree	new = NULL;

	bitree_callback(bt, edb_print_duplicate_print_tuple_info_func, &new);
	return new;
}

int
edb_print_tuple_register_func(EdbPrint ep, char *xmlname, int style, EdbPrintTupleFunc_t print_tuple)
{
#if 0 
	struct EDB_PRINT_TUPLE_INFO *info;
	EdbTableInfo	ti = edb_table_get_by_name(ep->ec, xmlname);

	if(style<0 || EP_PrintStyle_Variations<=style) return FALSE;
	if(ti==NULL) return FALSE;

	if(!ep->ct->print_tuple_tobe_free) {
		BiTree	bt;

		bt = ep->ct->bt_print_tuple_info;
		ep->ct->bt_print_tuple_info = edb_print_tuple_duplicate_info(bt);
		ep->ct->print_tuple_tobe_free = TRUE;
	}
	info = bitree_search(ep->ct->bt_print_tuple_info, NULL, TABLE_EID(ti), NULL);
	if(info==NULL) {
		info = malloc(sizeof(struct EDB_PRINT_TUPLE_INFO));
		if(info==NULL) return FALSE;
		bzero(info, sizeof(struct EDB_PRINT_TUPLE_INFO));
		info->ti = ti;
		BiTreeAdd(ep->ct->bt_print_tuple_info, info, TABLE_EID(info->ti), NULL, NULL, NULL);
	}
	info->func[style] = print_tuple;
#else
	struct EDB_PRINT_TUPLE_INFO *info;
	eid_t		teid = edb_table_xmlname_to_eid(ep->ec, (utf8_t *)xmlname);

	if(style<0 || EP_PrintStyle_Variations<=style) return FALSE;
	if(!EIDisVALID(teid)) return FALSE;

	if(!ep->ct->print_tuple_tobe_free) {
		BiTree	bt;

		bt = ep->ct->bt_print_tuple_info;
		ep->ct->bt_print_tuple_info = edb_print_tuple_duplicate_info(bt);
		ep->ct->print_tuple_tobe_free = TRUE;
	}
	info = bitree_search(ep->ct->bt_print_tuple_info, NULL, teid, NULL);
	if(info==NULL) {
		info = malloc(sizeof(struct EDB_PRINT_TUPLE_INFO));
		if(info==NULL) return FALSE;
		bzero(info, sizeof(struct EDB_PRINT_TUPLE_INFO));
		info->teid = teid;
		BiTreeAdd(ep->ct->bt_print_tuple_info, info, info->teid, NULL, NULL, NULL);
	}
	info->func[style] = print_tuple;
#endif
	return TRUE;
}

int
edb_print_tuple_register_language_func(EdbPrint ep, char *xmlname, EdbPrintTupleLangFunc_t lang)
{
#if 0
	struct EDB_PRINT_TUPLE_INFO *info;
	EdbTableInfo	ti = edb_table_get_by_name(ep->ec, (utf8_t *)xmlname);

	if(ti==NULL) return FALSE;

	if(!ep->ct->print_tuple_tobe_free) {
		BiTree	bt;

		bt = ep->ct->bt_print_tuple_info;
		ep->ct->bt_print_tuple_info = edb_print_tuple_duplicate_info(bt);
		ep->ct->print_tuple_tobe_free = TRUE;
	}
	info = bitree_search(ep->ct->bt_print_tuple_info, NULL, TABLE_EID(ti), NULL);
	if(info==NULL) {
		info = malloc(sizeof(struct EDB_PRINT_TUPLE_INFO));
		if(info==NULL) return FALSE;
		bzero(info, sizeof(struct EDB_PRINT_TUPLE_INFO));
		info->ti = ti;
		BiTreeAdd(ep->ct->bt_print_tuple_info, info, TABLE_EID(info->ti), NULL, NULL, NULL);
	}
	info->lang = lang;
#else
	struct EDB_PRINT_TUPLE_INFO *info;
	eid_t		teid = edb_table_xmlname_to_eid(ep->ec, (utf8_t *)xmlname);

	if(!EIDisVALID(teid)) return FALSE;

	if(!ep->ct->print_tuple_tobe_free) {
		BiTree	bt;

		bt = ep->ct->bt_print_tuple_info;
		ep->ct->bt_print_tuple_info = edb_print_tuple_duplicate_info(bt);
		ep->ct->print_tuple_tobe_free = TRUE;
	}
	info = bitree_search(ep->ct->bt_print_tuple_info, NULL, teid, NULL);
	if(info==NULL) {
		info = malloc(sizeof(struct EDB_PRINT_TUPLE_INFO));
		if(info==NULL) return FALSE;
		bzero(info, sizeof(struct EDB_PRINT_TUPLE_INFO));
		info->teid = teid;
		BiTreeAdd(ep->ct->bt_print_tuple_info, info, info->teid, NULL, NULL, NULL);
	}
	info->lang = lang;
#endif
	return TRUE;
}


