/*
	$Id: print_type.c,v 1.22 2008/08/09 09:22:20 alex Exp $
	print_type.c--- print type
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: print_type.c,v 1.22 2008/08/09 09:22:20 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	<ctype.h>
#include	"edb/edb.h"
#include	"edb/tex.h"
#include	"edb/publish.h"

static
int
edb_print_type_default(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	register int	status = FALSE;
	register int	japanese	= LangIsJapanese(ep);

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	if(japanese) {
		if(DatumJapaneseIsUsable(datum)) {
			doc_text_puts(doc, (char *)DatumJapanese(datum));
			status = TRUE;
		} else if(DatumEnglishIsUsable(datum)) {
			doc_text_puts(doc, (char *)DatumEnglish(datum));
			status = TRUE;
		} else {
			status = edb_print_alert_column(ep, 0, tc->ci);
		}
	} else {	/* English */
		if(DatumEnglishIsUsable(datum)) {
			doc_text_puts(doc, (char *)DatumEnglish(datum));
			status = TRUE;
		} else if(DatumJapaneseIsUsable(datum)) {
			if(str_isascii((char *)DatumJapanese(datum))) {
				doc_text_puts(doc, (char *)DatumJapanese(datum));
				status = TRUE;
			} else {
				if(!ep->ct->nocolor) doc_color_begin(ep->doc, WARNING_COLOR);
				if(ep->ct->persist_lang && DocIsTeX(doc)) doc_puts(doc, "\\EdbAlert{");
				doc_text_puts(doc, (char *)DatumJapanese(datum));
				if(ep->ct->persist_lang && DocIsTeX(doc)) doc_puts(doc, "}");
				if(!ep->ct->nocolor) doc_color_end(ep->doc);
				status = TRUE;
			}
		} else {
			status = edb_print_alert_column(ep, 0, tc->ci);
		}
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

static
int
edb_print_type_TEXT(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	register int	status = FALSE;

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	if(DatumEnglishIsUsable(datum)) {
		doc_text_puts(doc, (char *)DatumEnglish(datum));
		status = TRUE;
	} else {
		status = edb_print_alert_column(ep, 0, tc->ci);
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

static
int
edb_print_type_NAME(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	register int	status = FALSE;
	register int	japanese	= LangIsJapanese(ep);

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	ITEMIZE(ep) {
		if(japanese) {
			if(DatumJapaneseIsUsable(datum)) {
				doc_text_puts(doc, (char *)DatumJapanese(datum));
				status = TRUE;
			} else {
				status = edb_print_alert_column(ep, 0, tc->ci);
			}
		} else {	/* English */
			if(DatumEnglishIsUsable(datum)) {
				doc_text_puts(doc, (char *)DatumEnglish(datum));
				status = TRUE;
			} else {
				status = edb_print_alert_column(ep, 0, tc->ci);
			}
		}
	} else {
		if(japanese) {
			if(DatumJapaneseIsUsable(datum)) {
				doc_text_puts(doc, (char *)DatumJapanese(datum));
				status = TRUE;
			} else if(DatumEnglishIsUsable(datum)) {
				doc_text_puts(doc, (char *)DatumEnglish(datum));
				status = TRUE;
			} else {
				status = edb_print_alert_column(ep, 0, tc->ci);
			}
		} else {	/* English */
			if(DatumEnglishIsUsable(datum)) {
				doc_text_puts(doc, (char *)DatumEnglish(datum));
				status = TRUE;
			} else if(DatumJapaneseIsUsable(datum)) {
				if(str_isascii((char *)DatumJapanese(datum))) {
					doc_text_puts(doc, (char *)DatumJapanese(datum));
					status = TRUE;
				} else {
					if(!ep->ct->nocolor) doc_color_begin(ep->doc, WARNING_COLOR);
					if(ep->ct->persist_lang && DocIsTeX(doc)) doc_puts(doc, "\\EdbAlert{");
					doc_text_puts(doc, (char *)DatumJapanese(datum));
					if(ep->ct->persist_lang && DocIsTeX(doc)) doc_puts(doc, "}");
					if(!ep->ct->nocolor) doc_color_end(ep->doc);
					status = TRUE;
				}
			} else {
				status = edb_print_alert_column(ep, 0, tc->ci);
			}
		}
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

static
int
edb_print_type_MONETARY(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	register int	status = FALSE;

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	LaTeX(ep) {
		if(DatumEnglishIsValid(datum)) {
			doc_printf(doc, "\\EdbMonetary{%s}", (char *)DatumEnglish(datum));
			status = TRUE;
		}
	} else if(DocIsXML(doc)) {
		if(DatumEnglishIsValid(datum)) {
			doc_printf(doc, "%s", (char *)DatumEnglish(datum));
			status = TRUE;
		}
	} else if(DatumEnglishIsUsable(datum)) {
		doc_text_puts(doc, (char *)DatumEnglish(datum));
		status = TRUE;
	} else {
		status = edb_print_alert_column(ep, 0, tc->ci);
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

#define	DATE__Y		0
#define	DATE__YM	1
#define	DATE__YMD	2

static
int
edb_print_type_DATE_print_ymd(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum, int ymd)
{
	register Doc	doc = ep->doc;
	static char	*mname[] = {
	"Jan.",	"Feb.",	"March",	"April",	"May",	"June",
	"July",	"Aug.",	"Sep.",	"Oct.",	"Nov.",	"Dec."
	};
	int	japanese = LangIsJapanese(ep);

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	if(datum && DatumAvail(datum)) {
		char	*mn;
		int32_t	date1;

		date1 = ATOI((char *)DatumEnglish(datum));
		if(0<DATE_M(date1) && DATE_M(date1)<=12) {
			mn = mname[DATE_M(date1)-1];
		} else if(DATE_M(date1)<1) {
			mn = mname[0];
		} else {
			mn = mname[11];
		}
		LaTeX(ep) {
			doc_printf(doc, "\\EdbDate{%d}", date1);
		} else if(DocIsXML(doc)) {
			doc_printf(doc, "%08d", date1);
		} else {
#if 0
			switch(ymd) {
			case DATE__YMD:
				if(japanese) {
					doc_printf(doc, "%d年%d月%d日", DATE_Y(date1), DATE_M(date1), DATE_D(date1));
				} else {
					TeX(doc) {
						doc_printf(doc, "%s\\ %d, %d", mn, DATE_D(date1), DATE_Y(date1));
					} else {
						doc_printf(doc, "%s %d, %d", mn, DATE_D(date1), DATE_Y(date1));
					}
				}
				break;
			case DATE__YM:
				if(japanese) {
					doc_printf(doc, "%d年%d月", DATE_Y(date1), DATE_M(date1));
				} else {
					TeX(doc) {
						doc_printf(doc, "%s\\ %d", mn, DATE_Y(date1));
					} else {
						doc_printf(doc, "%s %d", mn, DATE_Y(date1));
					}
				}
				break;
			case DATE__Y:
				if(japanese) {
					doc_printf(doc, "%d年", DATE_Y(date1));
				} else {
					doc_printf(doc, "%d", DATE_Y(date1));
				}
				break;
			}
#else
			if(japanese) {
				doc_printf(doc, "%d年", DATE_Y(date1));
				if(ymd==DATE__YMD || ymd==DATE__YM) {
					if(!(0<DATE_M(date1) && DATE_M(date1)<=12)) {
						edb_print_alert(ep, EP_ALERT, "%d月", DATE_M(date1));
					} else {
						doc_printf(doc, "%d月", DATE_M(date1));
					}
				}
				if(ymd==DATE__YMD) {
					if(!(0<DATE_D(date1) && DATE_D(date1)<=31)) {
						edb_print_alert(ep, EP_ALERT, "%d日", DATE_D(date1));
					} else {
						doc_printf(doc, "%d日", DATE_D(date1));
					}
				}
			} else {
				if(ymd==DATE__YMD || ymd==DATE__YM) {
					if(!(0<DATE_M(date1) && DATE_M(date1)<=12)) {
						edb_print_alert(ep, EP_ALERT, "month");
					} else {
						TeX(doc) {
							doc_printf(doc, "%s\\ ", mn);
						} else {
							doc_printf(doc, "%s ", mn);
						}
					}
				}
				if(ymd==DATE__YMD) {
					if(!(0<DATE_D(date1) && DATE_D(date1)<=31)) {
						edb_print_alert(ep, EP_ALERT, "%d, ", DATE_D(date1));
					} else {
						doc_printf(doc, "%d, ", DATE_D(date1));
					}
				}
				doc_printf(doc, "%d", DATE_Y(date1));
			}
#endif
		}
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return TRUE;
}

int
edb_print_type_DATE(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{

	return edb_print_type_DATE_print_ymd(ep, tuple, tc, datum, DATE__YM);
}

int
edb_print_type_DATE_YMD(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{

	return edb_print_type_DATE_print_ymd(ep, tuple, tc, datum, DATE__YMD);
}

int
edb_print_type_DATE_YM(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{

	return edb_print_type_DATE_print_ymd(ep, tuple, tc, datum, DATE__YM);
}

int
edb_print_type_DATE_Y(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{

	return edb_print_type_DATE_print_ymd(ep, tuple, tc, datum, DATE__Y);
}

static
int
edb_print_type_DATE2(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	static char	*mname[] = {
	"Jan.",	"Feb.",	"March",	"April",	"May",	"June",
	"July",	"Aug.",	"Sep.",	"Oct.",	"Nov.",	"Dec."
	};
	int	status = FALSE;
	int	m, date1, date2;

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	edb_datum_get_DATE2(ep->ec, datum, &date1, &date2);
	LaTeX(ep) {
		doc_printf(doc, "\\EdbDatePeriod{%d}{%d}", date1, date2);
		status = TRUE;
	} else if(DocIsXML(doc)) {
		doc_printf(doc, "%08d %08d", date1, date2);
		status = TRUE;
	} else if(EpLang(ep)==EP_LANG_Japanese) {
		if(date1==EDB_DATE_MIN && date2==EDB_DATE_MAX) {
			status = edb_print_alert_column(ep, 0, tc->ci);
		} else if(date1==EDB_DATE_MIN) {
			doc_text_puts(doc, "〜");
			if(date2<EDB_DATE_MAX) {
				if(DATE_Y(date1) != DATE_Y(date2)) doc_text_printf(doc, "%d年", DATE_Y(date2));
				doc_text_printf(doc, "%d月", DATE_M(date2));
			}
			status = TRUE;
		} else {
			doc_text_printf(doc, "%d年%d月", DATE_Y(date1), DATE_M(date1));
			if(!(date1==date2 || (DATE_Y(date1)==DATE_Y(date2) && DATE_M(date1)==DATE_M(date2)))) {
				doc_text_puts(doc, "〜");
				if(date2<EDB_DATE_MAX) {
					if(DATE_Y(date1) != DATE_Y(date2)) doc_text_printf(doc, "%d年", DATE_Y(date2));
					doc_text_printf(doc, "%d月", DATE_M(date2));
				}
			}
			status = TRUE;
		}
	} else {
		if(date1==EDB_DATE_MIN && date2==EDB_DATE_MAX) {
			status = edb_print_alert_column(ep, 0, tc->ci);
		} else if(date2<EDB_DATE_MAX) {
			if(DATE_Y(date1) != DATE_Y(date2)) {
				m = DATE_M(date1);
				if(1<=m && m<=12) doc_text_printf(doc, "%s %d", mname[m-1], DATE_Y(date1));
				HTML(doc) doc_puts(doc, "-");
				TeX(doc) doc_puts(doc, "$\\sim$");
				PLAIN(doc) doc_puts(doc, "-");
			} else if(DATE_M(date1) != DATE_M(date2)) {
				m = DATE_M(date1);
				if(1<=m && m<=12) doc_text_printf(doc, "%s", mname[m-1]);
				HTML(doc) doc_puts(doc, "-");
				TeX(doc) doc_puts(doc, "$\\sim$");
				PLAIN(doc) doc_puts(doc, "-");
			}
			m = DATE_M(date2);
			if(1<=m && m<=12) {
				HTML(doc) doc_printf(doc, "%s %d", mname[m-1], DATE_Y(date2));
				TeX(doc) doc_printf(doc, "%s\\ %d", mname[m-1], DATE_Y(date2));
				PLAIN(doc) doc_printf(doc, "%s %d", mname[m-1], DATE_Y(date2));
			} else {
				doc_printf(doc, "%d", DATE_Y(date2));
			}
			status = TRUE;
		} else {
			m = DATE_M(date1);
			if(1<=m && m<=12) {
				HTML(doc) doc_printf(doc, "%s %d", mname[m-1], DATE_Y(date1));
				TeX(doc) doc_printf(doc, "%s\\ %d", mname[m-1], DATE_Y(date1));
				PLAIN(doc) doc_printf(doc, "%s %d", mname[m-1], DATE_Y(date1));
			} else {
				doc_printf(doc, "%d", DATE_Y(date1));
			}
			status = TRUE;
		}
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

static
int
edb_print_type_PHONE(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	register int	status = FALSE;

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	if(DatumEnglishIsUsable(datum)) {
		int	len;

		if(!strncmp((char *)DatumEnglish(datum), PHONE_EXCHANGE_NO, len=strlen(PHONE_EXCHANGE_NO))) {
			doc_text_puts(doc, (char *)&DatumEnglish(datum)[len]);
		} else {
			doc_text_puts(doc, (char *)DatumEnglish(datum));
		}
		status = TRUE;
	} else {
		status = edb_print_alert_column(ep, 0, tc->ci);
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

int
edb_print_emailaddress_in_HTML_avoid_SPAM(Doc doc, char *emailaddress)
{
	char	*s = emailaddress;
	static char *alpha[] = {
		"\xc4\x81",	/* a */
		"b",	/* b */
		"\xc4\x8b",	/* c */
		"\xc4\x91",	/* d */
		"\xc4\x97",	/* e */
		"f",	/* f */
		"\xc4\xa1",	/* g */
		"h",	/* h */
		"\xc4\xab",	/* i */
		"j",	/* j */
		"k",	/* k */
		"l",	/* l */
		"m",	/* m */
		"\xc5\x86",	/* n */
		"\xc5\x8d",	/* o */
		"p",	/* p */
		"q",	/* q */
		"\xc5\x97",	/* r */
		"\xc5\x9b",	/* s */
		"t",	/* t */
		"\xc5\xab",	/* u */
		"v",	/* v */
		"w",	/* w */
		"x",	/* x */
		"y",	/* y */
		"\xc5\xbc",	/* z */
	};

	if(!STRisValid(emailaddress)) return FALSE;
#if 0
	{ int i;
	for(i=0; i<26; ++i) {
		doc_puts(doc, "<span style=\"font-family:cursive\">");
		doc_puts(doc, alpha[i]);
		doc_puts(doc, "</span>");
	}
	}
#endif
	while(*s) {
		switch(*s) {
		case '@':
			doc_printf(doc, "<span style=\"font-family:cursive;color:transparent\">\xc2\xa0(</span>");
			doc_putc(doc, '@');
			doc_printf(doc, "<span style=\"font-family:cursive;color:transparent\">)\xc2\xa0</span>");
			break;
		case '.':
			doc_printf(doc, "<span style=\"font-family:cursive;color:transparent\">\xc2\xa0\xe2\x82\x8d</span>");
			doc_putc(doc, '.');
			doc_printf(doc, "<span style=\"font-family:cursive;color:transparent\">\xe2\x82\x8e\xc2\xa0</span>");
			break;
		default:
			doc_puts(doc, "<span style=\"font-family:cursive\">");
			if('a'<=*s && *s<='z') {
				doc_puts(doc, alpha[*s-'a']);
			} else {
				doc_putc(doc, *s);
			}
			doc_puts(doc, "</span>");
			break;
		}
		++s;
	}
	return TRUE;
}

int
edb_print_type_EMAIL_in_HTML_avoid_SPAM(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;

	if(DatumEnglishIsValid(datum)) {
		return edb_print_emailaddress_in_HTML_avoid_SPAM(ep->doc, (char *)DatumEnglish(datum));
	}
	return FALSE;
}

static
int
edb_print_type_EMAIL(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	register int	status = FALSE;

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	if(DatumEnglishIsUsable(datum)) {
		TeX(doc) {
			char *s;

			doc_printf(doc, "\\EdbEMAIL{");
#if 1
			for(s=(char *)DatumEnglish(datum); *s; ++s) {
				if(*s=='%') doc_putc(doc, '\\');
				doc_putc(doc, *s);
			}
#endif
			doc_puts(doc, "}{");
			doc_text_puts(doc, (char *)DatumEnglish(datum));
			doc_puts(doc, "}");
		} else HTML(doc) {
#if 0
			doc_printf(doc, "<a href=\"mailto:%s\">", (char *)DatumEnglish(datum));
			doc_text_puts(doc, (char *)DatumEnglish(datum));
			doc_puts(doc, "</a>");
#else
			doc_text_puts(doc, (char *)DatumEnglish(datum));
#endif
		} else {
			doc_text_puts(doc, (char *)DatumEnglish(datum));
		}
		status = TRUE;
	} else {
		status = edb_print_alert_column(ep, 0, tc->ci);
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

static
int
edb_print_type_URL(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	register int	status = FALSE;

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	if(DatumEnglishIsUsable(datum)) {
		TeX(doc) {
			char *s;

			doc_printf(doc, "\\EdbURL{");
#if 1
			for(s=(char *)DatumEnglish(datum); *s; ++s) {
				if(*s=='%' || *s=='#') doc_putc(doc, '\\');
				doc_putc(doc, *s);
			}
#endif
			doc_puts(doc, "}{");
			doc_text_puts(doc, (char *)DatumEnglish(datum));
			doc_puts(doc, "}");
		} else HTML(doc) {
			doc_printf(doc, "<a href=\"%s\">", (char *)DatumEnglish(datum));
			doc_text_puts(doc, (char *)DatumEnglish(datum));
			doc_puts(doc, "</a>");
		} else {
			doc_text_puts(doc, (char *)DatumEnglish(datum));
		}
		status = TRUE;
	} else {
		status = edb_print_alert_column(ep, 0, tc->ci);
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

static
int
edb_print_person_name(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	EdbTupleInfo ptuple;
	EdbContext	ec = ep->ec;
	EdbColumnInfo	ci = tc->ci;
	int	alert = FALSE;
	int	i, nfields;
	int	mode = ep->ct->tcd_mode;
	int	japanese = LangIsJapanese(ep);

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	if(datum && DatumEID(datum) && (ptuple = edb_tuple_get(ec, DatumEID(datum), 0))) {
		edb_book_print_person_info_name(ep, ptuple, japanese, mode);
		edb_print_tcd_postfix(ep, tuple, tc, datum);
		return TRUE;
	} else if(japanese) {
		if(DatumEnglishIsValid(datum) && !DatumJapaneseIsValid(datum)) {
			 ITEMIZE(ep) {
			} else {
				japanese = FALSE;
			}
		}
	} else {
		if(!DatumEnglishIsValid(datum) && DatumJapaneseIsValid(datum)) {
			 ITEMIZE(ep) {
			} else {
				japanese = TRUE;
			}
			alert = TRUE;
		}
	}
	if(alert) {
		if(!ep->ct->nocolor) doc_color_begin(ep->doc, ALERT_COLOR);
	}
	if(japanese && DatumJapaneseIsValid(datum)) {
		if(DatumJapanese(datum)[0]=='.') {
			doc_text_puts(doc, (char *)&(DatumJapanese(datum)[1]));
		} else {
			char	*s, *str;
			Link	lstr, ls;

			str = STRDUP((char *)DatumJapanese(datum));
			for(s=str; *s; ++s) {
				if(*s==',') *s = ' ';
			}
			str_oneline(str);
			lstr = str_split(str, ' ');

			TeX(doc) doc_puts(doc, "\\EdbSMG{");
			switch(nfields=link_atoms(lstr)) {
			case 0:
				edb_print_alert(ep, EP_ALERT, "%s", "姓");
				HTML(doc) doc_putc(doc, ' ');
				TeX(doc) doc_puts(doc, "}{}{");
				PLAIN(doc) doc_putc(doc, ' ');
				edb_print_alert(ep, EP_ALERT, "%s", "名");
				break;
			case 1:
				doc_text_puts(doc, lstr->d);
				HTML(doc) doc_putc(doc, ' ');
				TeX(doc) doc_puts(doc, "}{}{");
				PLAIN(doc) doc_putc(doc, ' ');
				edb_print_alert(ep, EP_ALERT, "%s", "名");
				break;
			case 2:
				doc_text_puts(doc, lstr->d);
				HTML(doc) doc_putc(doc, ' ');
				TeX(doc) doc_puts(doc, "}{}{");
				PLAIN(doc) doc_putc(doc, ' ');
				doc_text_puts(doc, lstr->n->d);
				break;
			case 3:
			default:
				s = link_get_obj(lstr, 0);
				doc_text_puts(doc, s);
				TeX(doc) doc_puts(doc, "}{");
				for(i=1; i<nfields-1; ++i) {
					HTML(doc) doc_putc(doc, ' ');
					TeX(doc) if(i>1) doc_puts(doc, "\\,");
					PLAIN(doc) doc_putc(doc, ' ');
					s = link_get_obj(lstr, i);
					doc_text_puts(doc, s);
				}
				HTML(doc) doc_putc(doc, ' ');
				TeX(doc) doc_puts(doc, "}{");
				PLAIN(doc) doc_putc(doc, ' ');
				s = link_get_obj(lstr, nfields-1);
				doc_text_puts(doc, s);
				break;
			}
			TeX(doc) doc_puts(doc, "}");
		}
	} else if(DatumEnglishIsValid(datum)) {
		if(DatumEnglish(datum)[0]=='.') {
			doc_text_puts(doc, (char *)&(DatumEnglish(datum)[1]));
		} else {
			char	*s, *str;
			Link	lstr;

			str = STRDUP((char *)DatumEnglish(datum));
			for(s=str; *s; ++s) {
				if(*s==',') *s = ' ';
			}
			str_oneline(str);
			lstr = str_split(str, ' ');

			switch(nfields=link_atoms(lstr)) {
			case 0:
				TeX(doc) doc_puts(doc, "\\EdbGMS{");
				edb_print_alert(ep, EP_ALERT, "%s", "名");
				HTML(doc) doc_putc(doc, ' ');
				TeX(doc) doc_puts(doc, "}{}{");
				PLAIN(doc) doc_putc(doc, ' ');
				edb_print_alert(ep, EP_ALERT, "%s", "姓");
				TeX(doc) doc_puts(doc, "}");
				break;
			case 1:
				TeX(doc) doc_puts(doc, "\\EdbGMS{");
				edb_print_alert(ep, EP_ALERT, "%s", "名");
				HTML(doc) doc_putc(doc, ' ');
				TeX(doc) doc_puts(doc, "}{}{");
				PLAIN(doc) doc_putc(doc, ' ');
				doc_text_puts(doc, lstr->d);
				TeX(doc) doc_puts(doc, "}");
				break;
			case 2:
				s = lstr->n->d;
				TeX(doc) doc_puts(doc, "\\EdbGMS{");
				if((ep->ct->person_name_mode==EP_PERSON_FULLNAME) || islower(s[0])) {
					if(strchr(s, '.') || strlen(s)<2) {
						if(!ep->ct->nocolor) doc_color_begin(ep->doc, WARNING_COLOR);
						doc_text_puts(doc, s);
						if(!ep->ct->nocolor) doc_color_end(ep->doc);
					} else {
						doc_text_puts(doc, s);
					}
				} else {
					doc_printf(doc, "%c.", s[0]);
				}
				HTML(doc) doc_putc(doc, ' ');
				TeX(doc) doc_puts(doc, "}{}{");
				PLAIN(doc) doc_putc(doc, ' ');
				s = lstr->d;
				if(strchr(s, '.') || strlen(s)<2) {
					if(!ep->ct->nocolor) doc_color_begin(ep->doc, WARNING_COLOR);
					doc_text_puts(doc, s);
					if(!ep->ct->nocolor) doc_color_end(ep->doc);
				} else {
					doc_text_puts(doc, s);
				}
				TeX(doc) doc_puts(doc, "}");
				break;
			case 3:
			default:
				s = link_get_obj(lstr, nfields-1);
				TeX(doc) doc_puts(doc, "\\EdbGMS{");
				if((ep->ct->person_name_mode==EP_PERSON_FULLNAME) || islower(s[0])) {
					if(strchr(s, '.') || strlen(s)<2) {
						if(!ep->ct->nocolor) doc_color_begin(ep->doc, WARNING_COLOR);
						doc_text_puts(doc, s);
						if(!ep->ct->nocolor) doc_color_end(ep->doc);
					} else {
						doc_text_puts(doc, s);
					}
				} else {
					doc_printf(doc, "%c.", s[0]);
				}
				TeX(doc) doc_puts(doc, "}{");
				for(i=1; i<nfields-1; ++i) {
					HTML(doc) doc_putc(doc, ' ');
					TeX(doc) if(i>1) doc_puts(doc, "\\,");
					PLAIN(doc) doc_putc(doc, ' ');
					s = link_get_obj(lstr, i);
					if((ep->ct->person_name_mode==EP_PERSON_FULLNAME) || islower(s[0])) {
						doc_text_puts(doc, s);
					} else {
						doc_printf(doc, "%c.", s[0]);
					}
				}
				HTML(doc) doc_putc(doc, ' ');
				PLAIN(doc) doc_putc(doc, ' ');
				TeX(doc) doc_puts(doc, "}{");
				s = link_get_obj(lstr, 0);
				if(strchr(s, '.') || strlen(s)<2) {
					if(!ep->ct->nocolor) doc_color_begin(ep->doc, WARNING_COLOR);
					doc_text_puts(doc, s);
					if(!ep->ct->nocolor) doc_color_end(ep->doc);
				} else {
					doc_text_puts(doc, s);
				}
				TeX(doc) doc_puts(doc, "}");
				break;
			}
		}
	} else {
		edb_print_alert_column(ep, EP_ALERT, ci);
	}
	if(alert) if(!ep->ct->nocolor) doc_color_end(ep->doc);
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return TRUE;
}

static
int
edb_print_page_number(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	int	status = FALSE;
	int	num;
	EdbDatum	d0, d1;
	char	*p1, *p2;

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	p1 = p2 = NULL;
	if(DatumEnglishIsValid(datum)) {
		Link	lstr;

		lstr = str_split((char *)DatumEnglish(datum), ' ');
		num = link_atoms(lstr);
		if(num>=2) {
			p1 = STRDUP(lstr->d);
			p2 = STRDUP(lstr->n->d);
			if(!strcmp(p1, p2)) p2 = NULL;
		} else if(num==1) {
			p1 = STRDUP(lstr->d);
		}
		link_free(lstr);
	}
	LaTeX(ep) {
		if(p1 && p2) {
			doc_text_puts3(doc, "\\EdbPage{", p1, "}");
			doc_text_puts3(doc, "{", p2, "}");
			status = TRUE;
		} else if(p1) {
			doc_text_puts3(doc, "\\EdbPage{", p1, "}");
			doc_text_puts3(doc, "{", p1, "}");
			status = TRUE;
		} else {
			status = edb_print_alert_column(ep, 0, tc->ci);
		}
	} else if(DocIsXML(doc)) {
		if(p1 && p2) {
			doc_text_puts(doc, p1);
			doc_text_puts(doc, " ");
			doc_text_puts(doc, p2);
			status = TRUE;
		} else if(p1) {
			doc_text_puts(doc, p1);
			doc_text_puts(doc, " ");
			doc_text_puts(doc, p1);
			status = TRUE;
		}
	} else {
#if 0	// change by Alex @ Sep. 8, 2017
		if(p1 && p2) {
			if(LangIsJapanese(ep)) {
				doc_text_puts(doc, p1);
				HTML(doc) doc_puts(doc, "-");
				pTeX(ep) doc_puts(doc, "--");
				LaTeX(ep) doc_puts(doc, "--");
				PLAIN(doc) doc_puts(doc, "-");
				doc_text_puts3(doc, NULL, p2, "頁");
			} else {
				doc_text_puts3(doc, "pp.", p1, NULL);
				HTML(doc) doc_puts(doc, "-");
				pTeX(ep) doc_puts(doc, "--");
				LaTeX(ep) doc_puts(doc, "--");
				PLAIN(doc) doc_puts(doc, "-");
				doc_text_puts(doc, p2);
			}
			status = TRUE;
		} else if(p1) {
			if(LangIsJapanese(ep)) {
				doc_text_puts3(doc, NULL, p1, "頁");
			} else {
				doc_text_puts3(doc, "p.", p1, NULL);
			}
			status = TRUE;
		} else {
			edb_print_alert_column(ep, 0, tc->ci);
		}
#else
		if(p1 && p2) {
			doc_text_puts(doc, p1);
			HTML(doc) doc_puts(doc, "-");
			pTeX(ep) doc_puts(doc, "--");
			LaTeX(ep) doc_puts(doc, "--");
			PLAIN(doc) doc_puts(doc, "-");
			doc_text_puts(doc, p2);
			status = TRUE;
		} else if(p1) {
			doc_text_puts(doc, p1);
			status = TRUE;
		} else {
			edb_print_alert_column(ep, 0, tc->ci);
		}
#endif
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

int
edb_print_patent_number(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	EdbColumnInfo		ci = tc->ci;
	char	*s;
	int	status;
	char	*prefix = "";
	char	*postfix = "";
	int	japanese = LangIsJapanese(ep);

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	if(japanese) {
		if(!strcmp((char *)ColumnXN(ci), "patent.apply")) {
			prefix = "特願";
			postfix = "";
		} else if(!strcmp((char *)ColumnXN(ci), "patent.disclosure")) {
			prefix = "特開";
			postfix = "";
		} else if(!strcmp((char *)ColumnXN(ci), "patent.number")) {
			prefix = "特許第";
			postfix = "号";
		}
	}
	if(DatumEnglishIsValid(datum)) {
		s = (char *)DatumEnglish(datum);
		doc_puts(doc, prefix);
		if(japanese) {
			switch(*s) {
			case 'M':	doc_text_puts(doc, "明");	++s;	break;
			case 'T':	doc_text_puts(doc, "大");	++s;	break;
			case 'S':	doc_text_puts(doc, "昭");	++s;	break;
			case 'H':	doc_text_puts(doc, "平");	++s;	break;
			case 'R':	doc_text_puts(doc, "令");	++s;	break;
			default:	break;
			}
		}
		doc_text_puts(doc, s);
		doc_puts(doc, postfix);
		status = TRUE;
	} else {
		status = edb_print_alert_column(ep, 0, ci);
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

int
edb_print_pmodel_number(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	register Doc	doc = ep->doc;
	EdbColumnInfo		ci = tc->ci;
	char	*s;
	int	status;
	int	japanese = LangIsJapanese(ep);

	edb_print_tcd_prefix(ep, tuple, tc, datum);
	if(DatumEnglishIsValid(datum)) {
		s = (char *)DatumEnglish(datum);
		if(japanese) {
			switch(*s) {
			case 'M':	doc_text_puts(doc, "明");	++s;	break;
			case 'T':	doc_text_puts(doc, "大");	++s;	break;
			case 'S':	doc_text_puts(doc, "昭");	++s;	break;
			case 'H':	doc_text_puts(doc, "平");	++s;	break;
			case 'R':	doc_text_puts(doc, "令");	++s;	break;
			default:	break;
			}
		}
		doc_text_puts(doc, s);
		status = TRUE;
	} else {
		status = edb_print_alert_column(ep, 0, ci);
	}
	edb_print_tcd_postfix(ep, tuple, tc, datum);
	return status;
}

static
int
edb_print_type_duplicate_info_func(BiTree_obj_t obj, void *data)
{
	register struct EDB_PRINT_TYPE_INFO	*src = obj;
	register BiTree						*new_bt = data;
	register struct EDB_PRINT_TYPE_INFO	*dst;

	dst = malloc(sizeof(struct EDB_PRINT_TYPE_INFO));
	bcopy(src, dst, sizeof(struct EDB_PRINT_TYPE_INFO));
	BiTreeAdd(*new_bt, dst, dst->typeID, NULL, NULL, NULL);
	return TRUE;
}

BiTree
edb_print_type_duplicate_info(BiTree bt)
{
	BiTree	new = NULL;

	bitree_callback(bt, edb_print_type_duplicate_info_func, &new);
	return new;
}

int
edb_print_type_register_func(EdbPrint ep, int typeID, int style, EdbPrintTCDFunc_t print_type)
{
	struct EDB_PRINT_TYPE_INFO *info;

	if(style<0 || EP_PrintStyle_Variations<=style) return FALSE;
	if(!ep->ct->print_type_tobe_free) {
		BiTree	bt;

		bt = ep->ct->bt_print_type_info;
		ep->ct->bt_print_type_info = edb_print_type_duplicate_info(bt);
		ep->ct->print_type_tobe_free = TRUE;
	}
	info = bitree_search(ep->ct->bt_print_type_info, NULL, typeID, NULL);
	if(info==NULL) {
		info = malloc(sizeof(struct EDB_PRINT_TYPE_INFO));
		if(info==NULL) return FALSE;
		bzero(info, sizeof(struct EDB_PRINT_TYPE_INFO));
		info->typeID = typeID;
		BiTreeAdd(ep->ct->bt_print_type_info, info, info->typeID, NULL, NULL, NULL);
	}
	info->func[style] = print_type;
	return TRUE;
}

void
edb_print_type_initialize(EdbPrint ep)
{
static struct {
	int				typeID;
	EdbPrintTCDFunc_t	std;
	EdbPrintTCDFunc_t	xml;
	EdbPrintTCDFunc_t	csv;
	EdbPrintTCDFunc_t	item;
	EdbPrintTCDFunc_t	cap;
} info2print[] = {{
	EDB_TYPE_TEXT,
	edb_print_type_TEXT,
	NULL,
	NULL,
	NULL,
	edb_print_type_TEXT,
},{
	EDB_TYPE_NAME,
	edb_print_type_NAME,
	NULL,
	NULL,
	NULL,
	edb_print_type_NAME,
},{
	EDB_TYPE_SENTENCE,
	edb_print_type_NAME,
	NULL,
	NULL,
	NULL,
	edb_print_type_NAME,
},{
	EDB_TYPE_PARAGRAPH,
	edb_print_type_NAME,
	NULL,
	NULL,
	NULL,
	edb_print_type_NAME,
},{
	EDB_TYPE_MONETARY,
	edb_print_type_MONETARY,
	NULL,
	NULL,
	NULL,
	edb_print_type_MONETARY,
},{
	EDB_TYPE_DATE,
	edb_print_type_DATE,
	NULL,
	NULL,
	edb_print_type_DATE,
	edb_print_type_DATE,
},{
	EDB_TYPE_DATE2,
	edb_print_type_DATE2,
	NULL,
	NULL,
	edb_print_type_DATE2,
	edb_print_type_DATE2,
},{
	EDB_TYPE_PHONE,
	edb_print_type_PHONE,
	NULL,
	NULL,
	NULL,
	edb_print_type_PHONE,
},{
	EDB_TYPE_EMAIL,
	edb_print_type_EMAIL,
	NULL,
	NULL,
	edb_print_type_EMAIL,
	edb_print_type_EMAIL,
},{
	EDB_TYPE_URL,
	edb_print_type_URL,
	NULL,
	NULL,
	edb_print_type_URL,
	edb_print_type_URL,
}};
	int	i;

	for(i=0; i<Ns(info2print); ++i) {
		edb_print_type_register_func(ep, info2print[i].typeID, PrintStyle_STANDARD, info2print[i].std);
		edb_print_type_register_func(ep, info2print[i].typeID, PrintStyle_XML, info2print[i].xml);
		edb_print_type_register_func(ep, info2print[i].typeID, PrintStyle_CSV, info2print[i].csv);
		edb_print_type_register_func(ep, info2print[i].typeID, PrintStyle_ITEMIZE, info2print[i].item);
		edb_print_type_register_func(ep, info2print[i].typeID, PrintStyle_CAPTION, info2print[i].cap);
	}
}

int
edb_print_type(EdbPrint ep, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	EdbContext	ec = ep->ec;
	Doc	doc = ep->doc;
	struct EDB_PRINT_TYPE_INFO		*type_info;
	register EdbPrintTCDFunc_t		print = NULL;
	register int	style = PrintStyle(ep);

	if(style<0 || EP_PrintStyle_Variations<=style) return FALSE;

	if(tuple == NULL) return edb_print_alert(ep, EP_ALERT, "!!!");
	if(tc == NULL) return edb_print_alert(ep, EP_ALERT, "???");
	if(datum==NULL || !DatumAvail(datum)) return edb_print_alert_column(ep, 0, tc->ci);

	if(tc->ci->personname) {
		print = edb_print_person_name;
		if(ep->ct->print_person_name[style]) print = ep->ct->print_person_name[style];
	} else if(tc->ci->pagenumber) {
		print = edb_print_page_number;
		if(ep->ct->print_page_number[style]) print = ep->ct->print_page_number[style];
	} else if(tc->ci->patentnumber) {
		print = edb_print_patent_number;
		if(ep->ct->print_patent_number[style]) print = ep->ct->print_patent_number[style];
	} else {
		type_info = bitree_search(ep->ct->bt_print_type_info, NULL, tc->ci->typeInfo->typeID, NULL);
		if(type_info==NULL) type_info = bitree_search(ep->ct->bt_print_type_info, NULL, tc->ci->typeInfo->basicTypeID, NULL);
		if(type_info) print = type_info->func[style];
	}
	if(print) return (*print)(ep, tuple, tc, datum);
	return edb_print_type_default(ep, tuple, tc, datum);
}

