/*
	$Id: utlf.c,v 1.3 2012/04/03 03:51:40 alex Exp alex $
	utlf.c--- utlf
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: utlf.c,v 1.3 2012/04/03 03:51:40 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/publish.h"
#include	"edb/utlf.h"
#include	"edb/pki.h"

#define	E_RAD_ID	"E-RAD.go.jp"
#define	PA_ID	"PA.jim.tokushima-u.ac.jp"
#define	SA_ID	"SA.jim.tokushima-u.ac.jp"

#define	UTLF_RID_FMT		UTLF_NS ":RID=%s/" E_RAD_ID
#define	UTLF_PID_FMT		UTLF_NS ":PID=%s/" PA_ID
#define	UTLF_StaffID_FMT	UTLF_NS ":StaffID=%s/" SA_ID

UTLF
utlf_new(char *id)
{
	UTLF	utlf;

	utlf = (UTLF)malloc(sizeof(struct _UTLF));
	if(utlf==NULL) return NULL;
	bzero(utlf, sizeof(struct _UTLF));
	if(id) utlf->id = strdup(id);
	return utlf;
}

void
utlf_free(UTLF utlf)
{

	FREE(utlf->id);
	FREE(utlf->title);
	FREE(utlf->title);
	FREE(utlf->creator);
	FREE(utlf->date);
	FREE(utlf->description);
	FREE(utlf->publisher);
	FREE(utlf->format);
	FREE(utlf->replaces);
	FREE(utlf);
}

static
char *
utlf_xml_text_encode(register char *s)
{
	register char *d, *dst, c;

	dst = d = (char *)alloca(sizeof(char)*strlen(s)*10);
	while((c=*s++)!='\0') {
		switch(c) {
		case	'<':	strcpy(d, "&lt;");	d += 4;	break;
		case	'>':	strcpy(d, "&gt;");	d += 4;	break;
		case	'&':	strcpy(d, "&amp;");	d += 5;	break;
		case	'\'':	strcpy(d, "&apos;");	d += 6;	break;
		case	'"':	strcpy(d, "&quot;");	d += 6;	break;
		default:	*d++ = c;			break;
		}
	}
	*d = '\0';
	return strdup(dst);
}

void
utlf_set_title(UTLF utlf, char *title)
{

	FREE(utlf->title);
	utlf->title = strdup(title);
}

void
utlf_set_creator(UTLF utlf, char *creator)
{

	FREE(utlf->creator);
	utlf->creator = strdup(creator);
}

void
utlf_set_creator_by_EID(UTLF utlf, eid_t op)
{

	FREE(utlf->creator);
	if(EIDisValid(op)) asprintf(&utlf->creator, UTLF_EID_FMT, op);
}

void
utlf_set_date_now(UTLF utlf)
{
	time_t	t;
	struct tm	tm;

	FREE(utlf->date);
	time(&t);
	gmtime_r(&t, &tm);
	//	<dc:date>2008-10-25T14:51:16,879Z</dc:date>
	asprintf(&utlf->date, "%04d-%02d-%02dT%02d:%02d:%02d,000Z",
		tm.tm_year+1900,
		tm.tm_mon + 1,
		tm.tm_mday,
		tm.tm_hour,
		tm.tm_min,
		tm.tm_sec);
}

void
utlf_set_description(UTLF utlf, char *description)
{

	FREE(utlf->description);
	utlf->description = strdup(description);
}

void
utlf_set_publisher(UTLF utlf, char *publisher)
{

	FREE(utlf->publisher);
	utlf->publisher = strdup(publisher);
}

void
utlf_set_format(UTLF utlf, char *format)
{

	FREE(utlf->format);
	utlf->format = strdup(format);
}

void
utlf_set_replaces(UTLF utlf, char *replaces)
{

	FREE(utlf->replaces);
	utlf->replaces = strdup(replaces);
}

static
Link
utlf_make_rdfxml(UTLF utlf, Link lcontent)
{
	Link	lstr = NULL, l;

	LinkAppend(lstr, link_atom_new("<?xml version=\"1.0\" encoding=\"UTF-8\"?>", NULL));
	LinkAppend(lstr, link_atom_new("<!DOCTYPE rdf:RDF[", NULL));
	LinkAppend(lstr, link_atom_new(" 	<!ENTITY dctypes 'http://purl.org/dc/dcmitype/'>", NULL));
	LinkAppend(lstr, link_atom_new("	<!ENTITY utlf 'http://utlf.db.tokushima-u.ac.jp/UTLF/utlf#'>", NULL));
	LinkAppend(lstr, link_atom_new("	<!ENTITY dcterms 'http://purl.org/dc/terms/'>", NULL));
	LinkAppend(lstr, link_atom_new("	<!ENTITY dc 'http://purl.org/dc/elements/1.1/'>", NULL));
	LinkAppend(lstr, link_atom_new("	<!ENTITY rdfs 'http://www.w3.org/2000/01/rdf-schema#'>", NULL));
	LinkAppend(lstr, link_atom_new("	<!ENTITY rdf 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'>", NULL));
	LinkAppend(lstr, link_atom_new("	<!ENTITY owl 'http://www.w3.org/2002/07/owl#'>", NULL));
	LinkAppend(lstr, link_atom_new(" ]>", NULL));
	LinkAppend(lstr, link_atom_new("<rdf:RDF ", NULL));
	LinkAppend(lstr, link_atom_new("	xmlns:dctypes=\"http://purl.org/dc/dcmitype/\"", NULL));
	LinkAppend(lstr, link_atom_new("	xmlns:utlf=\"http://utlf.db.tokushima-u.ac.jp/UTLF/utlf#\"", NULL));
	LinkAppend(lstr, link_atom_new("	xmlns:dcterms=\"http://purl.org/dc/terms/\"", NULL));
	LinkAppend(lstr, link_atom_new("	xmlns:dc=\"http://purl.org/dc/elements/1.1/\"", NULL));
	LinkAppend(lstr, link_atom_new("	xmlns:rdfs=\"http://www.w3.org/2000/01/rdf-schema#\"", NULL));
	LinkAppend(lstr, link_atom_new("	xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"", NULL));
	LinkAppend(lstr, link_atom_new("	xmlns:owl=\"http://www.w3.org/2002/07/owl#\" ", NULL));
	LinkAppend(lstr, link_atom_new(">", NULL));
	lstr = lstr_append_printf(lstr, "	<rdf:Description rdf:about=\"%s\">", utlf_xml_text_encode(STRisValid(utlf->id) ? utlf->id : UTLF_CONTAINER_ID));
	if(STRisValid(utlf->title)) {
		lstr = lstr_append_printf(lstr, "		<dc:title>%s</dc:title>", utlf_xml_text_encode(utlf->title));
	}
	if(STRisValid(utlf->creator)) {
		lstr = lstr_append_printf(lstr, "		<dc:creator>%s</dc:creator>", utlf_xml_text_encode(utlf->creator));
	}
	if(STRisValid(utlf->date)) {
		lstr = lstr_append_printf(lstr, "		<dc:date>%s</dc:date>", utlf_xml_text_encode(utlf->date));
	}
	if(STRisValid(utlf->description)) {
		lstr = lstr_append_printf(lstr, "		<dc:description>%s</dc:description>", utlf_xml_text_encode(utlf->description));
	}
	if(STRisValid(utlf->publisher)) {
		lstr = lstr_append_printf(lstr, "		<dc:publisher>%s</dc:publisher>", utlf_xml_text_encode(utlf->publisher));
	}
	if(STRisValid(utlf->format)) {
		lstr = lstr_append_printf(lstr, "		<dc:format>%s</dc:format>", utlf_xml_text_encode(utlf->format));
	}
	if(STRisValid(utlf->replaces)) {
		lstr = lstr_append_printf(lstr, "		<dcterms:replaces>%s</dcterms:replaces>", utlf_xml_text_encode(utlf->replaces));
	}

	LinkAppend(lstr, link_atom_new("		<utlf:content rdf:parseType=\"Literal\">", NULL));
	LinkAppend(lstr, lcontent);
	LinkAppend(lstr, link_atom_new("		</utlf:content>", NULL));

	LinkAppend(lstr, link_atom_new("	</rdf:Description>", NULL));
	LinkAppend(lstr, link_atom_new("</rdf:RDF>", NULL));
	return lstr;
}

void
utlf_print_header(UTLF utlf, FILE *fp)
{
	char	*s;

	fputs("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", fp);
	fputs("<!DOCTYPE rdf:RDF[\n", fp);
	fputs(" 	<!ENTITY dctypes 'http://purl.org/dc/dcmitype/'>\n", fp);
	fputs("	<!ENTITY utlf 'http://utlf.db.tokushima-u.ac.jp/UTLF/utlf#'>\n", fp);
	fputs("	<!ENTITY dcterms 'http://purl.org/dc/terms/'>\n", fp);
	fputs("	<!ENTITY dc 'http://purl.org/dc/elements/1.1/'>\n", fp);
	fputs("	<!ENTITY rdfs 'http://www.w3.org/2000/01/rdf-schema#'>\n", fp);
	fputs("	<!ENTITY rdf 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'>\n", fp);
	fputs("	<!ENTITY owl 'http://www.w3.org/2002/07/owl#'>\n", fp);
	fputs(" ]>\n", fp);
	fputs("<rdf:RDF \n", fp);
	fputs("	xmlns:dctypes=\"http://purl.org/dc/dcmitype/\"\n", fp);
	fputs("	xmlns:utlf=\"http://utlf.db.tokushima-u.ac.jp/UTLF/utlf#\"\n", fp);
	fputs("	xmlns:dcterms=\"http://purl.org/dc/terms/\"\n", fp);
	fputs("	xmlns:dc=\"http://purl.org/dc/elements/1.1/\"\n", fp);
	fputs("	xmlns:rdfs=\"http://www.w3.org/2000/01/rdf-schema#\"\n", fp);
	fputs("	xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"\n", fp);
	fputs("	xmlns:owl=\"http://www.w3.org/2002/07/owl#\" \n", fp);
	fputs(">\n", fp);
	fprintf(fp, "	<rdf:Description rdf:about=\"%s\">\n", s = utlf_xml_text_encode(STRisValid(utlf->id) ? utlf->id : UTLF_CONTAINER_ID));
	FREE(s);
	if(STRisValid(utlf->title)) {
		fprintf(fp, "		<dc:title>%s</dc:title>\n", s = utlf_xml_text_encode(utlf->title));
		FREE(s);
	}
	if(STRisValid(utlf->creator)) {
		fprintf(fp, "		<dc:creator>%s</dc:creator>\n", s = utlf_xml_text_encode(utlf->creator));
		FREE(s);
	}
	if(STRisValid(utlf->date)) {
		fprintf(fp, "		<dc:date>%s</dc:date>\n", s = utlf_xml_text_encode(utlf->date));
		FREE(s);
	}
	if(STRisValid(utlf->description)) {
		fprintf(fp, "		<dc:description>%s</dc:description>\n", s = utlf_xml_text_encode(utlf->description));
		FREE(s);
	}
	if(STRisValid(utlf->publisher)) {
		fprintf(fp, "		<dc:publisher>%s</dc:publisher>\n", s = utlf_xml_text_encode(utlf->publisher));
		FREE(s);
	}
	if(STRisValid(utlf->format)) {
		fprintf(fp, "		<dc:format>%s</dc:format>\n", s = utlf_xml_text_encode(utlf->format));
		FREE(s);
	}
	if(STRisValid(utlf->replaces)) {
		fprintf(fp, "		<dcterms:replaces>%s</dcterms:replaces>\n", s = utlf_xml_text_encode(utlf->replaces));
		FREE(s);
	}
	utlf->indent = 2;
}

void
utlf_print_trailer(UTLF utlf, FILE *fp)
{

	fputs("	</rdf:Description>\n", fp);
	fputs("</rdf:RDF>\n", fp);
	utlf->indent = 0;
}

static
void
put_indent(FILE *fp, int n)
{
	int	i;

	for(i=0; i<n; ++i) fputs("\t", fp);
}

void
utlf_content_begin(UTLF utlf, FILE *fp)
{

	put_indent(fp, utlf->indent);
	fputs("<utlf:content rdf:parseType=\"Literal\">\n", fp);
	++utlf->indent;
}

void
utlf_content_end(UTLF utlf, FILE *fp)
{

	--utlf->indent;
	put_indent(fp, utlf->indent);
	fputs("</utlf:content>\n", fp);
}

void
utlf_element_begin(UTLF utlf, char *en, FILE *fp)
{

	if(utlf->keyed) fputs("\n", fp);
	put_indent(fp, utlf->indent);
	fprintf(fp, "<%s>\n", en);
	++utlf->indent;
	utlf->keyed = FALSE;
}

void
utlf_element_end(UTLF utlf, char *en, FILE *fp)
{

	--utlf->indent;
	put_indent(fp, utlf->indent);
	fprintf(fp, "</%s>\n", en);
	utlf->keyed = FALSE;
}

void
utlf_element(UTLF utlf, char *en, FILE *fp, char *fmt, ...)
{
	va_list	ap;
	char	*s, *ss = NULL;

	if(!utlf->keyed) put_indent(fp, utlf->indent);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&ss, fmt, ap);
		va_end(ap);
	}
	if(ss) {
		fprintf(fp, "<%s>%s</%s>", en, s = utlf_xml_text_encode(ss), en);
		FREE(ss);
		FREE(s);
	} else {
		fprintf(fp, "<%s/>", en);
	}
	utlf->keyed = !strcmp(en, UTLF_KEY);
	if(!utlf->keyed) fputs("\n", fp);
}

static
Link
lstr_indent(Link lstr, int idx)
{
	Link	l, lstr2 = NULL;
	char	*indent;
	int	i;

	lstr2 = NULL;
	for(i=0; i<idx; ++i) lstr2 = lstr_append_printf(lstr2, "\t");
	indent = str_concatenate0(lstr2);
	link_free(lstr2);

	lstr2 = NULL;
	ForLink(l, lstr) lstr2 = lstr_append_printf(lstr2, "%s%s", indent, (char *)l->d);
	link_free(lstr);
	FREE(indent);
	return lstr2;
}

static	Link	utlf_make_tc_rdfxml	__P((EdbContext ec, EdbTupleInfo tuple, EdbTupleColumn tc));

static
Link
utlf_make_datum_rdfxml(EdbContext ec, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum)
{
	EdbColumnInfo		ci = tc->ci;
	char	*s, *ss, *seeAlso = NULL;
	Link	l, lstr = NULL, lcstr = NULL;
	utf8_t	*en, *ja, *pr;
	Caption	cap = NULL;

	if(datum->l_child) {
		ForLink(l, datum->l_child) {
			LinkAppend(lcstr, utlf_make_tc_rdfxml(ec, tuple, (EdbTupleColumn)l->d));
		}
	}
	if(DatumEIDisValid(datum)) {
		asprintf(&seeAlso, " utlf:seeAlso=\"" UTLF_EID_FMT "\"", DatumEID(datum));
	} else {
		seeAlso = strdup("");
	}
	en = DatumEnglishIsUsable(datum) ? DatumEnglish(datum) : (utf8_t *)"";
	ja = DatumJapaneseIsUsable(datum) ? DatumJapanese(datum) : (utf8_t *)"";
	pr = DatumPronounceIsUsable(datum) ? DatumPronounce(datum) : (utf8_t *)"";
	if(DatumEIDisValid(datum) && ci->typeInfo->multilingual && (cap = edb_make_caption(ec, DatumEID(datum), 0))) {
		en = cap->c_main_english;
		ja = cap->c_main_japanese;
		pr = cap->c_main_pronounce;
	}
	if(en[0] == '.') ++en;
	if(ja[0] == '.') ++ja;
	if(pr[0] == '.') ++pr;
	if(DatumEIDisValid(datum)) {
		if(STRisValid(cap->c_main)) {
			if(lcstr) {
				lstr = lstr_append_printf(lstr, "<utlf:dict>");
				lstr = lstr_append_printf(lstr, "	<utlf:key/><utlf:string%s>%s</utlf:string>", seeAlso, ss = utlf_xml_text_encode((char *)cap->c_main));
				LinkAppend(lstr, lstr_indent(lcstr, 1));
				lstr = lstr_append_printf(lstr, "</utlf:dict>");
			} else {
				lstr = lstr_append_printf(lstr, "<utlf:string%s>%s</utlf:string>", seeAlso, ss = utlf_xml_text_encode((char *)cap->c_main));
			}
			FREE(ss);
		}
	} else if(ci->typeInfo->multilingual) {
		if(STRisValid(ja)) {
			lstr = lstr_append_printf(lstr, "<utlf:key/><utlf:string%s>%s</utlf:string>", seeAlso, ss = utlf_xml_text_encode((char *)ja));
			FREE(ss);
		} else if(STRisValid(en)) {
			lstr = lstr_append_printf(lstr, "<utlf:key/><utlf:string%s>%s</utlf:string>", seeAlso, ss = utlf_xml_text_encode((char *)en));
			FREE(ss);
		}
		if(STRisValid(en)) {
			lstr = lstr_append_printf(lstr, "<utlf:key>English</utlf:key><utlf:string>%s</utlf:string>", ss = utlf_xml_text_encode((char *)en));
			FREE(ss);
		}
		if(STRisValid(ja)) {
			lstr = lstr_append_printf(lstr, "<utlf:key>Japanese</utlf:key><utlf:string>%s</utlf:string>", ss = utlf_xml_text_encode((char *)ja));
			FREE(ss);
		}
		if(STRisValid(pr)) {
			lstr = lstr_append_printf(lstr, "<utlf:key>Pronounce</utlf:key><utlf:string>%s</utlf:string>", ss = utlf_xml_text_encode((char *)pr));
			FREE(ss);
		}
		LinkAppend(lstr, lcstr);
		lstr = lstr_indent(lstr, 1);

		lstr = link_append(link_atom_new("<utlf:dict>", NULL), lstr);
		lstr = link_append(lstr, link_atom_new("</utlf:dict>", NULL));
	} else if(STRisValid(en)) {
		char	*uo = NULL;
		if(ci->typeInfo->basicTypeID==EDB_TYPE_INT) {
			asprintf(&uo, "<utlf:integer%s>%d</utlf:integer>", seeAlso, edb_datum_get_INT(ec, datum));
		} else if(ci->typeInfo->typeID==EDB_TYPE_MONETARY) {
			asprintf(&uo, "<utlf:real%s>%lf</utlf:real>", seeAlso, edb_datum_get_MONETARY(ec, datum)*1000.0);
		} else if(ci->typeInfo->basicTypeID==EDB_TYPE_REAL) {
			asprintf(&uo, "<utlf:real%s>%lf</utlf:real>", seeAlso, edb_datum_get_REAL(ec, datum));
		} else if(ci->typeInfo->basicTypeID==EDB_TYPE_DATE) {
			int	d = edb_datum_get_DATE(ec, datum);
			asprintf(&uo, "<utlf:date%s>%02d-%02d-%02d</utlf:date>", seeAlso, DATE_Y(d), DATE_M(d), DATE_D(d));
		} else if(ci->typeInfo->typeID==EDB_TYPE_URL) {
			asprintf(&uo, "<utlf:reference%s>%s</utlf:reference>", seeAlso, ss = utlf_xml_text_encode((char *)en));
			FREE(ss);
		} else if(!strcmp("person.personnelnumber", (char *)ColumnXN(ci))) {
			asprintf(&uo, "<utlf:reference%s>" UTLF_PID_FMT "</utlf:reference>", seeAlso, ss = utlf_xml_text_encode((char *)en));
			FREE(ss);
		} else if(!strcmp("person.sastaffid", (char *)ColumnXN(ci))) {
			asprintf(&uo, "<utlf:reference%s>" UTLF_StaffID_FMT "</utlf:reference>", seeAlso, ss = utlf_xml_text_encode((char *)en));
			FREE(ss);
		} else if(!strcmp("person.gasr", (char *)ColumnXN(ci))) {
			asprintf(&uo, "<utlf:reference%s>" UTLF_RID_FMT "</utlf:reference>", seeAlso, ss = utlf_xml_text_encode((char *)en));
			FREE(ss);
		} else {
			if(ci->typeInfo->typeID==EDB_TYPE_CERTIFICATE) {
				s = (char *)edb_pki_certificate_to_pem((utf8_t *)en);
				en = (utf8_t *)STRDUP((char *)s);
				FREE(s);
			} else if(ci->typeInfo->typeID==EDB_TYPE_PASSPHRASE) {
				en = (utf8_t *)STRDUP((char *)en);
				for(s = (char *)en; *s!='\0'; ++s) if(*s==' ') *s = '\n';
			}
			if(STRisValid(en)) {
				asprintf(&uo, "<utlf:string%s>%s</utlf:string>", seeAlso, ss = utlf_xml_text_encode((char *)en));
				FREE(ss);
			}
		}
		if(uo) {
			if(lcstr) {
				lstr = lstr_append_printf(lstr, "<utlf:dict>");
				lstr = lstr_append_printf(lstr, "	<utlf:key/>%s", uo);
				LinkAppend(lstr, lstr_indent(lcstr, 1));
				lstr = lstr_append_printf(lstr, "</utlf:dict>");
			} else {
				lstr = lstr_append_printf(lstr, "%s", uo);
			}
		}
	}
	FREE(seeAlso);
	if(cap) edb_caption_free(ec, cap);
	return lstr;
}

static
Link
utlf_make_tc_rdfxml(EdbContext ec, EdbTupleInfo tuple, EdbTupleColumn tc)
{
	EdbColumnInfo		ci = tc->ci;
	EdbDatum	datum;
	char	*ss;
	Link	lstr = NULL;
	char	*cxn;
	Link	l, l_datum = NULL;
	int	num;

	if(ci->obsoleted) return NULL;
	cxn = strrchr((char *)ColumnXN(ci), '.');
	if(cxn) ++cxn;
	ForTCDatum(datum, tc) {
		Link	ldatum_str;
		if(!edb_datum_at_present(ec, datum)) continue;
		ldatum_str = utlf_make_datum_rdfxml(ec, tuple, tc, datum);
		if(ldatum_str==NULL) continue;
		LinkAppend(l_datum, link_atom_new(ldatum_str, NULL));
	}
	num = link_atoms(l_datum);
	switch(num) {
	case 0:
		break;
	case 1:
		if(link_atoms((Link)l_datum->d)==1) {
			lstr = lstr_append_printf(lstr, "<utlf:key>@.%s</utlf:key>%s", cxn, ((Link)l_datum->d)->d);
			link_free((Link)l_datum->d);
		} else {
			lstr = lstr_append_printf(lstr, "<utlf:key>@.%s</utlf:key>", cxn);
			LinkAppend(lstr, (Link)l_datum->d);
		}
		link_free(l_datum);
		break;
	default:
		lstr = lstr_append_printf(lstr, "<utlf:key>@.%s</utlf:key>", cxn);
		lstr = lstr_append_printf(lstr, "<utlf:array>");
		ForLink(l, l_datum) LinkAppend(lstr, lstr_indent((Link)l->d, 1));
		link_free(l_datum);
		lstr = lstr_append_printf(lstr, "</utlf:array>");
		break;
	}
	return lstr;
}

char	*
utlf_make_tuple_rdfxml(EdbContext ec, EdbTupleInfo tuple)
{
	char	*id, *s, *ss, *rdfxml, *indent;
	UTLF		utlf;
	Link	lstr, l;
	Caption cap;
	EdbTableInfo	ti = TupleTable(tuple);
	EdbTupleColumn tc;

	asprintf(&id, "utlf:Form=%s/EID=%d/" UTLF_EDB_ID, TupleXN(tuple), TupleEID(tuple));
	utlf = utlf_new(id);

	utlf_set_creator_by_EID(utlf, ec->operator);

	cap = edb_tuple_make_caption(ec, tuple, 0);

	utlf_set_title(utlf, (char *)cap->c_main);

	asprintf(&s, "【%s】(%s)", ti->name.japanese, cap->c_main);
	utlf_set_description(utlf, s);
	FREE(s);

	utlf_set_publisher(utlf, "utlf:PROGRAM=EDB/" UTLF_EDB_ID);

	asprintf(&s, "utlf:Form=%s/" UTLF_EDB_ID, TABLE_XN(ti));
	utlf_set_format(utlf, s);
	FREE(s);

	asprintf(&s, UTLF_EID_FMT, TupleEID(tuple));
	utlf_set_replaces(utlf, s);
	FREE(s);

	utlf_set_date_now(utlf);

	lstr = NULL;

	lstr = lstr_append_printf(lstr, "<utlf:key>@.EID</utlf:key>");
	lstr = lstr_append_printf(lstr, "<utlf:reference>" UTLF_EID_FMT "</utlf:reference>", TupleEID(tuple));

	lstr = lstr_append_printf(lstr, "<utlf:key>@.CAPTION</utlf:key>");
	lstr = lstr_append_printf(lstr, "<utlf:dict>");
	lstr = lstr_append_printf(lstr, "	<utlf:key/><utlf:string>%s</utlf:string>", ss = utlf_xml_text_encode((char *)cap->c_main));	FREE(ss);
	lstr = lstr_append_printf(lstr, "	<utlf:key>English</utlf:key><utlf:string>%s</utlf:string>", ss = utlf_xml_text_encode((char *)cap->c_main_english));	FREE(ss);
	lstr = lstr_append_printf(lstr, "	<utlf:key>Japanese</utlf:key><utlf:string>%s</utlf:string>", ss = utlf_xml_text_encode((char *)cap->c_main_japanese));	FREE(ss);
	if(STRisValid(cap->c_main_pronounce)) {
		lstr = lstr_append_printf(lstr, "	<utlf:key>Pronounce</utlf:key><utlf:string>%s</utlf:string>", ss = utlf_xml_text_encode((char *)cap->c_main_pronounce));	FREE(ss);
	}
	lstr = lstr_append_printf(lstr, "</utlf:dict>");
	if(EIDisValid(TupleMapto(tuple))) {
		lstr = lstr_append_printf(lstr, "<utlf:key>@.MAP</utlf:key><utlf:reference>" UTLF_EID_FMT "</utlf:reference>", TupleMapto(tuple));
	}
	if(TupleIsPerson(tuple)) {
		if(edb_person_is_user(ec, tuple)) {
			lstr = lstr_append_printf(lstr, "<utlf:key>UID</utlf:key>");
			lstr = lstr_append_printf(lstr, "<utlf:dict>");
			lstr = lstr_append_printf(lstr, "	<utlf:key/><utlf:string>S%d</utlf:string>", TupleEID(tuple));
			lstr = lstr_append_printf(lstr, "	<utlf:key>hasPassword</utlf:key><utlf:%s/>", edb_tuple_seek_column(ec, tuple, (utf8_t *)"@.password")!=NULL ? "true":"false");
			lstr = lstr_append_printf(lstr, "</utlf:dict>");
		}
	}
	if(TupleIsPersonification(tuple)) {
		if(edb_personification_is_member(ec, tuple)) {
			lstr = lstr_append_printf(lstr, "<utlf:key>UID</utlf:key>");
			lstr = lstr_append_printf(lstr, "<utlf:dict>");
			lstr = lstr_append_printf(lstr, "	<utlf:key/><utlf:string>S%d</utlf:string>", TupleEID(tuple));
			lstr = lstr_append_printf(lstr, "	<utlf:key>hasPassword</utlf:key><utlf:%s/>", edb_tuple_seek_column(ec, tuple, (utf8_t *)"@.password")!=NULL ? "true":"false");
			lstr = lstr_append_printf(lstr, "</utlf:dict>");
		}
	}

	if((tc = edb_tuple_seek_column(ec, tuple, (utf8_t *)"@.certificate"))!=NULL) {
		EdbDatum	d;
		if(tc && (d=TCDatumFirst(tc))) {
			EdbPkiCertificateInfo	info = edb_pki_certificate_info(ec, DatumEnglish(d));

			if(info && STRisValid(info->S.DirName) && edb_pki_verify_certificate(ec, DatumEnglish(d))) {
				lstr = lstr_append_printf(lstr, "<utlf:key>X509DN</utlf:key><utlf:string>%s</utlf:string>", info->S.DirName);
			}
			edb_pki_certificate_info_free(ec, info);
		}
	}

	edb_caption_free(ec, cap);

	ForLink(l, tuple->l_column) {
		LinkAppend(lstr, utlf_make_tc_rdfxml(ec, tuple, (EdbTupleColumn)l->d));
	}

	lstr = utlf_make_rdfxml(utlf, lstr_indent(lstr, 3));

	utlf_free(utlf);
	free(id);

	rdfxml = str_concatenate(lstr, '\n');
	link_free(lstr);
	return rdfxml;
}

int
edb_print_tuple_UTLF(EdbPrint ep, EdbTupleInfo tuple)
{
	char	*rdfxml;

	rdfxml = utlf_make_tuple_rdfxml(ep->ec, tuple);
	doc_puts(ep->doc, rdfxml);
	FREE(rdfxml);
	return TRUE;
}

